package edu.gwu.blaise.version2.documents;
import java.util.*;
import java.io.*;
import java.util.jar.*;
import java.util.regex.*;
import java.awt.Color;
import java.awt.Point;
import java.awt.Rectangle;
/**
 * BasicFileHandler does most of the work to take apart a Bls file
 * and make it into a Blaise_Paper object.
 * This class is actually complete
 *
 * @author Michael Gaiman
 * @version 1 4/25/2004
 */
abstract public class BasicFileHandler{

    protected Blaise_Paper paper;
    protected Properties props;
    protected Vector drafts;
    
    /**
     * When true, BasicFileHandler will copy all external files
     * into byteArrays.  When false, BasicFileHandler will not.
     * True by default.
     */
    protected boolean makeArrays=true;
    
    private static final boolean debug=true;

    private JarFile jf;
    
    public final void setJarFile(JarFile jarFile){
        jf=jarFile;
    }
    
    /**
     * Returns a fresh InputStream from the JarFile.
     */
    protected final InputStream getInputStream(JarEntry je) throws IOException{
        return jf.getInputStream(je);
    }
    
    /**
     * Called with the Blaise Paper object to handle.
     * The BlaiseFile object will (the caller) will fill in
     * the general properties that are contained within the paper.props
     * file.  The BFHandler should just create either external links to the
     * attached file structures or it should include them embedded in the 
     * Blaise_Paper.
     */
    public void setPaper(Blaise_Paper paper){
        this.paper=paper;
    }
    
    public Blaise_Paper getPaper(){
        return paper;
    }
    
    /**
     * Called with the base blaise properties file.
     * The contract is that this method will be called first.
     * Note: The paper.props file, will also be passed via 
     * handleEntry.
     * @see BlaiseFile
     */
    public void setPaperProperties(Properties props){
        this.props=props;
        if(debug){
            System.out.println("Paper Props:");
            System.out.println(props);
        }
        drafts=new Vector();
        int numDrafts=Integer.parseInt(props.getProperty("2.numDrafts"));
        int i;
        int j,numComments;
        Blaise_Draft draft;
        Blaise_Comment comment;
        Vector comments;
        String tempD1,tempD2,tempC1,tempC2;//optimize getting draft/comment stuff, a little.
        String paperID=paper.getPaperID(),draftID;
        for(i=0;i<numDrafts;i++){
            draft=new Blaise_Draft();
            
            numComments=setDraftProperties(draft,props,i,paperID);
            draftID=draft.getDraftID();
            //comments, add to drafts.
            if(numComments>0){
                comments=new Vector();
                for(j=0;j<numComments;j++){
                    comment=new Blaise_Comment();
                    setCommentProperties(comment,props,i,j,paperID,draftID);
                    comments.add(comment);
                }
                draft.setComments(comments);
            }
            drafts.add(draft);
        }
        paper.setDrafts(drafts);
        
        //Check any additional WorkingDraft properties.
        String s;
        if((s=props.getProperty("2.workingDraft.extBodyFilename"))!=null){
            paper.getWorkingDraft().setExternalBodyFilename(s);
        }
    }
    
    /**
     * Sets the Blaise_Draft from the Properties object.
     * @param dirNum is the directory number this draft is (draft_number-1)
     * @returns the number of comments this draft has (according to the properties)
     */
    protected static int setDraftProperties(Blaise_Draft draft, Properties props,int dirNum,String paperID){
        String tempD1="1.draft"+dirNum;
        String tempD2="2.draft"+dirNum;
        //set the various properties for this draft.
        draft.setDraftID(props.getProperty(tempD1+".draftID"));
        draft.setPaperID(paperID);
        draft.setDraftNumber(props.getProperty(tempD1+".draft_number"));
        draft.setTitle(props.getProperty(tempD1+".title"));
        draft.setTurninDate(props.getProperty(tempD2+".turninDate"));
        draft.setIsModified(props.getProperty(tempD1+".isModified"));

        //only exists if the draft is external.
        String s;
        if((s=props.getProperty(tempD2+".extBodyFilename"))!=null){
            draft.setExternalBodyFilename(s);
            if(debug)System.out.println("draft "+dirNum+" external body file name found: "+s);
        }
        try{
            return Integer.parseInt(props.getProperty(tempD2+".numComments"));
        }catch(NumberFormatException e){
            return 0;//assume that it is missing...
        }
    }
    
    /**
     * Sets the Blaise_Comment from the Properties object.
     * @param draftDirNum the directory number of the draft that is the parent of the comment.
     */
    protected static void setCommentProperties(Blaise_Comment comment, Properties props, int draftDirNum, int commentDirNum, String paperID, String draftID){
        String tempC1="1.draft"+draftDirNum+".comment"+commentDirNum;
        String tempC2="2.draft"+draftDirNum+".comment"+commentDirNum;

        comment.setCommentID(props.getProperty(tempC1+".commentID"));
        comment.setPaperID(paperID);
        comment.setGraderID(props.getProperty(tempC1+".graderID"));
        comment.setDraftID(draftID);
        comment.setIsPublic(props.getProperty(tempC1+".isPublic"));
        comment.setGraderName(props.getProperty(tempC1+".grader_name"));
        comment.setAspectX(props.getProperty(tempC1+".aspect_x"));
        comment.setAspectY(props.getProperty(tempC1+".aspect_y"));
        String s;
        if((s=props.getProperty(tempC2+".extBodyFilename"))!=null){
            comment.setExternalBodyFilename(s);
            if(debug)System.out.println("comment "+commentDirNum+" of draft"+draftDirNum+" external body file name found: "+s);
        }
    }
    
    /**
     * Called once for each entry in a blaise file.  This
     * gives the handler the chance to output the file to a real
     * location (or take some other action).
     */
    public void handleEntry(JarEntry entry) throws IOException{
        if(debug){
            System.out.println("handleEntry called with:"+entry);
        }
    
        if(entry.isDirectory()){
            //only really care about for subclasses that need to create a directory here.
            handleDirectoryEntry(entry);
            return;
        }
        else{
            String name=entry.getName();
            int dn=-1;
            int cn=-1;
            
            //figure out whether it belongs to a draft, and if so, what number.
            dn=getDraftNumber(name);
            if(dn!=-1){
                //figure out if it also belongs to a comment.
                cn=getCommentNumber(name);
                handleDraftEntry(entry, dn, cn);
            }
            else if(name.startsWith("WorkingDraft/")){//working draft
                handleWorkingDraftEntry(entry);
            }
            else{//paper.props maybe...
                handleMiscEntry(entry);
            }
        }
    }
    
    /**
     * Called when a directory is detected in handleEntry.
     * This method is empty, but can be overridden by subclasses to
     * do something useful.
     */
    protected void handleDirectoryEntry(JarEntry entry)  throws IOException{
    }
    
    
    /**
     * This method allows you to specify the draft and comment that the entry should be put in.
     * Basically, this method was added as a retro-fit so the BlaiseFile could open and save to
     * just Blaise_Drafts and Comments.
     * note: draftNum and commentNum are the draft and comment directory numbers. (starting from 0, not 1).
     */
    protected ExternalFile handleDraftEntry(Blaise_Draft draft, Blaise_Comment comment, JarEntry entry, int draftNum, int commentNum) throws IOException{
        String name=entry.getName();
        InputStream stream=getInputStream(entry);
        if(commentNum==-1){//draft only stuff
            if(name.equals("Draft"+draftNum+"/draft.html")){
                draft.setBody(DocumentUtilities.getStreamContentsAsString(stream));
                stream.close();
                return null;
            }
            
            if(debug){
                System.out.println("Draft ("+draftNum+")= "+draft);
            }
            
            //check if this draft has an external body filename, and if this file is it.
            String extBfn=draft.getExternalBodyFilename();
            //if(debug)System.out.println("'"+extBfn+"'");
            if(extBfn!=null && name.endsWith("Files/"+extBfn)){
                //if(debug)System.out.println("Found external body file");
                if(makeArrays){
                    draft.setExternalBody(DocumentUtilities.getInputStreamAsByteArray(stream));
                    stream.close();
                }
                return null;
            }
            
            //a random File, nothing special about it.  Add it to the external Files list.
            
            //it might be a prudent idea to try and load it as something more than a byte array.
            //but for the moment, that's what we'll do.
            
            //cut out the directory data from the filename.
            int i=name.lastIndexOf('/');
            ExternalFile ef=new ExternalFile(name.substring(i+1,name.length()),(makeArrays)?(DocumentUtilities.getInputStreamAsByteArray(stream)):(null),null,null);
            draft.addExternalFile(ef);
            if(debug){
                System.out.println("added");
            }
            stream.close();
            return ef;
        }
        else{//comment related file
            String cNameStart="Draft"+draftNum+"/Comment"+commentNum+"/";
            if(name.equals(cNameStart+"comment.html")){
                comment.setBody(DocumentUtilities.getStreamContentsAsString(stream));
                stream.close();
                return null;
            }
            else if(name.equals(cNameStart+"scribbles.props")){
                comment.setScribbles(makeScribbles(entry));
                stream.close();
                return null;
            }
            else if(name.equals(cNameStart+"notes.props")){
                comment.setNotes(makeNotes(entry));
                stream.close();
                return null;
            }
            
            //check if this comment has an external body filename, and if this file is it.
            if(makeArrays){
                String extBfn=comment.getExternalBodyFilename();
                if(extBfn!=null && name.endsWith("Files/"+extBfn)){
                    comment.setExternalBody(DocumentUtilities.getInputStreamAsByteArray(stream));
                    stream.close();
                    return null;
                }
            }
            
            //a random File, nothing special about it.  Add it to the external Files list.
            
            //it might be a prudent idea to try and load it as something more than a byte array.
            //but for the moment, that's what we'll do.
            
            //cut out the directory data from the filename.
            int i=name.lastIndexOf('/');
            ExternalFile ef=new ExternalFile(name.substring(i+1,name.length()),(makeArrays)?(DocumentUtilities.getInputStreamAsByteArray(stream)):(null),null,null);
            comment.addExternalFile(ef);
            stream.close();
            return ef;
        }
    }
    
    /**
     * Called when the file is part of a Draft.  (not WorkingDraft).
     * If draft or comment number do not make sense, they should be set to -1.
     *
     * @returns the external file entry the it created, or null if one wasn't created.
     */
    protected ExternalFile handleDraftEntry(JarEntry entry, int draftNum, int commentNum) throws IOException{
        Blaise_Draft draft=paper.getDraft(draftNum);
        Blaise_Comment comment=null;
        if(commentNum>=0){
            comment=(Blaise_Comment)draft.getComments().elementAt(commentNum);
        }
        return handleDraftEntry(draft,comment,entry,draftNum,commentNum);
    }
    
    /**
     * Called when a jar entry that does not fall into Draft, Comment, or Directory entries
     * is found.  This could be, for example, paper.props, or the manifest, or it could be
     * a file added in a future version of blaise.
     * This method is empty, but can be overridden by subclasses to
     * do something useful.
     */
    protected void handleMiscEntry(JarEntry entry)  throws IOException{
    }
    
    /**
     * Called when a jar entry that is part of the working draft is detected.
     * Note that the actual "WorkingDraft" directory is passed to handleDirectoryEntry
     * not handleWorkingDraftEntry.
     *
     * @returns the external file entry the it created, or null if one wasn't created.
     */
    protected ExternalFile handleWorkingDraftEntry(JarEntry entry)  throws IOException{
        String name=entry.getName();
        Blaise_Draft draft=paper.getWorkingDraft();
        InputStream stream=getInputStream(entry);
        if(name.equals("WorkingDraft/draft.html")){
            draft.setBody(DocumentUtilities.getStreamContentsAsString(stream));
            stream.close();
            return null;
        }
        //check if this draft has an external body filename, and if this file is it.

        String extBfn=draft.getExternalBodyFilename();
        if(extBfn!=null && name.endsWith("Files/"+extBfn)){
            if(makeArrays){
                draft.setExternalBody(DocumentUtilities.getInputStreamAsByteArray(stream));
                stream.close();
            }
            return null;
        }
        
        //a random File, nothing special about it.  Add it to the external Files list.
    
        //it might be a prudent idea to try and load it as something more than a byte array.
        //but for the moment, that's what we'll do.
    
        //cut out the directory data from the filename.
        int i=name.lastIndexOf('/');
        ExternalFile ef=new ExternalFile(name.substring(i+1,name.length()),(makeArrays)?(DocumentUtilities.getInputStreamAsByteArray(stream)):(null),null,null);
        draft.addExternalFile(ef);
        stream.close();
        return ef;
    }
    
    /**
     * Given a JarEntry that points to a scribbles file, this method
     * puts together a Vector of Scribbles objects.
     */
    protected Vector makeScribbles(JarEntry entry) throws IOException{
        InputStream stream=jf.getInputStream(entry);
        Properties sprops=new Properties();
        sprops.load(stream);
        stream.close();
        
        Vector scribs=new Vector();
        Vector lines;
        Color color;
		String numScribs=sprops.getProperty("numScribs");
		if(numScribs==null){
			return scribs;
		}
        int size=Integer.parseInt(numScribs);
        int width,offset;
        double height,xDist;
        double prev_x,prev_y,new_x,new_y;
        StringTokenizer tok1,tok2;
        String base;
        for(int i=0;i<size;i++){
            base=new StringBuffer().append(i).append('.').toString();
            //get the basic properties
            color = new Color(Integer.parseInt(sprops.getProperty(base+"red")),
                              Integer.parseInt(sprops.getProperty(base+"green")),
                              Integer.parseInt(sprops.getProperty(base+"blue")));
            width=Integer.parseInt(sprops.getProperty(base+"width"));
            offset=Integer.parseInt(sprops.getProperty(base+"offset"));
            height=Double.parseDouble(sprops.getProperty(base+"height"));
            xDist=Double.parseDouble(sprops.getProperty(base+"xDist"));
            //get and parse the line property
            tok1 = new StringTokenizer(sprops.getProperty(base+"lines"),"#");
            prev_x = -1;
            prev_y = -1;
            lines=new Vector();
            while(tok1.hasMoreTokens()){
                tok2 = new StringTokenizer((String)tok1.nextElement(), "*");
                if(prev_x == -1){
                    prev_x = Double.parseDouble((String)tok2.nextElement());
                    prev_y = Double.parseDouble((String)tok2.nextElement());
                }
                else{
                    new_x = Double.parseDouble((String)tok2.nextElement());
                    new_y = Double.parseDouble((String)tok2.nextElement());
                    lines.add(new ScribbleLine(prev_x, prev_y, new_x, new_y));
                    prev_x = new_x;
                    prev_y = new_y;
                }
            }
            scribs.addElement(new Scribble(width, color, offset, height, xDist, lines));
        }
        return scribs;
    }
    
    /**
     * Given a JarEntry that points to a notes file, this method
     * puts together a Vector of Notes objects.
     */
    protected Vector makeNotes(JarEntry entry) throws IOException{
        InputStream stream=jf.getInputStream(entry);
        Properties nprops=new Properties();
        nprops.load(stream);
        stream.close();
        
        Vector notes=new Vector();
        Color color;
        Point point;
        Rectangle rect;
        String base,nID,title,text;
        boolean isOpen;
		String numNotes=nprops.getProperty("numNotes");
		if(numNotes==null){
			return notes;//empty
		}
        int size=Integer.parseInt(numNotes);
        for(int i=0;i<size;i++){
            base=new StringBuffer().append(i).append('.').toString();
            color = new Color(Integer.parseInt(nprops.getProperty(base+"red")),
                              Integer.parseInt(nprops.getProperty(base+"green")),
                              Integer.parseInt(nprops.getProperty(base+"blue")));
            rect= new Rectangle(Integer.parseInt(nprops.getProperty(base+"fPx")),
                                Integer.parseInt(nprops.getProperty(base+"fPy")),
                                Integer.parseInt(nprops.getProperty(base+"fPw")),
                                Integer.parseInt(nprops.getProperty(base+"fPh")));
            point=new Point(Integer.parseInt(nprops.getProperty(base+"iPx")),
                            Integer.parseInt(nprops.getProperty(base+"iPy")));
            nID=nprops.getProperty(base+"nID");
            title=nprops.getProperty(base+"title");
            text=nprops.getProperty(base+"text");
            isOpen=Boolean.getBoolean(nprops.getProperty(base+"isOpen"));
            notes.addElement(new Note(nID,point,color,title,rect,text,isOpen));
        }
        return notes;
    }
    
    /**
     * Attacks name with a regular expression to discover the draft number
     * that is contained within the name.
     * Draft numbers start at 0.  Returns -1 if no draft numbers are found.
     * @see D_DIR
     */
    protected int getDraftNumber(String name){
        int dn=-1;
        Matcher m=D_DIR.matcher(name);
        if(m.find()){
            try{
                dn=Integer.parseInt(name.substring(D_DIR_BEFORE_LENGTH,m.end()-D_DIR_AFTER_LENGTH));
            }catch(NumberFormatException e){
                System.out.println("parse error in name:"+name+" in getDraftNumber");
            }
        }
        return dn;
    }
    
    /**
     * Attacks name with a regular expression to discover the comment number
     * that is contained within the name.
     * Comment numbers start at 0.  Returns -1 if no comment numbers are found.
     * @see C_DIR
     */
    protected int getCommentNumber(String name){
        int cn=-1;
        Matcher m=C_DIR.matcher(name);
        if(m.find()){
            try{
                cn=Integer.parseInt(name.substring(C_DIR_BEFORE_LENGTH+m.start(),m.end()-C_DIR_AFTER_LENGTH));
            }catch(NumberFormatException e){
                System.out.println("parse error in name:"+name+" in getCommentNumber");
            }
        }
        return cn;
    }
    
    /**
     * A Regular Expression that matches the Directory structure for a draft dir.
     * "^Draft[0-9]+/"
     */
    protected static final Pattern D_DIR=Pattern.compile("^(Draft)[0-9]+/");
    /**Length of the string before the numbers start.*/
    protected static final int D_DIR_BEFORE_LENGTH=5;
    protected static final int D_DIR_AFTER_LENGTH=1;
    /**
     * A Regular Expression that matches the dir struct for a comment dir.
     * "/Comment[0-9]+/"
     */
    protected static final Pattern C_DIR=Pattern.compile("/Comment[0-9]+/");
    protected static final int C_DIR_BEFORE_LENGTH=8;
    protected static final int C_DIR_AFTER_LENGTH=1;
}
