package edu.gwu.blaise.version2.documents;
import java.io.*;
import java.util.jar.*;
import java.util.zip.*;
import java.util.*;
import java.awt.Color;
import java.awt.Point;
import java.awt.Rectangle;
/**
 * Handles the jarring and unjarring of Blaise Papers
 * (those with .bls extensions).
 *
 * @author Michael Gaiman
 * @version 1 4/8/2004
 */ 
public class BlaiseFile{
    private static final String PAPERPROPS="paper.props";
    private static final String PAPERID_PROP="1.paperID";//used multiple places in this file...
    
    private String note="";
    private int progress=0;

    private Blaise_Paper paper;
    private File jarFile;
    private BasicFileHandler handler;
    
    /**
     * true when this instance of BlaiseFile is used by the static methods.
     * They're allowed to do things that aren't usually allowed...
     */
    private boolean privateInstance=false;
    
    /**
     * Creates a Blaise File instance with the data
     * going to/from the supplied Blaise Paper and File
     * objects.
     */
    public BlaiseFile(Blaise_Paper bp, File jarFile,BasicFileHandler handler){
        newJob(bp,jarFile,handler);
    }
    
    /**
     * Creates a default instance of BlaiseFile.  In order to use this without crashing
     * things, newJob needs to be called.
     */
    public BlaiseFile(){
    }
    
    /**
     * Creates an instance of BlaiseFile only for saving papers to alternative
     * output streams.
     */
    private BlaiseFile(Blaise_Paper bp){
        privateInstance=true;
        newJob(bp,null,null);
    }
    
    /**
     * Gets the Blaise_Paper instance that this BlaiseFile
     * is currently using.
     */
    public Blaise_Paper getPaper(){
        return paper;
    }
    
    /**
     * Accessor method to get the File currently
     * being operated upon.
     */
    public File getJarFile(){
        return jarFile;
    }
    
    /**
     * Accessor method to get the filename of the File
     * currently being operated upon.
     */
    public String getJarFilename(){
        if(jarFile==null){
            return null;
        }
        return jarFile.getAbsolutePath();
    }
    
    /**
     * Courtesy method to use this BlaiseFile instance for a new
     * Blaise_Paper, File combination.
     */
    public void newJob(Blaise_Paper bp, File jarFile, BasicFileHandler handler){
        //privateInstances can have null args...
        if(!privateInstance && (bp==null||jarFile==null||handler==null)){
            throw new IllegalArgumentException("BlaiseFile args cannot be null.");
        }
        
        paper=bp;
        this.jarFile=jarFile;
        this.handler=handler;
        progress=0;
        note="";
    }
    
    /**
     * Saves the paper to the specified outputstream.  Usually it is a file output stream.
     */
    private void savePaper(OutputStream outputStream) throws IOException{
        note="Preparing to save Blaise document...";
        progress=1;
        
        //ready the outputstream.
        JarOutputStream ostream=new JarOutputStream(new BufferedOutputStream(outputStream));
        progress=50;
        
        note="Setting document properties";
        Properties props=makePaperPropsFromBP();//get the general paper-wide properties.
        progress=100;
        
        note="Creating document structure";
        createDirectories(ostream);//go through and setup all of the subdirectories first.
        progress=110;
        
        JarEntry entry;
        Vector drafts,comments;
        Blaise_Draft draft;
        Blaise_Comment comment;
        int i,j,dsize,csize;
        drafts=paper.getDrafts();
        dsize=drafts.size();
        
        int beginP=progress;//the progress when the for loop starts.
        int completed=0;
        int endP=(DONE_PROGRESS-200)-beginP;
        for(i=0;i<dsize;i++){
            note="Adding Draft "+(i+1);
            draft=(Blaise_Draft)drafts.elementAt(i);
            appendDraftProps(props,draft,false);
            addDraft(draft,ostream,false);//add the draft (though not the comment) to the bls
            completed++;
            progress=beginP + (int)((float)endP*((float)completed/(float)dsize));
            note="Adding Comments to draft";
            csize=draft.getCommentSize();
            comments=draft.getComments();
            for(j=0;j<csize;j++){
                comment=(Blaise_Comment)comments.elementAt(j);
                appendCommentProps(props,comment,i,j);
                addComment(comment,ostream,i,j);
            }
        }
        note="Adding Working Draft";
        draft=paper.getWorkingDraft();
        appendDraftProps(props,draft,true);
        addDraft(draft,ostream,true);
        progress+=100;
        
        //add the properties...
        note="Adding properties";
        entry=new JarEntry("paper.props");
        ostream.putNextEntry(entry);
        props.store(ostream,null);
        ostream.finish();
        ostream.flush();
        progress=DONE_PROGRESS;
        note="Complete";
    }
    
    /**
     * Causes the current Blaise_Paper to be saved to the current jarFile.
     */
    public void savePaper() throws IOException{
        FileOutputStream f=new FileOutputStream(jarFile);
        savePaper(f);
        f.close();
    }
    
    /**
     * Causes the current jarFile to be unpacked (into tempDir) and
     * read into the current Blaise_Paper.
     */
    public void openPaper() throws IOException{
        //open a JarFile from the File
        note="Preparing to open Blaise document...";
        progress=1;
        
        JarFile jf=new JarFile(jarFile);
        JarEntry entry=jf.getJarEntry(PAPERPROPS);
        Properties props=new Properties();
        
        //System.out.println("entry="+entry);
        
        InputStream stream=jf.getInputStream(entry);
        progress=10;
        props.load(stream);
        progress=50;
        if(paper==null){
            paper=new Blaise_Paper();
        }
        paper.setVersion(props.getProperty("1.blaise_version"));
        paper.setPaperID(props.getProperty("1.paperID"));
        paper.setWriterID(props.getProperty("1.writerID"));
        paper.setUniversityID(props.getProperty("1.universityID"));
        paper.setCourseID(props.getProperty("1.courseID"));
        paper.setFirst(props.getProperty("1.first"));
        paper.setLast(props.getProperty("1.last"));
        paper.setStudentIdentifier(props.getProperty("1.student_identifier"));
        paper.setEmail(props.getProperty("1.email"));
        paper.setDept(props.getProperty("1.dept"));
        paper.setNumber(props.getProperty("1.number"));//course number
        paper.setSection(props.getProperty("1.section"));
        paper.setSemester(props.getProperty("2.semester"));
        paper.setYear(props.getProperty("2.year"));
        paper.setAssignmentID(props.getProperty("2.assignID"));
        paper.setAssignmentNumber(props.getProperty("2.assignNum"));
        paper.setAssignmentName(props.getProperty("2.assignName"));
        //load the rest of paper.props into the various
        //blaise drafts and comments.
        progress=100;
        int size=jf.size();
        int completed=0;
        
        note="Preparing directory structure...";
        handler.setJarFile(jf);
        handler.setPaper(paper);
        handler.setPaperProperties(props);
        progress=150;
        
        int beginP=progress;//the progress when the for loop starts.
        int endP=DONE_PROGRESS-beginP;
        for(Enumeration e=jf.entries();e.hasMoreElements();){
            entry=(JarEntry)e.nextElement();
            handler.handleEntry(entry);
            completed++;
            progress=beginP + (int)((float)endP*((float)completed/(float)size));
            note="Handling "+entry.getName();
        }
        
        note="Complete";
        progress=DONE_PROGRESS;
    }
    
    /**
     * While jarring and unjarring are occurring this method may be
     * polled to get a String describing the current progress in a way that
     * is appropriate for user display (in a progress monitor, for example).
     * Please Note: the term "Note" in this context has nothing to do with the
     * Note grading tool.
     */
    public String getNote(){
        return note;
    }
    
    /**
     * While jarring and unjarring are occurring this method may be
     * polled to get the current status of the operation.  The number returned
     * is from 0 to DONE_PROGRESS where 0 is not started, and DONE_PROGRESS is completed.
     */ 
    public int getProgress(){
        return progress;
    }
    
    public static final int DONE_PROGRESS=1000;

/*------------------------------------------------------------------------------
 * Static To and From methods
 *----------------------------------------------------------------------------*/
    
    /**
     * A utility method that given a Blaise_Paper will return a jarred
     * representation of it as a byte array.
     * This is done without the use of an intermediate File.
     */
    public static byte[] paperToByteArray(Blaise_Paper p) throws IOException{
        BlaiseFile bf=new BlaiseFile(p);
        ByteArrayOutputStream baos=new ByteArrayOutputStream(1024);
        bf.savePaper(baos);
        
        return baos.toByteArray();
    }
    
    /**
     * A utility method that writes the specified blaise paper directly to the specified OutputStream.
     * This is done without the use of an intermediate File.
     * Note: this method does not close the stream afterwards.
     */
    public static void paperToOutputStream(Blaise_Paper p, OutputStream o) throws IOException{
        BlaiseFile bf=new BlaiseFile(p);
        bf.savePaper(o);
    }
    
    /**
     * Given a byte array containing a Bls file, the corresponding Blaise_Paper will
     * be returned.
     * The opening is done using the BackendFileHandler.
     * This is just a convenience method, it will output the byte array to a
     * File first and open it that way.
     */
    public static Blaise_Paper byteArrayToPaper(byte[] ba) throws IOException{
        File temp=File.createTempFile("blaise",null,null);
        DocumentUtilities.bytesToFile(ba,temp);
        Blaise_Paper p=fileToPaper(temp);
        temp.delete();
        return p;
    }
    
    /**
     * Given an InputStream containing a Bls file, the corresponding Blaise_Paper will
     * be returned.
     * The opening is done using the BackendFileHandler.
     * This is just a convenience method, it will output the stream contents to a
     * File first and open it that way.
     */
    public static Blaise_Paper inputStreamToPaper(InputStream istream) throws IOException{
        File temp=File.createTempFile("blaise",null,null);
        DocumentUtilities.inputStreamToFile(istream, temp);
        Blaise_Paper p=fileToPaper(temp);
        temp.delete();
        return p;
    }
    
    /**
     * A utility method that given a Blaise_Draft will return a serialized
     * representation of it as a byte array.  Note: the Blaise_Comment objects
     * that this Blaise_Draft has (if any) are not included.
     * This is done without the use of an intermediate File.
     */
    public static byte[] draftToByteArray(Blaise_Draft d) throws IOException{
        ByteArrayOutputStream baos=new ByteArrayOutputStream(1024);
        draftToOutputStream(d,baos);
        return baos.toByteArray();
    }
    
    /**
     * A utility method that writes the specified blaise draft directly to the specified OutputStream.
     * This is done without the use of an intermediate File.
     * Note: this method does not close the stream afterwards, and note, also, that
     * internal Blaise_Comment objects are not included.
     */
    public static void draftToOutputStream(Blaise_Draft d, OutputStream o) throws IOException{
        Properties props=new Properties();
        appendDraftProps(props,d,false);//assume this is not the working draft.
        //add on a property that will make unpacking the draft easier.
        props.setProperty("2.draftDirNum",Integer.toString(Integer.parseInt(d.getDraftNumber())-1));
        props.setProperty(PAPERID_PROP,d.getPaperID());
        
        JarOutputStream ostream=new JarOutputStream(new BufferedOutputStream(o));
        addDraft(d,ostream,false);
        ostream.putNextEntry(new JarEntry(PAPERPROPS));
        props.store(ostream,null);
        ostream.finish();
        ostream.flush();
    }
    
    /**
     * Given a byte array containing a serialized Blaise_Draft 
     * (serialized via the draftTo methods in BlaiseFile), 
     * the corresponding Blaise_Draft will be returned.
     * The opening is done using the BackendFileHandler, 
     * so all external files will be restored using as byte arrays.
     */
    public static Blaise_Draft byteArrayToDraft(byte[] ba) throws IOException{
        File temp=File.createTempFile("blaise",null,null);
        DocumentUtilities.bytesToFile(ba,temp);
        Blaise_Draft d=fileToDraft(temp);
        temp.delete();
        return d;
    }
    
    /**
     * Given an InputStream containing a serialized Blaise_Draft 
     * (serialized via the draftTo methods in BlaiseFile), 
     * the corresponding Blaise_Draft will be returned.
     * The opening is done using the BackendFileHandler, 
     * so all external files will be restored using as byte arrays.
     */
    public static Blaise_Draft inputStreamToDraft(InputStream istream) throws IOException{
        File temp=File.createTempFile("blaise",null,null);
        DocumentUtilities.inputStreamToFile(istream, temp);
        Blaise_Draft d=fileToDraft(temp);
        temp.delete();
        return d;
    }
    
    /**
     * A utility method that given a Blaise_Comment will return a serialized
     * representation of it as a byte array.
     * This is done without the use of an intermediate File.
     */
    public static byte[] commentToByteArray(Blaise_Comment c) throws IOException{
        ByteArrayOutputStream baos=new ByteArrayOutputStream(1024);
        commentToOutputStream(c,baos);
        return baos.toByteArray();
    }
    
    /**used in serializing comments*/
    private static final int COMMENT_DDIR_NUM=0;
    /**used in serializing comments*/
    private static final int COMMENT_CDIR_NUM=0;
    
    /**
     * A utility method that writes the Blaise_Comment directly to the specified OutputStream.
     * This is done without the use of an intermediate File.
     */
    public static void commentToOutputStream(Blaise_Comment c, OutputStream o) throws IOException{
        Properties props=new Properties();
        appendCommentProps(props,c,COMMENT_DDIR_NUM,COMMENT_CDIR_NUM);
        //add on a few properties that will make unpacking the comment easier.
        props.setProperty("2.draftDirNum",Integer.toString(COMMENT_DDIR_NUM));
        props.setProperty("2.commentDirNum",Integer.toString(COMMENT_CDIR_NUM));
        props.setProperty(PAPERID_PROP,c.getPaperID());
        props.setProperty("1.draft"+COMMENT_DDIR_NUM+".draftID",c.getDraftID());
        
        JarOutputStream ostream=new JarOutputStream(new BufferedOutputStream(o));
        addComment(c,ostream,COMMENT_DDIR_NUM,COMMENT_CDIR_NUM);
        ostream.putNextEntry(new JarEntry(PAPERPROPS));
        props.store(ostream,null);
        ostream.finish();
        ostream.flush();
    }
    
    /**
     * Given a byte array containing a serialized Blaise_Comment 
     * (serialized via the commentTo methods in BlaiseFile), 
     * the corresponding Blaise_Comment will be returned.
     * The opening is done using the BackendFileHandler, 
     * so all external files will be restored using as byte arrays.
     */
    public static Blaise_Comment byteArrayToComment(byte[] ba) throws IOException{
        File temp=File.createTempFile("blaise",null,null);
        DocumentUtilities.bytesToFile(ba,temp);
        Blaise_Comment c=fileToComment(temp);
        temp.delete();
        return c;
    }
    
    /**
     * Given an InputStream containing a serialized Blaise_Comment 
     * (serialized via the commentTo methods in BlaiseFile), 
     * the corresponding Blaise_Comment will be returned.
     * The opening is done using the BackendFileHandler, 
     * so all external files will be restored using as byte arrays.
     */
    public static Blaise_Comment inputStreamToComment(InputStream istream) throws IOException{
        File temp=File.createTempFile("blaise",null,null);
        DocumentUtilities.inputStreamToFile(istream, temp);
        Blaise_Comment c=fileToComment(temp);
        temp.delete();
        return c;
    }
    
    /**
     * Used privately by the static ToPaper methods to do the real work.
     */
    private static Blaise_Paper fileToPaper(File f) throws IOException{
        Blaise_Paper p=new Blaise_Paper();
        BlaiseFile bf=new BlaiseFile(p,f,new BackendFileHandler());
        bf.openPaper();
        return p;
    }
    
    /**
     * Used privately by the static ToDraft methods to do the real work.
     */
    private static Blaise_Draft fileToDraft(File f) throws IOException{
        Blaise_Draft draft=new Blaise_Draft();
        
        JarFile jf=new JarFile(f);
        JarEntry entry=jf.getJarEntry(PAPERPROPS);
        Properties props=new Properties();
                
        InputStream stream=jf.getInputStream(entry);
        props.load(stream);
        
        int dirNum=Integer.parseInt(props.getProperty("2.draftDirNum"));
        String paperID=props.getProperty(PAPERID_PROP);
        
        BackendFileHandler handler=new BackendFileHandler();
        handler.setJarFile(jf);
        
        handler.setDraftProperties(draft,props,dirNum,paperID);
        for(Enumeration e=jf.entries();e.hasMoreElements();){
            entry=(JarEntry)e.nextElement();
            if(entry.getName().equals(PAPERPROPS)){
                continue;//skip paper props.
            }
            handler.handleDraftEntry(draft,null,entry,dirNum,-1);
        }
        
        return draft;
    }
    
    /**
     * Used privately by the static ToComment methods to do the real work.
     */
    private static Blaise_Comment fileToComment(File f) throws IOException{
        Blaise_Comment comment=new Blaise_Comment();
        
        JarFile jf=new JarFile(f);
        JarEntry entry=jf.getJarEntry(PAPERPROPS);
        Properties props=new Properties();
                
        InputStream stream=jf.getInputStream(entry);
        props.load(stream);
        
        int dDirNum=Integer.parseInt(props.getProperty("2.draftDirNum"));
        int cDirNum=Integer.parseInt(props.getProperty("2.commentDirNum"));
        String paperID=props.getProperty(PAPERID_PROP);
        String draftID=props.getProperty("1.draft"+dDirNum+".draftID");
        
        BackendFileHandler handler=new BackendFileHandler();
        handler.setJarFile(jf);
        handler.setCommentProperties(comment,props,dDirNum,cDirNum,paperID,draftID);
        
        for(Enumeration e=jf.entries();e.hasMoreElements();){
            entry=(JarEntry)e.nextElement();
            if(entry.getName().equals(PAPERPROPS)){
                continue;//skip paper props.
            }
            handler.handleDraftEntry(null,comment,entry,dDirNum,cDirNum);
        }
        
        return comment;
    }
    
    
    ///////////////////////////////////////////////////////////////////////////
    // Saving Related Methods
    ///////////////////////////////////////////////////////////////////////////
    
    /**
     * Returns a Properties object that contains the various properties related
     * to the Blaise_Paper.  Note: the properties file does not contain properties
     * of drafts, comments, or the working draft.
     */
    private Properties makePaperPropsFromBP(){
        Properties props=new Properties();
        props.setProperty("1.blaise_version",paper.getVersion());
        props.setProperty(PAPERID_PROP,paper.getPaperID());
        props.setProperty("1.writerID",paper.getWriterID());
        props.setProperty("1.universityID",paper.getUniversityID());
        props.setProperty("1.courseID",paper.getCourseID());
        props.setProperty("1.first",paper.getFirst());
        props.setProperty("1.last",paper.getLast());
        props.setProperty("1.student_identifier",paper.getStudentIdentifier());
        props.setProperty("1.email",paper.getEmail());
        props.setProperty("1.dept",paper.getDept());
        props.setProperty("1.number",paper.getNumber());
        props.setProperty("1.section",paper.getSection());
        props.setProperty("2.semester",paper.getSemester());
        props.setProperty("2.year",paper.getYear());
        props.setProperty("2.assignID",paper.getAssignmentID());
        props.setProperty("2.assignNum",paper.getAssignmentNumber());
        props.setProperty("2.assignName",paper.getAssignmentName());
        props.setProperty("2.numDrafts",new Integer(paper.getDrafts().size()).toString());
        
        return props;
    }
    
    /**
     * Makes all of the standard directories.
     */
    private void createDirectories(JarOutputStream stream) throws IOException{
        JarEntry entry;
        int i,j;
        Vector drafts=paper.getDrafts();
        int dsize=drafts.size();
        int csize;
        String currentDir,currentCDir;
        for(i=0;i<dsize;i++){
            currentDir="Draft"+i+'/';
            entry=new JarEntry(currentDir);
            stream.putNextEntry(entry);
            entry=new JarEntry(currentDir+"Files/");
            stream.putNextEntry(entry);
            csize=((Blaise_Draft)drafts.elementAt(i)).getCommentSize();
            for(j=0;j<csize;j++){
                currentCDir=currentDir+"Comment"+j+'/';
                entry=new JarEntry(currentCDir);
                stream.putNextEntry(entry);
                entry=new JarEntry(currentCDir+"Files/");
                stream.putNextEntry(entry);
            }
        }
        entry=new JarEntry("WorkingDraft/");
        stream.putNextEntry(entry);
        entry=new JarEntry("WorkingDraft/Files/");
        stream.putNextEntry(entry);
    }
    
    /**
     * Adds the draft specific properties to the properties file
     * @param workingD true if this is the working draft, false otherwise.
     */
    private static void appendDraftProps(Properties props,Blaise_Draft draft,boolean workingD){
        if(workingD){
            if(draft.hasExternalBody()){
                props.setProperty("2.workingDraft.extBodyFilename",draft.getExternalBodyFilename());
            }
        }
        else{
            int draftNum=Integer.parseInt(draft.getDraftNumber())-1;
            String draft1="1.draft"+draftNum+'.';
            String draft2="2.draft"+draftNum+'.';
            props.setProperty(draft1+"draftID",draft.getDraftID());
            props.setProperty(draft1+"draft_number",draft.getDraftNumber());
            props.setProperty(draft1+"title",draft.getTitle());
            props.setProperty(draft2+"turninDate",draft.getTurninDate());
            props.setProperty(draft1+"isModified",draft.getIsModified());
            props.setProperty(draft2+"numComments",new Integer(draft.getCommentSize()).toString());
            if(draft.hasExternalBody()){
                props.setProperty(draft2+"extBodyFilename",draft.getExternalBodyFilename());
            }
        }
    }
    
    /**
     * Puts the various files of the draft in the bls file.
     * @param workingD true if this is the working draft, false otherwise.
     */ 
    private static void addDraft(Blaise_Draft draft, JarOutputStream stream, boolean workingD) throws IOException{
        JarEntry entry;
        String base;
        File file;
        byte[] bArray;
        if(workingD){
            base="WorkingDraft/";
        }
        else{
            int num=Integer.parseInt(draft.getDraftNumber())-1;
            base="Draft"+num+'/';
        }
        addEFContainer(draft,stream,base+"Files/");
        
        if(!draft.hasExternalBody()){
            //add the draft.html file.
            entry=new JarEntry(base+"draft.html");
            stream.putNextEntry(entry);
            Writer w=new BufferedWriter(new OutputStreamWriter(stream));
            w.write(draft.getBody());
            w.flush();
            stream.closeEntry();
        }
    }
    
    /**
     * Adds the comment specific properties to the props file.
     * @param draftNum The draft this comment belongs to.  Starts at 0.
     */
    private static void appendCommentProps(Properties props,Blaise_Comment comment,int draftNum,int commentNum){
        String comment1="1.draft"+draftNum+".comment"+commentNum+'.';
        String comment2="2.draft"+draftNum+".comment"+commentNum+'.';
        props.setProperty(comment1+"commentID",comment.getCommentID());
        props.setProperty(comment1+"graderID",comment.getGraderID());
        props.setProperty(comment1+"isPublic",comment.getIsPublic());
        props.setProperty(comment1+"grader_name",comment.getGraderName());
        props.setProperty(comment1+"aspect_x",comment.getAspectX());
        props.setProperty(comment1+"aspect_y",comment.getAspectY());
        if(comment.hasExternalBody()){
            props.setProperty(comment2+"extBodyFilename",comment.getExternalBodyFilename());
        }
    }
    
    /**
     * Puts the various files of the comment in the bls file.
     * @param draftNum The draft this comment belongs to.  Starts at 0.
     */
    private static void addComment(Blaise_Comment comment,JarOutputStream stream, int draftNum,int commentNum) throws IOException{
        JarEntry entry;
        String base;
        base="Draft"+draftNum+"/Comment"+commentNum+'/';
        addEFContainer(comment,stream,base+"Files/");
        Writer w;
        if(!comment.hasExternalBody()){
            //add the comment.html file.
            entry=new JarEntry(base+"comment.html");
            stream.putNextEntry(entry);
            w=new BufferedWriter(new OutputStreamWriter(stream));
            w.write(comment.getBody());
            w.flush();
            stream.closeEntry();
            
            //add the comment scribbles and notes files...
            entry=new JarEntry(base+"scribbles.props");
            stream.putNextEntry(entry);
            getScribblesProps(comment).store(stream,null);
            
            entry=new JarEntry(base+"notes.props");
            stream.putNextEntry(entry);
            getNotesProps(comment).store(stream,null);
			stream.closeEntry();
        }
    }
    
    /**
     * Returns a Properties object that contains all of the scribbles in the comment.
     */
    private static Properties getScribblesProps(Blaise_Comment comment){
        Properties sprops=new Properties();
        Vector scribs=comment.getScribbles();
        int size=scribs.size();
        sprops.setProperty("numScribs",Integer.toString(size));
        Scribble scrib;
        Color c;
        String base;
        for(int i=0;i<size;i++){
            base=new StringBuffer().append(i).append('.').toString();
            scrib=(Scribble)scribs.elementAt(i);
            c=scrib.getColor();
            sprops.setProperty(base+"red",Integer.toString(c.getRed()));
            sprops.setProperty(base+"blue",Integer.toString(c.getBlue()));
            sprops.setProperty(base+"green",Integer.toString(c.getGreen()));
            sprops.setProperty(base+"width",Integer.toString(scrib.getWidth()));
            sprops.setProperty(base+"offset",Integer.toString(scrib.getOffset()));
            sprops.setProperty(base+"height",Double.toString(scrib.getHeight()));
            sprops.setProperty(base+"xDist",Double.toString(scrib.getXDistanceFromOffset()));
            sprops.setProperty(base+"lines",scrib.getLineString());
        }
        return sprops;
    }
    
    /**
     * Returns a Properties object that contains all of the notes in the comment.
     */
    private static Properties getNotesProps(Blaise_Comment comment){
        Properties nprops=new Properties();
        Vector notes=comment.getNotes();
        int size=notes.size();
        nprops.setProperty("numNotes",Integer.toString(size));
        Note note;
        Color c;
        Point p;
        Rectangle r;
        String base;
        for(int i=0;i<size;i++){
            base=new StringBuffer().append(i).append('.').toString();
            note=(Note)notes.elementAt(i);
            c=note.getNoteColor();
            nprops.setProperty(base+"red",Integer.toString(c.getRed()));
            nprops.setProperty(base+"blue",Integer.toString(c.getBlue()));
            nprops.setProperty(base+"green",Integer.toString(c.getGreen()));
            nprops.setProperty(base+"nID",note.getNoteID());
            p=note.getIconPoint();
            nprops.setProperty(base+"iPx",Integer.toString((int)p.getX()));
            nprops.setProperty(base+"iPy",Integer.toString((int)p.getY()));
            r=note.getFrameRectangle();
            nprops.setProperty(base+"fPx",Integer.toString((int)r.getX()));
            nprops.setProperty(base+"fPy",Integer.toString((int)r.getY()));
            nprops.setProperty(base+"fPw",Integer.toString((int)r.getWidth()));
            nprops.setProperty(base+"fPh",Integer.toString((int)r.getHeight()));
            nprops.setProperty(base+"title",note.getTitle());
            nprops.setProperty(base+"text",note.getText());
            nprops.setProperty(base+"isOpen",Boolean.toString(note.isOpen()));
        }
        return nprops;
    }
    
    /**
     * Called by addComment and addDraft.  Does the real work of adding the externalFiles...
     * ExternalFiles are serialized as follows: first try the File, then the byte[], finally try to 
     * serialize the Object.
     * @param base the base of the name for the entries.  Ex: "Draft0/Comment2/Files/"
     */
    private static void addEFContainer(ExternalFileContainer cont,JarOutputStream stream, String base)throws IOException{
        JarEntry entry;
        File file;
        byte[] bArray;
        if(cont.hasExternalBody()){
            entry=new JarEntry(base+cont.getExternalBodyFilename());//don't put the entry in the stream until we know the data is there...
            
            //first try the File, then the byte array.
            file=cont.getExternalBodyLink();
            if(file==null){
                bArray=cont.getExternalBody();
                if(bArray==null){
                    //some sort of exception should be thrown here.
                    throw new IllegalStateException("Container claims to have external body, but does not.\nContainer:"+cont);
                }
                stream.putNextEntry(entry);
                stream.write(bArray);
            }
            else{
                stream.putNextEntry(entry);
                entry.setSize(file.length());
                DocumentUtilities.fileToStream(file,stream);
            }
        }
        
        //now do all of the external files...
        Vector files=cont.getExternalFiles();
        int size;
        if(files==null || (size=files.size())==0){
            return;//we're done.
        }
        
        ExternalFile ef;
        Serializable ser;
        for(int i=0;i<size;i++){
            ef=(ExternalFile)files.elementAt(i);
            entry=new JarEntry(base+ef.getFilename());
            //first try the File, then the byte[], finally try to serialize the Object.
            file=ef.getFileReference();
            if(file!=null){
                stream.putNextEntry(entry);
                entry.setSize(file.length());
                DocumentUtilities.fileToStream(file,stream);
            }
            else{
                bArray=ef.getByteArray();
                if(bArray!=null){
                    stream.putNextEntry(entry);
                    stream.write(bArray);
                }
                else{
                    try{
                        ser=(Serializable)ef.getObject();
                        //Note implimented for the moment.
                    }
                    catch(ClassCastException e){}//object isn't serializable...
                    catch(NullPointerException e){
                        throw new IllegalStateException("External File contains no data...");
                    }
                }
            }
        }
    }
}