package edu.gwu.blaise.version2.documents;
import edu.gwu.blaise.version2.util.Constants;
import java.io.*;
import java.util.jar.*;

/**
 * Adds Bls file opening functionality that 
 * is specific to the client.
 * 
 * @author Michael Gaiman
 * @version 5/16/2004
 */
public class ClientFileHandler extends BasicFileHandler{
    private String dir=null;
    public ClientFileHandler(){
        super.makeArrays=false;//don't waste the user's time by putting the files in memory.
    }
    
    /**
     * Override this method to have the paper unpack in to a different directory.
     * The contract is this method will be called after setPaper so that
     * the directory can, if needed, be based upon the properties of the paper being opened.
     */
    protected String getPaperDirectory(){
       return Constants.getClientPaperDirectory(); 
    }
    
    /**
     * Overridden to support the getPaperDirector() contract.
     */
    public void setPaper(Blaise_Paper paper){
        super.setPaper(paper);
        dir=getPaperDirectory();
    }
    
    /**
     * Duplicate a directory in the bls file in the file system.
     */
    protected void handleDirectoryEntry(JarEntry entry) throws IOException{
        super.handleDirectoryEntry(entry);
        String name=entry.getName();
        if(name.startsWith("META-INF/")){//don't output the meta-inf stuff
            return;
        }
        File f=new File(dir+name);
        f.mkdirs();
    }
    
    protected ExternalFile handleDraftEntry(JarEntry entry, int draftNum, int commentNum) throws IOException{
        ExternalFile ef=super.handleDraftEntry(entry,draftNum,commentNum);
        String name=entry.getName();
        Blaise_Draft draft=paper.getDraft(draftNum);
        
        //make the file in the file system, and copy the bytes there
        File file=makeFileAndCopy(entry);
        
        //find the right place in the Blaise_Paper heirarchy to put the File and put it there.
        if(commentNum==-1){//draft only stuff
        
            //check if this draft has an external body filename, and if this file is it.
            String extBfn=draft.getExternalBodyFilename();
            if(extBfn!=null && name.endsWith("Files/"+extBfn)){
                draft.setExternalBodyLink(file);
                return null;
            }
            
            if(ef==null){//some file that super didn't make an EF for... so we don't care now.
                return null;
            }
            
            //a random File, nothing special about it.  Add it to the external Files list.
            //already exists, for this draft (made via the super call).
            ef.setFileReference(file);
            return ef;
        }
        else{//comment related file
            Blaise_Comment comment=(Blaise_Comment)draft.getComments().elementAt(commentNum);
            
            //check if this comment has an external body filename, and if this file is it.
            String extBfn=comment.getExternalBodyFilename();
            if(extBfn!=null && name.endsWith("Files/"+extBfn)){
                comment.setExternalBodyLink(file);
                return null;
            }
            
            if(ef==null){//some file that super didn't make an EF for... so we don't care now.
                return null;
            }
            
            //a random File, nothing special about it.  Add it to the external Files list.
            //already exists, for this draft (made via the super call).
            ef.setFileReference(file);
            return ef;
        }
    }
    
    protected ExternalFile handleWorkingDraftEntry(JarEntry entry)  throws IOException{
        ExternalFile ef=super.handleWorkingDraftEntry(entry);
        String name=entry.getName();
        Blaise_Draft draft=paper.getWorkingDraft();
        
        //make the file in the file system, and copy the bytes there
        File file=makeFileAndCopy(entry);
        
        String extBfn=draft.getExternalBodyFilename();
        if(extBfn!=null && name.endsWith("Files/"+extBfn)){
            draft.setExternalBodyLink(file);
            return null;
        }
        
        if(ef==null){
            return null;
        }
        
        ef.setFileReference(file);
        return ef;
    }
    
    protected void handleMiscEntry(JarEntry entry)  throws IOException{
        //ignore anything in the META-INF dir.
        String name=entry.getName();
        if(name.startsWith("META-INF/")){
            return;
        }
        
        makeFileAndCopy(entry);
    }
    
    /**
     * Creates the File location on the filesystem and copies the data to that file.
     * @return a File reference
     */
    protected File makeFileAndCopy(JarEntry entry) throws IOException{
        //make the file in the file system, and copy the bytes there
        File file=new File(dir+entry.getName());
        /* check to see if the file already exists.  If so, either delete it
         * or keep it, depending upon its date and size.
         */
        if(file.exists()){
            System.out.println("existing entry[ name="+entry.getName()+" size="+entry.getSize()+"] file[ name="+file+" length="+file.length()+"]");
            if(entry.getSize()==file.length()){
                return file;
            }
            else{
                //delete it, just to be safe.
                file.delete();
            }
        }
        InputStream stream=getInputStream(entry);
        DocumentUtilities.inputStreamToFile(stream,file);
        stream.close();
        return file;
    }
    
    /**
     * Causes the directory structure to be deleted from the filesystem.
     */
    public void closePaper() throws IOException{
        //because this method is often used on its own, the dir
        //may not have been initialized.  check and get if needed
        if(dir==null){
            dir=getPaperDirectory();
        }
        deleteRecursively(new File(dir));
    }
    
    /**
     * Causes the File, and all of it's subfiles (if this is a directory)
     * to be deleted.
     */
    private void deleteRecursively(File f) throws IOException{
        if(f.isDirectory()){
            File[] files=f.listFiles();
            for(int i=0;i<files.length;i++){
                deleteRecursively(files[i]);
            }
        }
        
        System.out.println("Deleting file:"+f);
        f.delete();
    }
}
