package edu.gwu.blaise.version2.documents;
import java.io.*;
import java.util.*;
/**
 * DocumentUtilities is a collection of utility methods useful
 * in creating, opening, and saving Blaise documents.
 * @author  Michael Gaiman
 * @version 1 4/25/2004
 */
public class DocumentUtilities {
    
    /** Creates a new instance of DocumentUtilities */
    public DocumentUtilities() {}
    
    /**
     * Given an input stream, this method will collect all data from the
     * stream, and return it as a properly sized byte array.
     */
    public static byte[] getInputStreamAsByteArray(InputStream stream) throws IOException{
        ByteArrayOutputStream out=new ByteArrayOutputStream(1000);
        BufferedInputStream in=new BufferedInputStream(stream);
        int i;
        i=in.read();
        while(i>=0){
            out.write(i);
            i=in.read();
        }
        
        return out.toByteArray();
    }
    
    /**
     * Writes the byte array to the File specified, creating the file
     * and supporting directories if needed.
     */
    public static void bytesToFile(byte[] bytes, File file) throws IOException{
        if(!file.exists()){
            makeDirs(file);
            file.createNewFile();
        }
        
        FileOutputStream fos=new FileOutputStream(file);
        fos.write(bytes);
    }
    
    /**
     * Copies the data from an input stream and outputs it to the file specified,
     * creating the file and supporting directories, if needed.
     */
    public static void inputStreamToFile(InputStream stream, File file) throws IOException{
        if(!file.exists()){
            makeDirs(file);
            file.createNewFile();
        }
        
        int b;
        BufferedInputStream in=new BufferedInputStream(stream);
        BufferedOutputStream out=new BufferedOutputStream(new FileOutputStream(file));
        b=in.read();
        while(b>=0){
            out.write((byte)b);
            b=in.read();
        }
        out.close();
    }
    
    /**
     * Given a valid InputStream, the method will convert the contents of the
     * stream into a single String.
     */
    public static String getStreamContentsAsString(InputStream stream) throws IOException{
        BufferedReader br=new BufferedReader(new InputStreamReader(stream));
        StringBuffer sb=new StringBuffer();
        int i;
        
        i=br.read();
        while(i>=0){
            sb.append((char)i);
            i=br.read();
        }
        return sb.toString();
    }
    
    /**
     * Outputs the contents of File to the output stream.
     */
    public static void fileToStream(File file, OutputStream ostream) throws IOException{
        BufferedInputStream istream=new BufferedInputStream(new FileInputStream(file));
        BufferedOutputStream out=new BufferedOutputStream(ostream);
        int i;
        i=istream.read();
        while(i>=0){
            out.write(i);
            i=istream.read();
        }
        istream.close();
        out.flush();
    }
    
    /**
     * Copies the data in file original to file destination, creating destination
     * if necessary.
     */
    public static void duplicateFile(File original, File destination) throws IOException{
        if(!destination.exists()){
            makeDirs(destination);
            destination.createNewFile();
        }
        OutputStream out=new BufferedOutputStream(new FileOutputStream(destination));
        fileToStream(original,out);
        out.close();
    }
    
    /**
     * Given an ExternalFileContainer which uses File Links this utility method
     * will move all of the files to a specified directory and update the File Links.
     * @param newDir a File pointing to the new directory.
     * @returns true if successful.  False otherwise.
     * @see java.io.File.renameTo()
     */
    public static boolean moveExternalFiles(ExternalFileContainer container, File newDir) throws IOException{
        Vector files=container.getExternalFiles();
        File rename,temp;
        ExternalFile ef;
        int size=(files==null)?0:files.size();
        int i;
        boolean retVal=true;
        
        if(!newDir.exists()){
            makeDirs(newDir);
        }
        
        ((Blaise_Draft)container).print();
                
        for(i=0;i<size;i++){
            ef=(ExternalFile)files.elementAt(i);
            rename=new File(newDir,ef.getFilename());
            temp=ef.getFileReference();
            if(temp!=null){
                retVal=temp.renameTo(rename);
            }
            if(!retVal){//a failure occurred.
                System.out.println("Couldn't rename "+ef.getFileReference()+" to "+temp);
                return false;
            }
        }
        
        //now do the body, if it has one.
        if(container.hasExternalBody()){
            rename=new File(newDir,container.getExternalBodyFilename());
            temp=container.getExternalBodyLink();
            if(temp!=null){
                retVal=temp.renameTo(rename);
            }
        }
        return retVal;
    }
    
    /**
     * Creates the directories this file is contained in.
     * Does not check whether the file exists, or perform other checks.
     */
    public static void makeDirs(File file) throws IOException{
        File p=file.getParentFile();
        if(p==null){
            return; //no parent
        }
        p.mkdirs();
    }
}
