package edu.gwu.blaise.version2.grader;
import javax.swing.*;
import java.awt.*;
import java.beans.*;
/**
 * AbstractPortalPane is basically a template.
 * It defines four different areas of the pane, and
 * handles the laying out of those four general areas.
 * Subclasses override to specify what goes in the areas.
 * Laying out is done using a GridBagLayout.
 * @author Michael Gaiman
 * @version 1 6/7/2004
 */
abstract public class AbstractPortalPane extends JPanel implements PropertyChangeListener{
    /**
     * Calls the doLayout method, and registers to receive onlineStatus changes
     */
    public AbstractPortalPane(BlaiseGradingPortal parent){
        //notify kids of onlineStatus change...
        parent.addPropertyChangeListener("onlineStatus",this);
        SwingUtilities.invokeLater(new Runnable(){
            public void run(){
                doPaneLayout();
            }
        });
    }
    
    /**
     * Called by BlaiseGradingPortal when the online status changes.
     * Calls onlineStatusChange.
     */
    public void propertyChange(PropertyChangeEvent e){
        if(e.getPropertyName().equals("onlineStatus")){
            int type=((Integer)e.getNewValue()).intValue();
            System.out.println("onlineStatus property changed");
            onlineStatusChange(type);
        }
    }
    
    /**
     * Called when the grading portal has either gone online, or
     * offline.  (after initial invocation).
     * @param loginType one of the constant BlaiseGradingPortal login types
     * (eg LIVE or CACHED)
     */
    abstract protected void onlineStatusChange(int loginType);
    
    /**
     * PortalPanes have a JLabel offering a brief description
     * of the use of the pane.
     */
    abstract protected JLabel getHeader();
    
    /**
     * The upper panel is usually filled with options that effect
     * the lower panel.
     */
    abstract protected Component getUpperPanel();
    
    /**
     * The default grid bag constraints for the upper panel.
     * The default constraints are as follows:
     * c.fill=GridBagConstraints.NONE
     * c.gridx=0
     * c.gridy=1
     * c.weighty=.2
     */
    protected GridBagConstraints getUpperPanelConstraints(){
        GridBagConstraints c=new GridBagConstraints();
        //c.anchor=GridBagConstraints.PAGE_START;
        c.fill=GridBagConstraints.NONE;
        c.gridx=0;
        c.gridy=1;
        c.weighty=.1;
        return c;
    }
    
    /**
     * The lower panel is usually the main data view of the pane.
     */
    abstract protected Component getLowerPanel();
    
    /**
     * The default grid bag constraints for the lower panel.
     * The default constraints are as follows:
     * c.anchor=GridBagConstraints.PAGE_START
     * c.fill=GridBagConstraints.BOTH
     * c.gridx=0
     * c.gridy=2
     * c.weighty=1
     */
    protected GridBagConstraints getLowerPanelConstraints(){
        GridBagConstraints c=new GridBagConstraints();
        c.anchor=GridBagConstraints.PAGE_START;
        c.fill=GridBagConstraints.BOTH;
        c.gridx=0;
        c.gridy=2;
        c.weighty=1;
        c.weightx=1;
        return c;
    }
    
    /**
     * An array of buttons to be laid out along the bottom
     */
    abstract protected AbstractButton[] getBottomButtons();
    
    /**
     * Performs the actual layout.
     */
    protected void doPaneLayout(){
        setLayout(new GridBagLayout());
        GridBagConstraints c=new GridBagConstraints();
        c.gridx=0;
        c.gridy=0;
        c.fill=GridBagConstraints.HORIZONTAL;
        c.anchor=GridBagConstraints.FIRST_LINE_START;
        c.insets=new Insets(5,0,0,0);
        add(getHeader(),c);
        add(getUpperPanel(),getUpperPanelConstraints());
        add(getLowerPanel(),getLowerPanelConstraints());
        AbstractButton[] buttons=getBottomButtons();
        if(buttons.length>0){
            JPanel buttonPanel=new JPanel(new GridBagLayout());
            c=new GridBagConstraints();
            c.weightx=.5;
            c.gridx=0;
            c.gridy=0;
            for(int i=0;i<buttons.length;i++){
                c.gridx=i;
                buttonPanel.add(buttons[i],c);
            }
            c=new GridBagConstraints();
            c.gridx=0;
            c.gridy=3;
            c.weighty=.05;
            c.anchor=GridBagConstraints.PAGE_END;
            add(buttonPanel,c);
        }
    }
}
