package edu.gwu.blaise.version2.grader;
import edu.gwu.blaise.version2.documents.*;
import edu.gwu.blaise.version2.util.Constants;
import java.util.*;
import java.io.*;
/**
 * PaperStore handles keeps track of local papers for the grading portal.
 * PaperStore has multiple parts.
 * It runs a properties file that gets written to the filesystem when the
 * application closes.  The properties file stores information about the checked
 * out papers (the informaion that the tables show).  As well as the paper's
 * location on the filesystem.
 * @author Michael Gaiman
 * @version 1 6/7/2004
 */
class PaperStore {
    private boolean writeNeeded=true;
    private static final boolean debug=true;
    private String graderID;
    private String filename;
    private String backUpFilename;
    
    private Hashtable store;//by PaperID
    //stores course to Vector (of paperID String objects) mappings
    private Hashtable byCourse;
    private Hashtable byAssignment;
    private Vector paperIDs;
    
    /**
     * Open the specified grader's store.
     */
    public PaperStore(String graderID){
        this.graderID=graderID;
        store=new Hashtable();
        byCourse=new Hashtable();
        byAssignment=new Hashtable();
        paperIDs=new Vector();
        //the filename adds a W or a G so that there are no possible filename conflicts
        filename=Constants.getPaperStoreBase()+graderID+(BlaiseGradingPortal.isGrader()?"G":"W")+".props";
        backUpFilename=filename+".bak";
        //install a shutdown hook
        Runtime rt=Runtime.getRuntime();
        rt.addShutdownHook(new Thread(){
            public void run(){
                if(debug)System.out.println("In PaperStore shutdown hook");
                try{
                    if(writeNeeded){
                        writeStore();
                    }
                }catch(IOException e){
                    System.out.println("PaperStore could not serialize changes.  (Error: "+e+")");
                }
            }
        });
        
        //read in the properties...
        try{
            parseStore();
        }catch(IOException e){
            System.out.println("PaperStore could not parse properties. (Error: "+e+")");
        }
    }
    
    /**
     * Reads in the paper store file and sets up the data structures.
     */
    public void parseStore() throws IOException{
        File f=new File(filename);
        if(!f.exists()){
            System.out.println("Couldn't find PaperStore.");//not a big deal--probably just means this is the first time running...
            return;
        }
        Properties props=new Properties();
        InputStream in=new BufferedInputStream(new FileInputStream(f));
        props.load(in);
        in.close();
        int size=Integer.parseInt(props.getProperty("numEntries"));
        Entry entry;
        String base,subbase;
        String paperID,graderID;
        Vector assigns;
        int numIDs,j;
        for(int i=0;i<size;i++){
            base=new StringBuffer().append(i).append('.').toString();
            entry=new Entry();
            paperID=props.getProperty(base+"paperID");
            entry.setPaperID(paperID);
            entry.setCourseID(props.getProperty(base+"courseID"));
            entry.setCourse(props.getProperty(base+"course"));
            entry.setAssignmentID(props.getProperty(base+"assignmentID"));
            entry.setWriterID(props.getProperty(base+"writerID"));
            entry.setFirstName(props.getProperty(base+"firstName"));
            entry.setLastName(props.getProperty(base+"lastName"));
            graderID=props.getProperty(base+"graderID");
            entry.setGraderID(graderID);
            entry.setGrader(props.getProperty(base+"grader"));
            entry.setDraftCount(props.getProperty(base+"draftCount"));
            entry.setTurninDate(props.getProperty(base+"turninDate"));
            entry.setPaperFilePath(props.getProperty(base+"paperFilePath"));
            entry.setStatus(props.getProperty(base+"status"));
            addToAll(paperID,entry);
        }
    }
    
    /**
     * Adds the entry to all of the different data structures.
     */
    private void addToAll(String paperID,Entry entry){
        Object old=store.get(paperID);
        if(old==null){
            //simple case--not in store already
            store.put(paperID,entry);
            addToByTables(paperID,entry,false);
            paperIDs.addElement(paperID);
        }
        else{
            store.put(paperID,entry);
            addToByTables(paperID,entry,true);
            //no need to add to paperIDs.
        }
    }
    
    /**
     * Adds the entry's paperID to the byCourse and byAssignment tables.
     * note, no duplication checking is done.
     * @param exists true if an earlier version of this paperID already exists in the store.
     */
    private void addToByTables(String paperID,Entry entry,boolean exists){
        Vector v;
        String cID=entry.getCourseID();
        v=(Vector)byCourse.get(cID);
        if(v==null){
            //add new vector
            v=new Vector();
            v.addElement(paperID);
            byCourse.put(cID,v);
        }
        else{
            //vector already exists, just add it to the vector.
            if(exists){
                if(v.indexOf(paperID)==-1){
                    v.addElement(paperID);
                }
            }
            else{
                v.addElement(paperID);
            }
        }
        
        //for each assignment id, add it to the byAssignment table.
        String aID=entry.getAssignmentID();
        v=(Vector)byAssignment.get(aID);
        if(v==null){
            //add new vector
            v=new Vector();
            v.addElement(paperID);
            byAssignment.put(aID,v);
        }
        else{
            //vector already exists, just add it to the vector.
            if(exists){
                if(v.indexOf(paperID)==-1){
                    v.addElement(paperID);
                }
            }
            else{
                v.addElement(paperID);
            }
        }
    }

    
    /**
     * By default the paper store gets written out as a runtime shutdown hook.
     * To turn that off, set write needed to be false.
     */
    public void setWriteNeeded(boolean needed){
        writeNeeded=needed;
    }
    
    /**
     * Saves the current status of the store to the properties file.
     * This method (unless disabled via setWriteNeeded) is called upon
     * shutdown.
     */
    public void writeStore() throws IOException{
        Properties props=new Properties();
        int size=store.size();
        props.setProperty("numEntries",Integer.toString(size));
        String base;
        Entry entry;
        int i=0,j,numIDs;
        for(Enumeration e=store.elements();e.hasMoreElements();i++){
            base=new StringBuffer().append(i).append('.').toString();
            entry=(Entry)e.nextElement();
            props.setProperty(base+"paperID",entry.getPaperID());
            props.setProperty(base+"courseID",entry.getCourseID());
            props.setProperty(base+"course",entry.getCourse());
            props.setProperty(base+"assignmentID",entry.getAssignmentID());
            props.setProperty(base+"writerID",entry.getWriterID());
            props.setProperty(base+"firstName",entry.getFirstName());
            props.setProperty(base+"lastName",entry.getLastName());
            props.setProperty(base+"graderID",entry.getGraderID());
            props.setProperty(base+"grader",entry.getGrader());
            props.setProperty(base+"draftCount",entry.getDraftCount());
            props.setProperty(base+"turninDate",entry.getTurninDate());
            props.setProperty(base+"paperFilePath",entry.getPaperFilePath());
            props.setProperty(base+"status",entry.getStatus());
        }
        
        File f=new File(backUpFilename);
        File f2=new File(filename);
        try{
            if(!f.exists()){
                File p=f.getParentFile();
                if(p!=null){
                    p.mkdirs();
                }
                f.createNewFile();
            }
            OutputStream out=new BufferedOutputStream(new FileOutputStream(f));
            props.store(out,null);
            out.flush();
            out.close();
        }
        catch(IOException e){
            throw new IOException("Exception while writing backup paper store.  The original contents are at "+f2.getCanonicalPath()+" whereas the failed back up is at"+f.getCanonicalPath()+" Exception:"+e);
        }
        boolean internalThrow=false;
        try{
            f2.delete();//windows seems to require that rename not be to an existing file...
            if(!f.renameTo(f2)){
                System.out.println("Couldn't rename backup file.");
                internalThrow=true;
                throw new IOException("PaperStore couldn't rename backup file.  Contents are stored in "+f.getCanonicalPath()+" please rename to "+f2.getCanonicalPath()+" manually");
            }
        }catch(IOException e){
            if(internalThrow){
                throw e;
            }
            else{
                throw new IOException("Exception while renaming backup paper store to primary.  The newest contents are at "+f.getCanonicalPath()+" please rename manually to "+f2.getCanonicalPath()+" (Exception:"+e+')');
            }
        }
    }
    
    /**
     * Add a paper to the store.  (In memory only, to write out call writeStore()
     * after this)
     * The extra parameters apart from the paper are for human readable table fields.
     */
    public synchronized void addPaper(Blaise_Paper paper, String course, String grader, String paperFilePath, String status){
        Entry entry=new Entry();
        String paperID=paper.getPaperID();
        entry.setPaperID(paperID);
        entry.setCourseID(paper.getCourseID());
        entry.setCourse(course);
        
        entry.setAssignmentID(paper.getAssignmentID());
        entry.setWriterID(paper.getWriterID());
        entry.setFirstName(paper.getFirst());
        entry.setLastName(paper.getLast());
        entry.setGraderID(graderID);
        entry.setGrader(grader);
        entry.setDraftCount(Integer.toString(paper.getDrafts().size()));
        Blaise_Draft draft=(Blaise_Draft)paper.getDrafts().lastElement();
        entry.setTurninDate(draft.getTurninDate());
        entry.setPaperFilePath(paperFilePath);
        entry.setStatus(status);
        if(debug)System.out.println("PaperStore adding entry:"+entry);
        addToAll(paperID,entry);
    }
    
    /**
     * Given a paperID, getEntry returns the corrsponding PaperStore.Entry object.
     * (null if no matching paperID).
     */
    public synchronized PaperStore.Entry getEntry(String paperID){
        return (PaperStore.Entry)store.get(paperID);
    }
    
    /**
     * Returns a vector containing paperIDs (as Strings) that match that courseID.
     * Or null if there are none.
     * Vector should be thought of as read-only
     */
    public synchronized Vector getEntriesByCourse(String courseID){
       return (Vector)byCourse.get(courseID);
    }
    
    /**
     * Returns a vector containing paperIDs (as Strings) that match that assignmentID.
     * Or null if there are none.
     * Vector should be thought of as read-only
     */
    public synchronized Vector getEntriesByAssignment(String assignmentID){
        return (Vector)byAssignment.get(assignmentID);
    }
    
    /**
     * Returns a vector containing paperIDs (as Strings) of all the papers in the store.
     * Vector should be thought of as read-only.
     */
    public synchronized Vector getPaperIDs(){
        return paperIDs;
    }
    
    /**
     * Returns an enumeration of all the paper IDs (in String form)
     */
    public Enumeration ids(){
        return store.keys();
    }
    
    /**
     * Returns an enumeration of all of the Entry objects.
     */
    public Enumeration entries(){
        return store.elements();
    }
    
    /**
     * PaperStore.Entry objects can, externally, be thought of as read only.
     * This method allows callers to change the values in the paperFilePath field,
     * and the status field.
     * (if you don't want to change both fields, leave one null.)
     */
    public void modifyEntry(PaperStore.Entry entry, String paperFilePath, String status){
        if(paperFilePath!=null){
            entry.setPaperFilePath(paperFilePath);
        }
        if(status!=null){
            entry.setStatus(status);
        }
    }
    
    /**
     * A convenience method.  PaperStore will find the entry matching paperID and
     * will then call modifyEntry(PaperStore.Entry,String,String)
     * @throws IllegalArgumentException if paperID is not in the store.
     */
    public void modifyEntry(String paperID, String paperFilePath, String status){
        Entry e=getEntry(paperID);
        if(e==null){
            throw new IllegalArgumentException("PaperID does not match a paper in the store");
        }
        modifyEntry(e,paperFilePath,status);
    }
    
    /**
     * Removes the entry from the store.  (This is done in memory only,
     * to save changes you must call writeStore() after this).
     * @throws NullPointerException if entry is null or not in the paperStore
     */
    public synchronized void removeEntry(PaperStore.Entry entry){
        String paperID=entry.getPaperID();
        store.remove(paperID);
        paperIDs.remove(paperID);
        Vector v=(Vector)byCourse.get(entry.getCourseID());
        v.remove(paperID);
        if(v.size()==0){
            byCourse.remove(entry.getCourseID());
        }
        v=(Vector)byAssignment.get(entry.getAssignmentID());
        v.remove(paperID);
        if(v.size()==0){
            byAssignment.remove(entry.getAssignmentID());
        }
    }
    
    /**
     * removes the entry with the specified paperID from the store.
     */
    public synchronized void removeEntry(String paperID){
        removeEntry((Entry)store.get(paperID));
    }
    
    /**
     * A paper store Entry holds information about a single paper in the
     * paper store.
     * @author Michael Gaiman
     * @version 1 6/7/2004
     */
    public class Entry{
        private String courseID;
        private String course;
        private String assignmentID;
        private String writerID;
        private String firstName;
        private String lastName;
        private String paperID;
        private String graderID;
        private String grader;
        private String draftCount;
        private String turninDate;
        private String paperFilePath;
        private String status;
        
        Entry(){
            /*courseID="";
            course="";
            assignmentIDs=new Vector();
            writerID="";
            firstName="";
            lastName="";
            paperID="";
            graderID="";
            grader="";
            draftCount="";
            turninDate="";
            paperFilePath="";
            status="";
            */
        }
        
        void setCourseID(String val){
            courseID=val;
        }
        public String getCourseID(){
            return courseID;
        }
        void setCourse(String val){
            course=val;
        }
        /**
         * String that serves as a written representation of a course (usually
         * DEPT NUMBER-SECTION
         */
        public String getCourse(){
            return course;
        }
        void setAssignmentID(String val){
            assignmentID=val;
        }
        public String getAssignmentID(){
            return assignmentID;
        }
        void setWriterID(String val){
            writerID=val;
        }
        public String getWriterID(){
            return writerID;
        }
        void setFirstName(String val){
            firstName=val;
        }
        public String getFirstName(){
            return firstName;
        }
        void setLastName(String val){
            lastName=val;
        }
        public String getLastName(){
            return lastName;
        }
        void setPaperID(String val){
            paperID=val;
        }
        public String getPaperID(){
            return paperID;
        }
        void setGraderID(String val){
            graderID=val;
        }
        /**
         * Returns the graderID of the grader who checked the paper out.
         */
        public String getGraderID(){
            return graderID;
        }
        void setGrader(String val){
            grader=val;
        }
        /**
         * A string representing the grader who checked the paper out, usually their username.
         */
        public String getGrader(){
            return grader;
        }
        void setDraftCount(String val){
            draftCount=val;
        }
        public String getDraftCount(){
            return draftCount;
        }
        void setTurninDate(String val){
            turninDate=val;
        }
        public String getTurninDate(){
            return turninDate;
        }
        void setPaperFilePath(String val){
            paperFilePath=val;
        }
        public String getPaperFilePath(){
            return paperFilePath;
        }
        void setStatus(String val){
            status=val;
        }
        /**
         * String containing the grading status of this paper.
         */
        public String getStatus(){
            return status;
        }
        
        public String toString(){
            return '['+"paperID:"+paperID+" courseID:"+courseID+" course:"+
                course+" assignmentID:"+assignmentID+" writerID:"+writerID+
                " firstName:"+firstName+" lastName:"+lastName+" graderID:"+
                graderID+" grader:"+grader+" draftCount:"+draftCount+
                " turninDate:"+turninDate+" paperFilePath:"+paperFilePath+
                " status:"+status+']';
        }
    }
}
