//LOC:303
package edu.gwu.blaise.version2.recognizer;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import org.w3c.dom.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

import edu.gwu.blaise.version2.util.*;

/**
 * Creates or parses an XML file that contains gesture data (name and sample).
 *
 * @author Sean T. Hanlon
 */
public class GestureXML {

	private File file;			//file of gesture set
	private Vector names;		//names of gestures
	private Vector gestures;	//the gestures themselves

	/**********************************************************************************************
	* Constructor and Accessor/Mutators
	**********************************************************************************************/

	/**
	 * Constructor
	 *
	 * @param file file to be read-written into
	 */
	public GestureXML(File file){
		this.file = file;
	}

	/**
	 * Returns the vector of names
	 *
	 * @return <code>Vector</code> of names
	 */
	public Vector getNames(){
		return names;
	}

	/**
	 * Returns the vector of samples
	 *
	 * @return <code>Vector</code> of samples
	 */
	 public Vector getSamples(){
		return gestures;
	}

	/**********************************************************************************************
	* XML Generation
	**********************************************************************************************/

	/**
	 * Generates XML into file specified in constructor.
	 *
	 * @param names Names of gestures
	 * @param gests The gestures
	 */
	 public void generateXML(Vector names, Vector gests){

		this.names = names;

		StringBuffer buffer = new StringBuffer();
		buffer.append("<!DOCTYPE GESTURE_FILE[" + Constants.getGestureDTD() + "]>\n");
		buffer.append("<GESTURE_FILE>\n");

		for(int i=0;i<names.size();i++){
			buffer.append("<GESTURE>\n");
			buffer.append("     <name>"+(String)names.elementAt(i)+"</name>\n");
			buffer.append("     <sample>"+generateLines((Vector)gests.elementAt(i))+"</sample>\n");
			buffer.append("</GESTURE>\n");
		}

		buffer.append("</GESTURE_FILE>\n");

		writeToFile(file, buffer.toString());
	}

	//Returns a string representation of the lines of the gesture
	private String generateLines(Vector lines){

		StringBuffer buffer = new StringBuffer();
		for(int i=0;i<lines.size();i++){
			Line2D.Double line = (Line2D.Double)lines.elementAt(i);
			buffer.append("|"+line.getX1()+","+line.getY1());

			if(i == (lines.size()-1)){
				buffer.append("|"+line.getX2()+","+line.getY2());
			}
		}

		return (buffer.toString());
	}

	//Writes XML to file
	private void writeToFile(File file, String values){

		try{
			//if file exists, delete it
			if(file.exists()){
				file.delete();
			}

			if(!file.createNewFile()){
				System.out.println("GestureXML -- Could not create save file");
			}
			else{
				FileWriter out = new FileWriter(file,false);
				out.write(values,0,values.length());
				out.close();
			}
		}
		catch(IOException e){
			System.out.println("GestureXML -- IOException Found in writeToFile:" + e);
		}
	}

	/**********************************************************************************************
	* XML Parsing
	**********************************************************************************************/

	/**
	 * Parses XML from file specified in constructor.
	 *
	 * @return <code>true</code> if successful, <code>false</code> otherwise
	 */
	public boolean parseXML(){

		names = new Vector();
		gestures = new Vector();

		try{
			// create Document
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			DocumentBuilder db = dbf.newDocumentBuilder();
			Document doc = db.parse(file);
			traverseTree(doc);
		}
		catch(ParserConfigurationException e){
			return false;
		}
		catch(SAXException e){
			return false;
		}
		catch(IOException e){
			return false;
		}

		return true;
	}

	//Traverses a node and takes appropriate action
	private void traverseTree(Node node){
		int type = node.getNodeType();

        switch (type) {
        case Node.ATTRIBUTE_NODE:
            break;
        case Node.ELEMENT_NODE:

			// start of an event
			if (node.getNodeName().equals("GESTURE")) {

				NodeList childList = node.getChildNodes();

				// loop thru all the properties of the object
				for(int i=0;i<childList.getLength();i++) {

					Node childNode = childList.item(i);

					if (childNode.getNodeType() == Node.ELEMENT_NODE) {

						// get the node info
						String name = childNode.getNodeName();
						String value;

						// the if statement is there because description maybe empty therefore as no child
						if (childNode.getFirstChild() != null) {
							value = childNode.getFirstChild().getNodeValue();
						}
						else {
							value = "";
						}

						if(name.equals("name")){
							names.add(value);
						}
						else if(name.equals("sample")){
							gestures.add(generateVector(value));
						}
					}
				}
			}

			break;
        default:
            break;
        }

		// Recursively call children
        for (Node child = node.getFirstChild(); child != null; child = child.getNextSibling()) {
            traverseTree(child);
        }
	}

	//Returns a vector of line segments that represent a gesture
	private Vector generateVector(String input){
		Vector lines = new Vector();

		StringTokenizer token = new StringTokenizer(input, "|");
		double x0 = -1.0;
		double y0 = -1.0;
		while(token.hasMoreElements()){
			StringTokenizer token2 = new StringTokenizer((String)token.nextElement(),",");
			if(x0 == -1.0 && y0 == -1.0){
				x0 = Double.parseDouble((String)token2.nextElement());
				y0 = Double.parseDouble((String)token2.nextElement());
			}
			else{
				double x1 = Double.parseDouble((String)token2.nextElement());
				double y1 = Double.parseDouble((String)token2.nextElement());
				lines.add(new Line2D.Double(x0,y0,x1,y1));
				x0 = x1;
				y0 = y1;
			}
		}

		return lines;
	}

	/**********************************************************************************************
	* Local Testing
	**********************************************************************************************/

	public static void main(String[] argv){

		File f = new File("E:\\SeniorDesign\\Versions\\data\\test");

		Line2D.Double l1 = new Line2D.Double(0,0,1,1);
		Line2D.Double l2 = new Line2D.Double(1,1,2,2);
		Line2D.Double l3 = new Line2D.Double(2,2,3,3);
		Line2D.Double l4 = new Line2D.Double(3,3,4,4);
		Line2D.Double l5 = new Line2D.Double(4,4,5,5);

		Vector gestures = new Vector();

		Vector gest = new Vector();
		gest.add(l1);
		gest.add(l2);
		gest.add(l3);
		gest.add(l4);
		gest.add(l5);
		gestures.add(gest);
		Vector names = new Vector();
		names.add(new String("Circle"));

		Line2D.Double l6 = new Line2D.Double(0,0,1,1);
		Line2D.Double l7 = new Line2D.Double(1,1,2,2);
		Line2D.Double l8 = new Line2D.Double(2,2,3,3);
		Line2D.Double l9 = new Line2D.Double(3,3,4,4);
		Line2D.Double l10 = new Line2D.Double(4,4,5,5);
		Line2D.Double l11 = new Line2D.Double(5,5,10,10);

		gest = new Vector();
		gest.add(l6);
		gest.add(l7);
		gest.add(l8);
		gest.add(l9);
		gest.add(l10);
		gest.add(l11);
		gestures.add(gest);
		names.add(new String("Star"));

		GestureXML xml = new GestureXML(f);
		xml.generateXML(names, gestures);

		if(xml.parseXML()){
			Vector n = xml.getNames();
			Vector samples = xml.getSamples();

			for(int i=0;i<n.size();i++){
				System.out.println("____________________________________");
				System.out.println((String)n.elementAt(i));
				Vector lines = (Vector)samples.elementAt(i);
				for(int j=0;j<lines.size();j++){
					Line2D.Double l = (Line2D.Double)lines.elementAt(j);
					System.out.println(l.getX1()+","+l.getY1()+" --> "+l.getX2()+","+l.getY2());
				}
			}
		}
		else{
			System.out.println("Parsing Failed");
		}
	}
}