//LOC:642
package edu.gwu.blaise.version2.recognizer;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.*;

import edu.gwu.blaise.version2.util.*;

/**
 * The "brain" of the gesture recognition - the Recognizer performs the gesture recognition
 *
 * @author Sean T. Hanlon
 */
public class Recognizer{

	private boolean debug = false;

	Vector gestures;	//Vector of gestures
	int meshSize;		//mesh size
	double[][]weights;	//weight array used in matching
	String filename;	//filename load or save to

	/**
	 * Constructor
	 *
	 * @param f File that contains, or will contain the gesture dictionary
	 */
	public Recognizer(File f){
		gestures = new Vector();
		meshSize = Constants.getMeshSize();

		//open gesture file from gestureFile
		if(f.exists()){
			openFile(f);
		}
		else{

		}

		filename = f.getName();
	}

	///////////////////////////////////////////////////
	// Accessors and Mutators
	///////////////////////////////////////////////////

	/**
	 * Adds a new gesture to the currently active gesture set
	 *
	 * @param name The name of the new gesture
	 * @param gestu The gesture itself
	 */
	public boolean addGesture(String name, Vector gestu){

		//Create and add new gesture to dictionary
		gesture gest = new gesture(name, gestu, meshSize);
		gestures.add(gest);

		if(debug){
			System.out.println("---------------------------------------------");
			System.out.println("Added New Entry : "+name);
			System.out.print("Signature: ");
			Vector cells = gest.getCells();
			for(int i=0;i<cells.size();i++){
				System.out.print(((Integer)cells.elementAt(i)).intValue());
				System.out.print(", ");
			}
			System.out.println();
		}

		//Sort and Refill gestures vector
		Object[] gestures_sorted = gestures.toArray();
		Arrays.sort(gestures_sorted);
		gestures = new Vector();
		for(int i=0;i<gestures_sorted.length;i++){
			gestures.add((gesture)gestures_sorted[i]);
		}

		return true;
	}

	/**
	 * Matches a given gesture to the closest one in the dictionary
	 *
	 * @param lines The gesture to be matched
	 */
	public String match(Vector lines){

		//Create new gesture and get signature of it
		gesture input = new gesture("Input",lines,meshSize);
		Vector inputCells = input.getCells();

		//For each gesture in dictionary
		double min = -1;
		String min_name = "";
		for(int i=0;i<gestures.size();i++){
			gesture current = (gesture)gestures.elementAt(i);			//current gesture in dictionary
			if(debug)System.out.println("Matching "+current.getName());
			Vector dict = current.getCells();							//the signature of the current gesture

			Rectangle2D.Double[] inputRects = normalizeGrid(current.getRects(), input.getRects());	//normalize the grids to be the same size

			//initialize weights
			weights = new double[inputCells.size()+1][dict.size()+1];
			weights[0][0] = 0.0;
			for(int x=1;x<inputCells.size();x++){
				weights[x][0] = weights[x-1][0] + gapScore1(inputCells.elementAt(x),current.getRects());
			}
			for(int y=1;y<dict.size();y++){
				weights[0][y] = weights[0][y-1] + gapScore2(dict.elementAt(y),current.getRects());
			}

			//For every cell in source signature
			for(int a=1;a<=inputCells.size();a++){
				//For every cell in the current dictionary item's signature
				for(int b=1;b<=dict.size();b++){

					//Want to find the smallest value from a dirrect match, and then a gap in either direction
					double value = weights[a-1][b] + gapScore2(dict.elementAt(b-1), current.getRects());

					if(value > weights[a][b-1] + gapScore1(inputCells.elementAt(a-1),inputRects)){
						value = weights[a][b-1] + gapScore1(inputCells.elementAt(a-1),inputRects);
					}

					if(value > (weights[a-1][b-1] + matchScore((Integer)inputCells.elementAt(a-1), (Integer)dict.elementAt(b-1), inputRects, current.getRects()))){
						value = (weights[a-1][b-1] + matchScore((Integer)inputCells.elementAt(a-1), (Integer)dict.elementAt(b-1), inputRects,  current.getRects()));
					}

					weights[a][b] = value;
				}

			}

			//See if the current gesture from the dictionary has the smallest found score
			if(min == -1 || weights[inputCells.size()][dict.size()] < min){
				min = weights[inputCells.size()][dict.size()];
				min_name = current.getName();
			}

			if(debug)System.out.println("Value: "+weights[inputCells.size()][dict.size()]);
			if(debug)System.out.println("------------------------");
		}

		if(debug)System.out.println("***************************************");

		return min_name;
	}


	//Returns score of a match between two cells
	//Parameters are indexes into rects[]
	private double matchScore(Integer c1, Integer c2, Rectangle2D.Double[] rects1, Rectangle2D.Double[] rects2){

		int cell1 = c1.intValue();
		int cell2 = c2.intValue();

		if(cell1 < 0 || cell2 < 0){
			return rects1[0].getWidth();
		}

		Rectangle2D.Double rect1 = rects1[cell1];
		Rectangle2D.Double rect2 = rects2[cell2];

		Point2D.Double r1 = new Point2D.Double(rect1.getX(), rect1.getY());

		double d = r1.distance(rect2.getX(), rect2.getY());

		//Return distance squared so as to make this score more effective
		return (d*d);
	}

	//Returns value of matching the given index with a gap in first String
	private double gapScore1(Object cell, Rectangle2D.Double[] rect){

		if(rect.length > 0){
			return rect[0].getWidth();
		}
		else{
			return 0.0;
		}
	}

	//Returns value of matching the given index with a gap in second String
	private double gapScore2(Object cell, Rectangle2D.Double[] rect){

		if(rect.length > 0){
			return rect[0].getWidth();
		}
		else{
			return 0.0;
		}
	}

	//Normalizes the rectangle[]'s to be the same size
	private Rectangle2D.Double[] normalizeGrid(Rectangle2D.Double[] goal, Rectangle2D.Double[] change){

		double goal_width = goal[0].getWidth();
		double change_width = change[0].getWidth();

		double scale = goal_width/change_width;

		Rectangle2D.Double[] output = new Rectangle2D.Double[change.length];
		for(int i=0;i<output.length;i++){
			output[i] = new Rectangle2D.Double(change[i].getX()*scale, change[i].getY()*scale, change[i].getWidth()*scale, change[i].getHeight()*scale);
		}

		return output;
	}

	/**
	 * Opens a file
	 *
	 * @param file The file to be opened.
	 */
	public void openFile(File file){
		//Utilize GestureXML utiliy to parse XML and fill vector of gestures
		GestureXML xml = new GestureXML(file);
		if(xml.parseXML()){
			Vector names = xml.getNames();
			Vector samples = xml.getSamples();

			for(int i=0;i<names.size();i++){
				String n = (String)names.elementAt(i);
				Vector s = (Vector)samples.elementAt(i);

				gestures.add(new gesture(n,s,meshSize));
			}
		}
	}

	/**
	 * Saves current dictionary to a file
	 *
	 * @param file The file to write into
	 */
	public void writeToFile(File file){
		//Utilize GestureXML utility to generate XML and write to file
		GestureXML xml = new GestureXML(file);
		Vector names = new Vector();
		Vector samples = new Vector();
		for(int i=0;i<gestures.size();i++){
			gesture g = (gesture)gestures.elementAt(i);
			names.add((String)g.getName());
			samples.add((Vector)g.getSample());
		}

		xml.generateXML(names,samples);
	}

	/**
	 * Returns the names of all gestures in dictionary
	 *
	 * @returns <code>Vector</code> of names
	 */
	public Vector getNames(){

		Vector output = new Vector();
		for(int i=0;i<gestures.size();i++){
			gesture current = (gesture)gestures.elementAt(i);
			output.add(current.getName());
		}

		return output;
	}

	/**
	 * Returns the samples of all gestures in dictionary
	 *
	 * @returns <code>Vector</code> of samples
	 */
	public Vector getSamples(){

		Vector output = new Vector();
		for(int i=0;i<gestures.size();i++){
			gesture current = (gesture)gestures.elementAt(i);
			output.add(current.getSample());
		}

		return output;
	}

	/**
	 * Returns the rectangle grids of all gestures in dictionary
	 *
	 * @returns <code>Vector</code> of rectangle grids
	 */
	public Vector getRects(){

		Vector output = new Vector();
		for(int i=0;i<gestures.size();i++){
			gesture current = (gesture)gestures.elementAt(i);
			output.add(current.getRects());
		}

		return output;
	}

	/**
	 * Removes an entry from the dictionary
	 *
	 * @param name The name of the gesture to remove
	 */
	public void removeEntry(String name){

		for(int i=0;i<gestures.size();i++){
			gesture current = (gesture)gestures.elementAt(i);
			if((current.getName()).equals(name)){
				gestures.removeElementAt(i);
				break;
			}
		}
	}

	/**
	 * Checks to see if a name is already in use by a gesture in the dictionary
	 *
	 * @param name The name of the gesture
	 * @return <code>true</code> if the name exists, <code>false</code> otherwise
	 */
	public boolean nameExists(String name){

		for(int i=0;i<gestures.size();i++){
			gesture current = (gesture)gestures.elementAt(i);
			if((current.getName()).equalsIgnoreCase(name)){
				return true;
			}
		}

		return false;
	}

	/**
	 * Initializes a new dictionary
	 */
	public void newSet(){
		gestures = new Vector();
	}

	/**
	 * Returns a JPanel representation of an input gesture
	 *
	 * @param input The gesture to display
	 * @return <code>Gesture_Panel</code> of gesture
	 */
	public static Gesture_Panel getPanel(Vector input){
		gesture ges = new gesture("",input,Constants.getMeshSize());
		Vector cells = ges.getCells();
		return new Gesture_Panel(ges.getSample(), ges.getRects());
	}

	public static void main(String argv[]){

		//File f = new File("E:\\SeniorDesign\\Versions\\data\\test");
		//Recognizer rec = new Recognizer(f);
		File f = new File("E:\\SeniorDesign\\Versions\\data\\test2");
		Recognizer rec = new Recognizer(f);
		//rec.writeToFile(f);
		while(true){
			GestureCreator gest = new GestureCreator(new JFrame(),"Name of Gesture");
			Vector input = new Vector();
			while(true){
				if(gest.isDone()){
					input = gest.getGesture();
					break;
				}
			}
			rec.addGesture("Test",input);
			//System.out.println("Matched: "+rec.match(input));
		}
	}
}

//Object representation of a gesture
class gesture implements Comparable{

	String name;	//name of the gesture
	Vector sample;	//the sample of it
	Vector cells;	//the signature of it

	Rectangle2D.Double[] rects;	//the grid

	double min_x = -1.0;
	double min_y = -1.0;
	double max_x = -1.0;
	double max_y = -1.0;

	////////////////////////////////////////////////////////////////////////////////////////////
	//
	// Accessor and Mutators
	//
	///////////////////////////////////////////////////////////////////////////////////////////
	public String getName(){
		return name;
	}

	public Vector getSample(){
		return sample;
	}

	public Rectangle2D.Double[] getRects(){
		return rects;
	}

	public Vector getCells(){
		return cells;
	}


	public gesture(){}

	public gesture(String n, Vector s, int meshSize){
		name = n;
		sample = normalize(s);

		createRects(meshSize);
		createSignature();
	}

	private Vector normalize(Vector lines){

		min_x = -1.0;
		max_x = -1.0;
		min_y = -1.0;
		max_y = -1.0;

		//determine smallest and largest
		for(int i=0;i<lines.size();i++){
			Line2D.Double current = (Line2D.Double)lines.elementAt(i);

			//max x
			if(current.getX1() > max_x || max_x == -1.0){
				max_x = current.getX1();
			}
			if(current.getX2() > max_x || max_x == -1.0){
				max_x = current.getX2();
			}

			//min x
			if(current.getX1() < min_x || min_x == -1.0){
				min_x = current.getX1();
			}
			if(current.getX2() < min_x || min_x == -1.0){
				min_x = current.getX2();
			}


			//max y
			if(current.getY1() > max_y || max_y == -1.0){
				max_y = current.getY1();
			}
			if(current.getY2() > max_y || max_y == -1.0){
				max_y = current.getY2();
			}

			//min y
			if(current.getY1() < min_y || min_y == -1.0){
				min_y = current.getY1();
			}
			if(current.getY2() < min_y || min_y == -1.0){
				min_y = current.getY2();
			}
		}

		//check for bad input case
		if(max_x == -1.0 || max_y == -1.0 || min_x == -1.0 || min_y == -1.0){
			return new Vector();
		}

		//divide all points to be between 0 and 1
		double max = max_x;
		if(max_y > max_x){
			max = max_y;
		}

		if(max <= 1.0){
			max = 1.0;
		}

		min_x = min_x/max;
		min_y = min_y/max;
		max_x = max_x/max;
		max_y = max_y/max;

		Vector output = new Vector();
		//walk through points and normalize using largest found value
		for(int i=0;i<lines.size();i++){
			Line2D.Double current = (Line2D.Double)lines.elementAt(i);
			Line2D.Double newLine = new Line2D.Double(((current.getX1()/max)-min_x),((current.getY1()/max)-min_y),((current.getX2()/max)-min_x),((current.getY2()/max)-min_y));
			output.add(newLine);
		}

		//update values
		max_x = max_x - min_x;
		max_y = max_y - min_y;
		min_x = 0.0;
		min_y = 0.0;

		return output;
	}

	//Given the hash size, creates a grid of rectangles
	private void createRects(int mesh){

		double interval = -1.0;
		if(max_x > max_y){
			interval = (double)max_x/(double)mesh;
		}
		else{
			interval = (double)max_y/(double)mesh;
		}

		//Create rectangles of mesh
		rects = new Rectangle2D.Double[mesh*mesh];
		double x = 0.0;
		double y = 0.0;
		int location = 0;
		for(int i=0;i<mesh;i++){
			for(int j=0;j<mesh;j++){
				rects[location] = new Rectangle2D.Double(x,y,interval,interval);
				location++;
				x += interval;
			}
			x = 0.0;
			y += interval;
		}
	}

	//Creates the signature of the gesture
	private void createSignature(){

		int current = -1;
		cells = new Vector();
		//for every point
		for(int i=0;i<sample.size();i++){
			Line2D.Double line = (Line2D.Double)sample.elementAt(i);
			int next = getCellNumber(line.getX1(),line.getY1());

			if(current == -1){
				current = next;
			}

			if(next != current){
				cells.add(new Integer(current));
				current = next;
			}

			if(i == sample.size()-1){
				cells.add(new Integer(current));
				next = getCellNumber(line.getX2(), line.getY2());
				if(next != current){
					cells.add(new Integer(next));
				}
			}
		}
	}

	//Called from determineCells and return int of cell that line intersects
	private int getCellNumber(double x, double y){

		//check each rectangle to see if point is in rectangle
		for(int i=0;i<rects.length;i++){
			if(rects[i].contains(x,y) || rects[i].intersectsLine(x,y,x,y)){
				return i;
			}
		}

		return -1;
	}

	//Comparator Interface
	public int compare(Object o1, Object o2){
		gesture t1 = (gesture)o1;
		gesture t2 = (gesture)o2;

		return (t1.getName()).compareTo(t2.getName());
	}
	public int compareTo(Object o1){
		gesture g1 = (gesture)o1;

		return name.compareTo(g1.getName());
	}

	public boolean equals(Object o1){
		return super.equals(o1);
	}
}

//A JFrame entension used to display gestures (mostly debugging purposes)
class testGrid extends JFrame{

	Rectangle2D.Double[] rects;
	Vector gest;
	JPanel pane;
	int scale = 300;

	public testGrid(Rectangle2D.Double[] rects, String name){
		super(name);
		this.rects = rects;
		pane = new JPanel();
		this.getContentPane().add(pane);
		this.setSize(500,500);
		this.setVisible(true);
		gest = new Vector();
	}

	public testGrid(Rectangle2D.Double[] rects, Vector gest, String name){
		super(name);
		this.rects = rects;
		this.gest = gest;
		pane = new JPanel();
		this.getContentPane().add(pane);
		this.setSize(500,500);
		this.setVisible(true);
	}

	public void paint(Graphics graphics){
		super.paint(graphics);
		Graphics2D g = (Graphics2D)pane.getGraphics();

		for(int i=0;i<rects.length;i++){
			Rectangle2D.Double r = rects[i];
			g.draw(new Rectangle2D.Double(r.getX()*scale, r.getY()*scale, r.getWidth()*scale, r.getHeight()*scale));
		}

		for(int i=0;i<gest.size();i++){
			g.setColor(Color.RED);
			Line2D.Double l = (Line2D.Double)gest.elementAt(i);
			Line2D.Double line = new Line2D.Double(l.getX1()*scale, l.getY1()*scale, l.getX2()*scale, l.getY2()*scale);
			g.draw(line);
		}

		//mark starting point
		Line2D.Double l = (Line2D.Double)gest.elementAt(0);
		Line2D.Double line = new Line2D.Double(l.getX1()*scale, l.getY1()*scale,l.getX1()*scale, l.getY1()*scale);
		g.setStroke(new BasicStroke(7));
		g.draw(line);
	}
}