package edu.gwu.blaise.version2.transmission;
import java.io.*;
import java.util.*;
import java.nio.*;

public class DualByteArray{
  private int numEntries;
  private byte[] parray;
  private ByteArrayOutputStream baos;
  private Vector v;
  
  /**
   * Create a DualByteArray object with more than two entries
   * @param p the properties object that is always first in a DualByteArray
   * @param numEntries the number of entries in this array (excluding the Properties object)
   */
  public DualByteArray(Properties p){
    numEntries=0;
    baos=new ByteArrayOutputStream();
    v=new Vector();
    parray=toByteArray(p);
  }
  
  public void add(byte[] b){
    numEntries++;
    v.add(b);
  }
  
  public byte[] getSizeNDualByteArray(){
    //insert the number of elements and the size of the properties array
    byte[] temp=toBytes(numEntries);
    baos.write(temp,0,temp.length);
    temp=toBytes(parray.length);
    baos.write(temp,0,temp.length);
    //go through once and insert the sizes of the other arrays
    byte[] temp2=null;
    for(int i=0;i<numEntries;i++){
      temp=(byte[])v.elementAt(i);
      temp2=toBytes(temp.length);
      baos.write(temp2,0,temp2.length);
    }
    //insert the properties byte array
    baos.write(parray,0,parray.length);
    //put in the rest of the arrays
    for(int i=0;i<numEntries;i++){
      temp=(byte[])v.elementAt(i);
      baos.write(temp,0,temp.length);
    }
    return baos.toByteArray();
  }
  
  //Convert a props instance to a byte array
  public static byte[] toByteArray(Properties props){
    try{
      ByteArrayOutputStream byteOutputStream=new ByteArrayOutputStream();
      //**May want a better name.
      props.store(byteOutputStream,"Test");
      return byteOutputStream.toByteArray();
    }catch(IOException e){
      e.printStackTrace();
      return null;
    }
  }

  //Convert a byte array into a props instance  
  public static Properties toProperties(byte[] bytes){
    try{
      ByteArrayInputStream byteInputStream=new ByteArrayInputStream(bytes);
      Properties props=new Properties();
      props.load(byteInputStream);
      return props;
    }catch(IOException e){
      e.printStackTrace();
      return null;
    }
  }
  
  //Convert a byte array to a string.
  public static String toString(byte[] bytes){
    String retStr=null;
    try{
      retStr=new String(bytes, 0, bytes.length,"UTF-8");
    }catch(UnsupportedEncodingException e){
      return null;
    }
    return retStr;
  }
  
  //Vice=versa.
  public static byte[] toBytes(String str){
    byte[] b=null;
    try{
      b= str.getBytes("UTF-8");
    }catch(UnsupportedEncodingException e){
      return null;
    }
    return b;
  }
  
  //Convert an int to 4-byte array.  Why didn't they put
  //this in Integer?   
  public static byte[] toBytes(int k){
    ByteBuffer byteBuf=ByteBuffer.allocate(4);
    //putInt does the dirty work
    byteBuf.putInt(k);
    byte[] bytes=byteBuf.array();
    return bytes;
  }
  
  //from 4-bytes to the int.
  public static int toInt(byte[] bytes){
    ByteBuffer byteBuf=ByteBuffer.allocate(4);
    byteBuf.put(bytes);
    //getInt does it for you.
    int k=byteBuf.getInt(0);
    return k;
  }
  
  //same thing, but with an offset into the byte array
  public static int toInt(byte[] bytes, int offset){
    ByteBuffer byteBuf=ByteBuffer.allocate(4);
    byteBuf.put(bytes, offset, 4);
    int k=byteBuf.getInt(0);
    return k;
  }
  
  //returns the number of entries in this array, not counting the properties
  public static int getNumEntries(byte[] bytes){
    return toInt(bytes,0);    
  }
  
  //Make the dual array with each array's size encoded in 4-bytes.
  //Thus, the first four bytes tell you the length of the first
  //array, the second four bytes tell you the length of the second array
  //The arrays themselves follow the first 8 bytes.  
  public static byte[] makeDualArray(byte[] bytes1, byte[] bytes2){
    //Allocate the space.
    ByteBuffer byteBuf=ByteBuffer.allocate(12 + bytes1.length + bytes2.length);
    
    //store the number of entries that will be in this array, minus one for the props file
    byte[] numEntriesBytes=toBytes(1);
    byteBuf.put(numEntriesBytes);
    //Get the sizes and put them in the array.
    byte[] size1Bytes=toBytes(bytes1.length);
    byteBuf.put(size1Bytes);
    byte[] size2Bytes=toBytes(bytes2.length);
    byteBuf.put(size2Bytes);

    //Now put the arrays.  The ByteBuffer itself advances the cursor
    byteBuf.put(bytes1);
    byteBuf.put(bytes2);
    
    return byteBuf.array();
  }
  
  //Extract the first array (the props).
  public static byte[] extractFirst(byte[] bytes){
    return extractNth(bytes, 1);
  }
  
  //Extract the second.
  public static byte[] extractSecond(byte[] bytes){
    return extractNth(bytes, 2);
  }
  
  public static synchronized byte[] extractNth(byte[] bytes, int n){
    int num=getNumEntries(bytes);
    //get the lengths of all the ones leading up to this one    
    int len1=0;
    for(int i=1;i<n;i++){
      len1+=toInt(bytes,i*4);
    }
    int len2=toInt(bytes,(n*4));
    //So, the array to be extracted has indices from (((num+2)*4)+len1) ... (((num+2)*4)+len1+len2-1).
    byte[] nth=new byte[len2];
    for(int i=0; i<len2; i++){
      nth[i]=bytes[((num+2)*4)+len1+i];
    }
    return nth;
  }
  
  //For debugging:
  public static void printBytes(byte[] bytes){
    System.out.print("Bytes:");
    for(int i=0;i<bytes.length;i++){
      System.out.print(" "+bytes[i]);
    }
  }
  
  public static void main(String[] arg){
    //Test methods.

    //testIntByte();
    //testProps();
    //testString();
    testBoth();
    testWithThree();
  }
  
  public static void testWithThree(){
    System.out.println("Testing with three byte arrays.");
    //First, make some properties.
    Properties props=new Properties();
    props.put("prop1","111");
    props.put("prop2","222");
    props.put("prop3","333333");

    //Next, two strings (instead of jar files)
    String str1="Hello World!";
    String str2="Goodbye Cruel World!";
    String str3="Why isn't this working?";
    String str4="I just hate you so much!";
    
    //make the object and add the strings to it
    DualByteArray dba=new DualByteArray(props);
    byte[] str1bytes=toBytes(str1);
    byte[] str2bytes=toBytes(str2);
    byte[] str3bytes=toBytes(str3);
    byte[] str4bytes=toBytes(str4);
    dba.add(str1bytes);
    dba.add(str2bytes);
    dba.add(str3bytes);
    dba.add(str4bytes);
    //extract the byte array
    byte[] combinedBytes=dba.getSizeNDualByteArray();

    //extract the pieces and print them
    byte[] propsAsBytes=extractNth(combinedBytes,1);
    Properties props2=toProperties(propsAsBytes);
    Enumeration e=props2.propertyNames();
    while(e.hasMoreElements()){
      String propName=(String)e.nextElement();
      String propVal=props2.getProperty(propName);
      System.out.println("testWithThree(): propName="+propName+" propVal="+propVal);
    }

    byte[] str1AsBytes=extractNth(combinedBytes,2);
    String str1a=toString(str1AsBytes);
    System.out.println("testWithThree(): str1="+str1+" str1a="+str1a);

    byte[] str2AsBytes=extractNth(combinedBytes,3);
    String str2a=toString(str2AsBytes);
    System.out.println("testWithThree(): str2="+str2+" str2a="+str2a);

    byte[] str3AsBytes=extractNth(combinedBytes,4);
    String str3a=toString(str3AsBytes);
    System.out.println("testWithThree(): str3="+str3+" str3a="+str3a);

    byte[] str4AsBytes=extractNth(combinedBytes,5);
    String str4a=toString(str4AsBytes);
    System.out.println("testWithThree(): str4="+str4+" str4a="+str4a);
  }
  
  public static void testBoth(){
    //First, make some properties.
    Properties props=new Properties();
    props.put("prop1","111");
    props.put("prop2","222");
    props.put("prop3","333333");
    //Convert to bytes.
    byte[] propsAsBytes=toByteArray(props);

    //Next, a string (instead of a jar file)
    String str="Hello World!";
    byte[] stringAsBytes=toBytes(str);
    
    //combine
    byte[] combinedBytes=makeDualArray(propsAsBytes, stringAsBytes);
    
    //Now see if it works:
    //Now see if it works:
    byte[] propsAsBytes2=extractFirst(combinedBytes);
    Properties props2=toProperties(propsAsBytes2);
    Enumeration e=props2.propertyNames();
    while(e.hasMoreElements()){
      String propName=(String)e.nextElement();
      String propVal=props2.getProperty(propName);
      System.out.println("testBoth(): propName="+propName+" propVal="+propVal);
    }
    byte[] stringAsBytes2=extractSecond(combinedBytes);
    String str2=toString(stringAsBytes2);
    System.out.println("testBoth: str="+str+" str2="+str2);
  }
    
  public static void testString(){
    String str="Hello World!";
    byte[] stringAsBytes=toBytes(str);
    String str2=toString(stringAsBytes);
    System.out.println("testString(): str="+str+" str2="+str2);
  }
  
  public static void testProps(){
    //First, make some properties.
    Properties props=new Properties();
    props.put("prop1","111");
    props.put("prop2","222");
    props.put("prop3","333333");
    //Convert to bytes.
    byte[] propsAsBytes=toByteArray(props);
    
    //now reverse process
    Properties props2=toProperties(propsAsBytes);
    Enumeration e=props2.propertyNames();
    while(e.hasMoreElements()){
      String propName=(String)e.nextElement();
      String propVal=props2.getProperty(propName);
      System.out.println("propName="+propName+" propVal="+propVal);
    }
  }
  
  public static void testIntByte(){
    int m=5;
    byte[] bytes=toBytes(m);
    int n=toInt(bytes);
    System.out.println("testIntByte: m="+m+" n="+n);
    m=41234;
    bytes=toBytes(m);
    n=toInt(bytes);
    System.out.println("testIntByte: m="+m+" n="+n);
  }
}
