package edu.gwu.blaise.version2.transmission;
/**
 * <p>Used for in the second tier of the Blaise Backend.  It encapulates all of the actual data that
 * is returned to requesters.  This information includes status, response format, and the actual
 * response itself.</p>
 *
 * @author Michael Gaiman
 * @version 2 1/25/2004
 */

public class ResultWriter {
    
    private String status=null; //the status of this result, Usually either SUCCESS or FAILURE
    private String format=null; //the format that the body is in.  Used by the client encapsulator to
                                //align body data with names.
    private StringBuffer body=new StringBuffer("");//the actual body of the status.
    
    //constructors
    public ResultWriter(){}
    
    /**
     * assign initial values.
     * @param status <code>true</code> for success, <code>false</code> for failure.
     * @param format see {@link #setFormat} for the format of format.
     * @param body the message to set
     */
    public ResultWriter(boolean status,String format,String body){
        this.status=(status)?TransmissionConstants.SUCCESS:TransmissionConstants.FAILURE;
        this.format=format;
        this.body.append(body);
    }
    
    /**
     * clears all data from this instance.
     */
    public void reset(){
        status=null;
        format=null;
        body=new StringBuffer();
    }
    
    /**
     * sets the status of this result.  To guarantee data integrity, if the status has already been set, 
     * this will erase the call {@link #reset} before setting the new status.
     * @param newStatus <code>true</code> for success, <code>false</code> for failure.
     * @return <code>boolean</code> representing whether this result was reset (<code>true</code> 
     * or not (<code>false</code>).
     */ 
    public boolean setStatus(boolean newStatus){
        boolean isReset=false;
        
        //check to see if a reset is needed.
        if(status!=null){
            isReset=true;
            reset();
        }
        
        status=(newStatus)?TransmissionConstants.SUCCESS:TransmissionConstants.FAILURE;
        return isReset;
    }
    
    /**
     * returns the current status of this result.
     * @return <code>boolean</code> <code>true</code> if status is positive, <code>false</code> if there was
     * an error (status was negative).
     */
    public boolean getStatus(){
        return (status==TransmissionConstants.SUCCESS)?true:false;
    }
    
    /**
     * sets the response body format of this result.
     * The format should list one complete iteration of the body.
     * So, for example, if this body was the result of a university table show, it would look like this:
     * <p><code>universityID:name:contactInfo|</code></p><p> where ":" should actually be
     * {@link TransmissionConstants#COL_SEPARATOR} and "|" should actually be 
     * {@link TransmissionConstants#ROW_SEPARATOR}.</p>
     * <p>Here is another example, this time for IDs:  <p><code>universityID,</code></p>
     * <p>where "," should actually be {@link TransmissionConstants#ID_SEPARATOR}.</p>
     * <p>format maybe null if a pure message is passed as the body.  DO NOT use \n in your format string.</p>
     * 
     * @param newFormat is the format string.
     * @return <code>boolean</code> representing whether this result was reset (<code>true</code> 
     * or not (<code>false</code>).
     */
    public boolean setFormat(String newFormat){
        boolean isReset=false;
        
        //check to see if a reset is needed.
        if(format!=null){
            isReset=true;
            reset();
        }
        
        format=newFormat;
        return isReset;
    }
    
    /**
     * add something to the body.
     *
     * @param s what should be appended.
     */
    public void append(String s){
        body.append(s);
    }
    
    /**
     * @return The current body of this ResultWriter in String form.
     */
    public String getBody(){
        return body.toString();
    }

    /**
     * converts this result to a String that is appropriate for transmission.  Will throw an
     * {@link IllegalStateException} if called before status is set.
     * @return a String representation of this result that is appropriate for transmission.
     */ 
    public String toString(){
        if(status==null){
            throw new IllegalStateException("Result Status must be set before conversion to String.");
        }
        return status+
               TransmissionConstants.BREAK+
               format+
               TransmissionConstants.BREAK+
               body.toString();
    }
}
