package edu.gwu.blaise.version2.transmission;
import java.net.*;
import java.io.*;
import java.util.*;
import edu.gwu.blaise.version2.documents.*;
/**
 * SubmitQuery performs the acual communication with the server.  It puts in a request
 * and returns the results in String or byte array form.  This class is totally generic, and thus
 * contains no settings that could change from client to client (editor, grader, admin tool, etc).
 * @author Michael Gaiman 
 * @author Jared Kiraly
 * @version 2 6/4/2004
 */
public class SubmitQuery {
    private static final boolean debug=false;
    private Properties arguments=null;
    private String forwardingAddress=null;

    /**
     * Constructs a new SubmitQuery with the arguments provided.
     * @param arguments A Properties object containing all parameter, argument
     * pairs to be transmitted to the server.
     * @param forwardAddress The URL of the text file which has the address
     * of the server to make the query of.
     */
    public SubmitQuery(Properties arguments, String forwardingAddress){
        this.arguments=arguments;
        this.forwardingAddress=forwardingAddress;
    }

    /**
     * Performs the actual querying of the server, after getting the address via the
     * {@link #getServerAddress} method.
     * @return A String containing the response returned.
     * @throws IOException if either a network, or a data error happens.
     */
    public String submit() throws IOException{
        if(debug){ System.out.println("SubmitQuery.submit, arguments:"+arguments);}
        String serverAddress=getServerAddress();
        URL serverURL=new URL(serverAddress);

        HttpURLConnection connection=(HttpURLConnection)serverURL.openConnection();
        connection.setDoOutput(true);
        connection.setDoInput(true);
        connection.setRequestProperty("Content-type", "application/x-www-form-urlencoded");
        connection.setRequestMethod("POST");

        //do the output of the parameters/arguments
        DataOutputStream out =  new DataOutputStream(connection.getOutputStream());
        StringBuffer sb = new StringBuffer();
        Enumeration e=arguments.propertyNames();
        String s;
        boolean incAnd=false;
        while(e.hasMoreElements()){
            //first parameter should have no &, all others should.
            if(incAnd){
                sb.append("&");
            }else{
                incAnd=true;
            }
            s=(String)e.nextElement();
            sb.append(URLEncoder.encode(s,"UTF-8"));
            sb.append("=");
            sb.append(URLEncoder.encode(arguments.getProperty(s),"UTF-8"));
        }

        if(debug){System.out.println("SubmitQuery.submit, StringBuffer:"+sb);}

        out.writeBytes(sb.toString());
        out.flush();
        out.close();

        //now get the results
        BufferedReader br=new BufferedReader(new InputStreamReader(connection.getInputStream()));
        sb=new StringBuffer();
        s=br.readLine();
        while(s!=null){
            sb.append(s);
            sb.append("\n");
            s=br.readLine();
        }
        br.close();
        //now give back the results.
        return sb.toString().trim();
    }
    
    /**
     * Performs the actual querying of the server, after getting the address via the
     * {@link #getServerAddress} method.
     * @return A File containing the response returned.
     * @throws IOException if either a network, or a data error happens.
     */
    public byte[] submitBinary(Blaise_Paper bp) throws IOException{
        if(debug){ System.out.println("SubmitQuery.submitBinary, arguments:"+arguments);}
        String serverAddress=getServerAddress();
        URL serverURL=new URL(serverAddress);

        HttpURLConnection connection=(HttpURLConnection)serverURL.openConnection();
        connection.setDoOutput(true);
        connection.setDoInput(true);
        connection.setRequestProperty("Content-type", "application/octet-stream");
        connection.setRequestMethod("POST");

        byte[] props;
        byte[] paper=new byte[0];                
        
        if(bp!=null){
            paper=BlaiseFile.paperToByteArray(bp);
        }
        
        props=DualByteArray.toByteArray(arguments);
        
        byte[] both=DualByteArray.makeDualArray(props, paper);        
        
        //do the output of the parameters/arguments
        DataOutputStream out =  new DataOutputStream(connection.getOutputStream());

        out.write(both);
        out.flush();
        out.close();

        //now get the results
        //don't do anything to the results, just pass them along in a different format
        //now give back the results.
        InputStream in=connection.getInputStream();
        byte[] ba=DocumentUtilities.getInputStreamAsByteArray(in);
        in.close();
        return ba;        
    }

    /**
     * Performs a query to discover the address of the server.
     * @return The address of the server in String form.
     * @throws IOException if a network error happens.
     */
    public String getServerAddress() throws IOException{
        URL url=new URL(forwardingAddress);
        HttpURLConnection connection=(HttpURLConnection)url.openConnection();
        BufferedReader br=new BufferedReader(new InputStreamReader(connection.getInputStream()));
        StringBuffer sb=new StringBuffer();
        String s=br.readLine();
        br.close();
        return s;
    }

    public static void main(String[] agrv){
        Properties p=new Properties();
        p.setProperty("command","getPaperID");
        p.setProperty("version","1");
        p.setProperty("passCode","1");
        p.setProperty("graderID","1");
        SubmitQuery sq=new SubmitQuery(p,"http://blaise.gwu.edu/graderForward.txt");
        System.out.println("getPaperID results:");
        try{
            System.out.println(sq.submit());
        }catch(IOException e){
            System.out.println(e);
        }
    }
}
