package edu.gwu.blaise.version2.transmission;
import java.util.*;
/**
 * Takes a query result String from a getXXXQuery and parses it into a form that can
 * be read easily.
 * @author Michael Gaiman
 * @version 2 1/25/2004
 */
public class TableResultReader extends ResultReader{
    private static boolean debug=false;
    private Vector table=null;
    private Properties currentRow=null;
    private Vector columnNames=null;
    private int index=0;

    public TableResultReader(String result){
        super(result);
    }

    /**
     * Parses the contents of the result into status, format, and body field.
     * If the format contains a row and column spec, then the body is parsed into table form.
     * @throws ParseResultException if the result is not in the proper form.
     */
    public void parse() throws ParseResultException{
        //parse the basic stuff first.
        super.parse();
        //now try to parse the body into the table.
        if(!format.equals("") && status==true){
            //first parse the format to find out the column names.
            columnNames=new Vector();
            String[] fRows=format.split(TransmissionConstants.ROW_SEPARATOR_RX);
            if(fRows.length!=1){
                throw new ParseResultException("Invalid format field in result (no row separator).");
            }

            String[] fCols=fRows[0].split(TransmissionConstants.COL_SEPARATOR_RX);
            if(fCols[0].equals(fRows[0])){//no regex matches.
                throw new ParseResultException("Invalid format field in result (no column separators).");
            }

            for(int i=0;i<fCols.length;i++){
                columnNames.add(fCols[i]);
            }
            if(debug)System.out.println("columnNames:"+columnNames);

            //now we have the column names, let's parse the body
            table=new Vector();
            String s;
            Properties col;
            int formatIndex;
            if(!body.equals("")){
            String[] rows=body.split(TransmissionConstants.ROW_SEPARATOR_RX);
                for(int rowIndex=0;rowIndex<rows.length;rowIndex++){
                    s=rows[rowIndex];
                    if(debug)System.out.println("column:"+s);
                    String[] cells=s.split(TransmissionConstants.COL_SEPARATOR_RX);
                    col=new Properties();
                    for(formatIndex=0;formatIndex<cells.length;formatIndex++){
                        col.setProperty((String)columnNames.get(formatIndex),cells[formatIndex]);
                    }
                    if(debug)System.out.println("col (properties):"+col);
                    table.add(col);
                }
            }
        }
    }

    /**
     * If there is a next row in the table, the current row will be incremented to it.
     * The current row is initially before the table, so this method should be called at least
     * once, initially.
     * @return <code>true</code> if there is a next row, <code>false</code> otherwise.
     */
    public boolean next(){
        if(status==false){
            return false;
        }
        if((index+1)<=table.size()){
            currentRow=(Properties)table.get(index);
            index++;
            return true;
        }
        else{
            return false;
        }
    }

    /**
     * returns a Vector containing the names of all the columns.  This information was taken from
     * the format field, and as such will return <code>null</code> if the format field contained no
     * column names.
     */
    public Vector getColumnNames(){
        return columnNames;
    }

    /**
     * Retrieves the cell in the table from the current row and the column specified by the parameter.
     * @param columnName The name of the column from which to return the cell.
     * @return The cell contents in String form, or <code>null</code> if the columnName was invalid.
     */
    public String getCell(String columnName){
        if(currentRow==null){
            return null;
        }
        return currentRow.getProperty(columnName);
    }

    /**
     * Retrieves the cell in the table from the current row and the column specified by the parameter.
     * @param columnNumber The number of the column from which to return the cell.  (Column number start from 0.)
     * @return The cell contents in String form or <code>null</code> if the columnNumber was invalid.
     */
    public String getCell(int columnNumber){
        if(columnNames==null || columnNumber>=columnNames.size() || columnNumber<0 || currentRow==null){
            return null;
        }
        return currentRow.getProperty((String)columnNames.get(columnNumber));
    }

    /**
     * returns the number of columns.
     */
    public int getNumberOfColumns(){
        if(columnNames==null){
            return 0;
        }
        return columnNames.size();
    }

    /**
     * returns the number of rows.
     */
    public int getNumberOfRows(){
        if(table==null){
            return 0;
        }
        return table.size();
    }

    public static void main(String[] argv){
        String s= "SUCCESS\nwriterID:::universityID:::studentIdentifier:::firstName:::lastName:::email:::userName:::userNameType:::creationDate|||\n1:::1:::lalala:::Mike:::Gaiman:::mgaiman@gwu.edu:::mgaiman:::standard:::2003-06-24 23:22:27.0|||3:::2:::1010:::Thomas:::Anderson:::neo@matrix.com:::neo:::standard:::1999-03-31 00:00:00.0|||1:::1:::lalala:::Mike:::Gaiman:::mgaiman@gwu.edu:::mgaiman:::standard:::2003-06-24 23:22:27.0|||4:::3:::0101:::Unknown:::Unknown:::trinity@matrix.com:::trinity:::standard:::2003-06-25 02:19:16.0";
        //String s="FAILURE\n\nSilly boy...";

        TableResultReader rr=new TableResultReader(s);
        try{
            rr.parse();
        }catch(ParseResultException e){
            System.out.println(e);
        }
        System.out.println("Status:"+rr.getStatus());
        System.out.println("Format:"+rr.getFormat());
        System.out.println("ColumnNames:"+rr.getColumnNames());
        System.out.println("Body:"+rr.getBody());

        int size=rr.getNumberOfColumns();
        Vector names=rr.getColumnNames();
        while(rr.next()){
            for(int i=0;i<size;i++){
                System.out.println("Column "+i+": "+rr.getCell(i));
            }

            for(int i=0;i<names.size();i++){
                System.out.println("Column "+names.get(i)+": "+rr.getCell((String)names.get(i)));
            }
        }
    }
}
