//LOC:251
package edu.gwu.blaise.version2.util;

import java.awt.*;
import java.util.*;

/**
 * A CorrectionSet is a grading mark that reflects some grammatical error (spelling, grammar, capitalization)
 *
 * @author Sean T. Hanlon
 */
public class CorrectionSet{

	String type;	//the type of correction
	Color color;	//the color to show it

	Vector start;	//vector of start indexes
	Vector end;		//vector of end indexes

	/**
	 * Constructor
	 *
	 * @param t The type of correction set this is
	 */
	public CorrectionSet(String t){
		this(Color.RED, t);
	}

	/**
	 * Constructor
	 *
	 * @param c The color of this correction set
	 * @param t The type of this correction set
	 */
	public CorrectionSet(Color c, String t){
		color = c;
		type = t;
		start = new Vector();
		end = new Vector();
	}

	/**
	 * Constructor
	 *
	 * @param i The first start index of this correction set
	 * @param j The first end index of this correction set
	 * @param c The color of this correction set
	 * @param t The type of this correction set
	 */
	public CorrectionSet(int i, int j, Color c, String t){
		start = new Vector();
		end = new Vector();
		start.add(new Integer(i));
		end.add(new Integer(j));
		color = c;
		type = t;
	}

	/**
	 * Constructor
	 *
	 * @param t The type of this correction set
	 * @param c The color of this correction set
	 * @param starts The vector of starts
	 * @param end The vector of ends
	 */
	public CorrectionSet(String t, Color c, Vector starts, Vector ends){
		type = t;
		color = c;
		start = starts;
		end = ends;
	}

	/**
	 * Adds another start-end pair
	 *
	 * @param i The start index
	 * @param j The end index
	 */
	public void add(int i, int j){
		start.add(new Integer(i));
		end.add(new Integer(j));
	}

	/**
	 * Returns the starts indices of the correction set
	 *
	 * @return <code>Vector</code> of the starts
	 */
	public Vector getStarts(){
		return start;
	}

	/**
	 * Set the start index vector
	 *
	 * @param s Vector of the starts
	 */
	public void setStart(Vector s){
		start = s;
	}

	/**
	 * Returns the ends indices of the correction set
	 *
	 * @return <code>Vector</code> of the ends
	 */
	public Vector getEnds(){
		return end;
	}

	/**
	 * Set the end index vector
	 *
	 * @param e Vector of the ends
	 */
	public void setEnd(Vector e){
		end = e;
	}

	/**
	 * Set the color
	 *
	 * @param c The color
	 */
	public void setColor(Color c){
		color = c;
	}

	/**
	 * Get the Color
	 *
	 * @return <code>Color</code> of the correction set
	 */
	public Color getColor(){
		return color;
	}

	/**
	 * Set the type
	 *
	 * @param t The type to set this correction set
	 */
	public void setType(String t){
		type = t;
	}

	/**
	 * Get the type of this correction set
	 *
	 * @return <code>String</code> of the type
	 */
	public String getType(){
		return type;
	}

	/**
	 * Gets a string version of the color
	 *
	 * @return <code>String</code> in format RED|GREEN|BLUE
	 */
	public String getColorString(){
		return (color.getRed()+"|"+color.getGreen()+"|"+color.getBlue());
	}

	/**
	 * Gets a string version of the indices
	 *
	 * @return <code>String</code> in format S1,E1|S2,E2|S3,E3|...
	 */
	public String getLocationString(){
		String output = "";

		for(int i=0;i<start.size();i++){
			output += ((Integer)start.elementAt(i)).intValue();
			output += ",";
			output += ((Integer)end.elementAt(i)).intValue();

			if(i != start.size() - 1){
				output += "|";
			}
		}

		return output;
	}

	/**
	 * Increments all starts-end pairs after the passed value
	 *
	 * @param after what index to increment after
	 */
	public void increment(int after){

		Vector newStart = new Vector();
		Vector newEnd = new Vector();

		for(int i=0;i<start.size();i++){
			int s = ((Integer)start.elementAt(i)).intValue();
			int e = ((Integer)end.elementAt(i)).intValue();

			if(s >= after){
				newStart.add(new Integer(s+1));
				newEnd.add(new Integer(e+1));
			}
			else{
				newStart.add(new Integer(s));
				newEnd.add(new Integer(e));
			}
		}

		start = newStart;
		end = newEnd;
	}

	/**
	 * Generates the stored content of this correction set.
	 *
	 * @return <code>String</code> in format type@RED@GREEN@BLUE@S1*E1*S2*E2*...
	 */
	public String generateStoredContent(){
		String output = "";

		output += type;
		output += "@";
		output += color.getRed();
		output += "@";
		output += color.getGreen();
		output += "@";
		output += color.getBlue();
		output += "@";

		for(int i=0;i<start.size();i++){
			output += (Integer)start.elementAt(i);
			output += "*";
			output += (Integer)end.elementAt(i);
			if(i != start.size() - 1){
				output += "*";
			}
		}

		return output;
	}

	/**
	 * Returns the number of corrections
	 *
	 * @return <code>int</code> the amount of corrections
	 */
	public int size(){
		return start.size();
	}

	/**
	 * Prints correct set (primarily for debugging)
	 */
	public void print(){
		System.out.println("-----Correction Set------");
		System.out.println("Type: "+type);
		System.out.println("Color: "+color);

		for(int i=0;i<start.size();i++){
			System.out.println("start: "+start.elementAt(i)+" --> end: "+end.elementAt(i));
		}
		System.out.println("-----End Correction Set------");
	}

	public void removeLast(){

		if(start.size() > 0 && end.size() > 0){
			start.removeElement(start.lastElement());
			end.removeElement(end.lastElement());
		}
	}
}