package edu.gwu.blaise.version2.util;

// File: New_io.java
//
// Author: Rahul Simha
// Created: Sept 15, 1998
// Modified: October 7, 1998 (added read_long, formatting).
//
// A package of useful I/O functions.
// See the main function below for examples of use.

import java.io.*;
import java.util.*;
import java.text.*;

public class New_io {

  // Class data for screen input.
  private static InputStreamReader isr_screen
    = new InputStreamReader (System.in);
  private static LineNumberReader lr_screen
    = new LineNumberReader (isr_screen);

  // Class methods:

  // Put out a prompt and read in a string.
  public static String read_string (String prompt)
  {
    System.out.print (prompt);
    try {
      String s = lr_screen.readLine ();
      return s;
    }
    catch (IOException e) {
      System.out.println ("ERR: New_io::read_string: exception\n");
      System.exit (1);
      return "";
    }
  }

  // Put out a prompt and read in an int.
  public static int read_int (String prompt)
  {
    String s = read_string (prompt);
    while (true) {
      try {
        int i = Integer.parseInt (s);
        return i;
      }
      catch (NumberFormatException e) {
        s = read_string (prompt);
      }
    }
  }

  // Put out a prompt and read in a double.
  public static double read_double (String prompt)
  {
    String s = read_string (prompt);
    while (true) {
      try {
	Double d = Double.valueOf (s);
	return d.doubleValue();
      }
      catch (NumberFormatException e) {
        s = read_string (prompt);
      }
    }
  }


  // Put out a prompt and read in a long.
  public static long read_long (String prompt)
  {
    String s = read_string (prompt);
    while (true) {
      try {
	Long L = Long.valueOf (s);
	return L.longValue();
      }
      catch (NumberFormatException e) {
        s = read_string (prompt);
      }
    }
  }


  // Given a property-value combination (e.g., "radius=0.9")
  // and the property name ("radius"), return the value ("0.9")
  // as a string.
  public static String read_property_string (String in_string, String property)
  {
    StringTokenizer st = new StringTokenizer (in_string);

    String first_part = st.nextToken ("=");
    first_part = first_part.trim ();

    if (!first_part.equals (property)){
      System.out.println ("ERR:New_io.read_property: " +
                          "Improper string: " + in_string);
      System.exit (0);
    }
    String second_part = st.nextToken (" =\t\n\r");
    second_part = second_part.trim();

    return second_part;
  }

  // Given a property-value combination (e.g., "radius=0.9")
  // and the property name ("radius"), return the value (0.9).
  public static double read_property_double (String in_string, String property)
  {
    String s = read_property_string (in_string, property);
    try {
      Double d = Double.valueOf (s);
      return d.doubleValue();
    }
    catch (NumberFormatException e) {
      System.out.println ("ERR:New_io.read_property: " +
                          "Second part not a number: " + s);
      System.exit (0);
    }
    return 0;
  }


  /////////////////////////////////////////////////////////
  // The next section has instance data and methods
  // for file I/O:

  // Instance data:
  private String filename;
  private int operation = 0; // 1=read, 2=write, 3=append.
  private FileReader fr;
  private LineNumberReader lr_file;
  private FileWriter fw;
  private PrintWriter pw;

  // Constructor.
  //   Takes in a filename and an action: "read",
  //   "write" or "append"
  public New_io (String fname, String action)
  {
    filename = fname;
    if (action.equalsIgnoreCase("read")) {
      try {
	fr = new FileReader (filename);
      }
      catch (FileNotFoundException e) {
        System.out.println ("ERR: New_io.New_io: file " +
                            filename + "not found\n");
        System.exit(0);
      }
      lr_file = new LineNumberReader (fr);
      operation = 1;
    }
    else if (action.equalsIgnoreCase("write")) {
      try {
	fw = new FileWriter (filename);
      }
      catch (IOException e) {
        System.out.println ("ERR: New_io.New_io: file " +
                            filename + "cannot be opened\n");
        System.exit(0);
      }
      pw = new PrintWriter (fw, true);
      operation = 2;
    }
    else if (action.equalsIgnoreCase("append")) {
      try {
	fw = new FileWriter (filename, true);
      }
      catch (IOException e) {
        System.out.println ("ERR: New_io.New_io: file " +
                            filename + "cannot be opened\n");
        System.exit(0);
      }
      pw = new PrintWriter (fw, true);
      operation = 3;
    }
  }

  // Close the file.
  public void close ()
  {
    try {
      if (operation == 1)
	lr_file.close();
      else
	pw.close();
    }
    catch (IOException e) {
      System.out.println ("ERR: New_io.close(): cannot close\n");
    }
  }

  // Read a line from the file.
  public String read_string ()
  {
    try {
      String s = lr_file.readLine ();
      return s;
    }
    catch (IOException e) {
      System.out.println ("ERR: New_io.read_string() exception\n");
      System.exit (1);
      return "";
    }
  }

  // Read an integer all by itself on a line in the file.
  public int read_int ()
  {
    String s = this.read_string ();
    try {
      int i = Integer.parseInt (s);
      return i;
    }
    catch (NumberFormatException e) {
      System.out.println ("ERR: New_io.read_int() : " +
                          "NumberFormat exception: " + s + "\n");
      System.exit (1);
      return 0;
    }
  }

  // Read a double all by itself on a line in the file.
  public double read_double ()
  {
    String s = this.read_string ();
    try {
      Double d = Double.valueOf (s);
      return d.doubleValue();
    }
    catch (NumberFormatException e) {
      System.out.println ("ERR: New_io::read_double file: " +
                          "NumberFormat exception:" + s + "\n");
      System.exit (1);
      return 0;
    }
  }

  // Write a string to the file.
  public void print (String s)
  {
    pw.print (s);
  }

  // Write a string to the file with an end-of-line.
  public void println (String s)
  {
    pw.println (s);
  }

  /////////////////////////////////////////////////////////
  // Formatting output:

  // Pad blanks as required (for both int's and double's).
  static String pad_blanks (String s, int target_len, boolean is_int)
  {
    int slen = s.length();

    // No decimal point to worry about with int's.
    if (is_int) {
      if (slen >= target_len)
	return s;
      // Otherwise we pad from front
      for (int i=1; i <= target_len-slen; i++)
	s = ' ' + s;
      return s;
    }

    // Otherwise check up to decimal point
    int i = s.indexOf ('.');
    if ((i < 0) || (i >= slen)) {
      System.out.println ("ERR: pad_blanks: no decimal point");
      System.exit(1);
    }
    if (i+1 >= target_len)
      return s;

    // Otherwise, pad with blanks
    for (int j=1; j <=target_len-(i+1); j++)
      s = ' ' + s;
    return s;
  }

  // Format integers.
  public static String format (long i, int width)
  {
    DecimalFormat df_i = new DecimalFormat ("#");
    String s_i = df_i.format (i);
    s_i = pad_blanks (s_i, width, true);
    return s_i;
  }

  // Format double's.
  public static String format (double d, int before_dec_pt,
                                    int after_dec_pt)
  {
    // Use '#' symbols before decimal point.
    String s = "";
    for (int i=1; i<before_dec_pt; i++)
      s = s + '#';

    // Force a zero for numbers less than 1.0
    s = s + "0.";

    // Include zeroes as specified by width.
    for (int i=1; i<=after_dec_pt; i++)
      s = s + '0';

    DecimalFormat df_d = new DecimalFormat (s);
    String s_d = df_d.format (d);

    // Pad blanks.
    s_d = pad_blanks (s_d, before_dec_pt, false);
    return s_d;
  }

  /////////////////////////////////////////////////////////
  // Testing:

  public static void main (String[] argv)
  {
    // Test screen input:
    String s = New_io.read_string ("Enter a string: ");
    int i = New_io.read_int ("Enter an int: ");
    double d = New_io.read_double ("Enter a double: ");
    System.out.println ("String s = " + s + ", int i = " + i +
                        ", double d = " + d);

    // Open a file called testdata for reading lines:
    // Suppose testdata consists of the lines:
    //   Circle:
    //     center.x=50
    //     center.y=100
    //     radius=10
    New_io nio = new New_io ("testdata", "read");
    boolean over = false;
    int n_lines = 0;
    do {
      String input_line = nio.read_string ();
      if (input_line != null)
	n_lines ++;
      else
	over = true;
    } while (! over);
    nio.close();
    System.out.println ("Number of lines in testdata: " + n_lines);

    // Read properties from testdata:
    nio = new New_io ("testdata", "read");
    String input_line = nio.read_string ();
    double x=0, y=0, r=0;
    if (input_line.equals("Circle:")) {
      input_line = nio.read_string();
      x = read_property_double (input_line, "center.x");
      input_line = nio.read_string();
      y = read_property_double (input_line, "center.y");
      input_line = nio.read_string();
      r = read_property_double (input_line, "radius");
      System.out.println ("c.x=" + x + ", c.y=" + y + ", r=" + r);
    }
    nio.close ();

    // Write to the same file:
    nio = new New_io ("testdata", "append");
    nio.println ("c.x=" + x + ", c.y=" + y + ", r=" + r);
    nio.close ();

    // Test string property:
    s = New_io.read_property_string ("name= Smith ", "name");
    System.out.println ("Name = " + s);
  }

} // End of class "new_io"
