package edu.gwu.blaise.version2.util;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import edu.gwu.blaise.version2.util.Constants;

/**
 * Handles the opening of Windows so that the previous closing state is remembered
 * and restored.
 * It is designed to be handled statically.
 * <p>To use it make a Window (or subclass) like normal:
 * <code>JFrame f=new JFrame("Window Opener Test");</code>
 * Then call <code>WindowOpener.open(f,"Test");</code> (replacing "Test" with something more useful).
 * </p>
 *
 *
 * @author Michael Gaiman mgaiman@gwu.edu
 * @version 1.1 3/12/04
 * @history 1.0 Initial Version
 *          1.1 3/12/04 Blaise integration
 */
public class WindowOpener implements WindowListener{

    /** Extension of the X location coordinate.  Used when serializing the properties.*/
    public static final String LOCATION_X=".x";
    /** Extension of the Y location coordinate.  Used when serializing the properties.*/
    public static final String LOCATION_Y=".y";
    /** Extension for the height of the window.  Used when serializing the properties.*/
    public static final String HEIGHT=".height";
    /** Extension for the width of the window.  Used when serializing the properties.*/
    public static final String WIDTH=".width";
    /** What is written out to the properties file as a comment*/
    private static final String HEADER="WindowOpener window Properties.  Delete this file to reset window size and positions.";

    private static String filename=Constants.getWindowOpenerFile();
    private static Hashtable names=new Hashtable();
    private static Properties props=new Properties();
    private static final boolean debug=true;
    
    //Static initializer.
    static{
        try{
            loadProperties(filename);
        }
        catch(IOException e){
        //could not open it for whatever reason.  
        //Don't be too worried, we'll probably recreate it again during saving.
            if(debug)System.out.println(e);
        }
        
        Runtime rt=Runtime.getRuntime();
        rt.addShutdownHook(new Thread(){
            public void run(){
                if(debug)System.out.println("In WindowOpener shutdown hook");
                //go names hashtable and ask each Window its settings.
                //All windows left in the hashtable, are on close (as we're closing now, and they
                //still in the hashtable.)
                
                try{
                    save();
                }catch(IOException e){
                    System.out.println("WindowOpener could not serialize changes.  (Error: "+e+")");
                }
            }
        });
    }
    
    /**
     * Causes the local filename to be set.  This will change the file the properties will be 
     * written to at the end, but will not load a new file.
     * @see loadProperties
     */
    public static void setFilename(String filename){
        WindowOpener.filename=filename;
    }
    public static String getFilename(){
        return filename;
    }
    
    /**
     * Sets WindowOpener to load, and eventually save to, a different properties file.
     */
    public static void loadProperties(String filename) throws IOException{
        if(debug)System.out.println("Loading properties...");
        WindowOpener.filename=filename;
        props.load(new BufferedInputStream(new FileInputStream(filename)));
    }
    
    /**
     * Causes the size and location to be set to the previous saved value (if one exists).
     * @param window The Actual Window (frame, dialog, etc) to be opened.
     * @param name a semi-unique identifier for this window.  It is was is used when serializing the settings, 
     * so if you name multiple windows the same, when you reopen them next time, they will all have the settings
     * of the last one closed.
     */
    public static void open(Window window,String name){
        int x=-1,y=-1,h=-1,w=-1;
        String s;
        //see if this thing has a past.
        s=props.getProperty(name+LOCATION_X);
        if(s!=null){
            x=Integer.parseInt(s);
        }
        s=props.getProperty(name+LOCATION_Y);
        if(s!=null){
            y=Integer.parseInt(s);
        }
        s=props.getProperty(name+HEIGHT);
        if(s!=null){
            h=Integer.parseInt(s);
        }
        s=props.getProperty(name+WIDTH);
        if(s!=null){
            w=Integer.parseInt(s);
        }
        
        //set its values
        if(x>=0 && y>=0){//valid location
            window.setLocation(x,y);
        }
        if(w>=0 && h>=0){//valid height and width
            window.setSize(w,h);
        }
        
        //register to be a windowListener.
        window.addWindowListener(new WindowOpener());
        
        //add it to the hashtable
        names.put(window,name);
        
        window.setVisible(true);
    }
    
    
    /**
     * Saves the settings of all Windows that have been opened with WindowOpener since
     * This JVM was instantiated.  This happens automatically when the program exits.
     */
    public static void save() throws IOException{
        //get all currently open window's settings. (added to all of the closed ones).
        Window w;
        for(Enumeration e=names.keys();e.hasMoreElements();){
            w=(Window)e.nextElement();
            recordSetting(w);
        }
        
        props.store(new BufferedOutputStream(new FileOutputStream(filename)),HEADER);
    }
    
    
    //WindowListener methods (only care about closing)
    public void windowActivated(WindowEvent e){}
    public void windowClosed(WindowEvent e){
        //remove the window from the names hashtable (to prevent memory leaks).
        if(debug)System.out.println("window closed.");
        names.remove(e.getWindow());
    }
    public void windowOpened(WindowEvent e){}
    public void windowIconified(WindowEvent e){}
    public void windowDeiconified(WindowEvent e){}
    public void windowDeactivated(WindowEvent e){}
    /**
     * Record information about the window's state, to be used when opening it next time.
     */
    public void windowClosing(WindowEvent e){
        if(debug)System.out.println("windowClosing event.");
        /*The windowevent has the object.  Look it up in the hashtable
         *for the String name for use as the base property.
         */
        Window w=e.getWindow();
        recordSetting(w);
    }
    
    /**
     * Given a window, it will add the settings to the properties.  This will not write the file.
     * @see save
     */
    private static void recordSetting(Window w){
        String name=(String)names.get(w);
        if(name==null){//don't have a name for it...which is odd.  Give up on this window.
            if(debug)System.out.println(
            "WindowOpener: Received a window closing event from a window who's name is unknown. ("+w+")");
            return;
        }
        props.setProperty(name+LOCATION_X,  Integer.toString(w.getX())       );
        props.setProperty(name+LOCATION_Y,  Integer.toString(w.getY())       );
        props.setProperty(name+WIDTH,       Integer.toString(w.getWidth())   );
        props.setProperty(name+HEIGHT,      Integer.toString(w.getHeight())  );
    }
    
    public static void main(String[] argv){
        JFrame f=new JFrame("Window Opener Test");
        f.setResizable(true);
        f.getContentPane().add(new JLabel("Window Opener Test"));
        f.setSize(400,400);
        f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        WindowOpener.open(f,"Test");
        JFrame f2=new JFrame("A smaller frame");
        f2.getContentPane().add(new JLabel("Another"));
        f2.setSize(100,200);
        f2.setLocation(300,300);
        WindowOpener.open(f2,"Another");
    }
}