package edu.gwu.blaise.version2.util.contextmenu;

import java.util.*;
import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
/**
 * <p>A utility class that allows multiple interested parties to build a combined
 * contextual (aka right click, or popup) menu.  Interested parties should implement
 * the ContextMenuListener interface, and register with the manager via the addContextMenuListener
 * method.</p>
 * <p>Please note: ContextMenuManager will consume mouse events for the component that are found
 * to trigger a contextual menu.</p>
 * <p>ContextMenuManager allows the listeners to specify which order they would like their items
 * displayed on the combined menu.  This is done via the order parameter.  All effort is made to
 * honor the request for a specific order, but if multiple listeners request the same order, then
 * following requested orders will be further by the factor of overlapping requests.
 * If order does not matter, choose AUTO_ORDER, which will assign order in a constant manner (provided
 * no new listeners are registered).</p>
 *
 * @author Michael Gaiman mgaiman@gwu.edu
 * @version 1.1 3/16/2004
 *
 * @history 1.0 3/16/2004 Initial version
 *          1.1 3/16/2004 moved into Blaise
 *              3/18/2004 Added Separator support
 */
public class ContextMenuManager extends MouseAdapter{
    
    /**
     * Context menu order doesn't matter.
     */
    public static final int AUTO_ORDER=-1;
    private static final Integer AUTO_ORDER_I=new Integer(AUTO_ORDER);
    /**
     * Should always be the last.  If multiple LAST listeners are added, the last one to be added
     * will be the true last.
     */
    public static final int LAST=-2;
    private static final Integer LAST_I=new Integer(LAST);
    /**
     * Should always be the first.  If multiple FIRST listeners are added, the first one to be added
     * will be the true first.
     */
    public static final int FIRST=0;
    private static final Integer FIRST_I=new Integer(FIRST);
    
    /**
     * As a work around for not having access to the JPopUpMenu to call the addSeparator()
     * method, one can return as one of the JMenuItems this JMenuItem.  It will be converted
     * to an addSeparator() call.
     */
    public static final JMenuItem Separator=new JMenuItem("__ContextMenuManager.SEPARATOR__");
    
    /**
     * Contains the relationships of Component to ContextMenuListeners.
     * Table actually maps Components to ArrayLists which contain ContextMenuManager.CMMListener objects.
     */
    private static Hashtable table=new Hashtable();
    
    
    /**
     * Registers a ContextMenuListener to contribute to the Contextual Menu of Component
     * listenTo.
     * @param order In what order should your menus appear in the context menu.  These can be AUTO_ORDER,
     * FIRST, LAST, or any number from 1 to Integer.MAX_VALUE.
     */
    public static void addContextMenuListener(ContextMenuListener listener, Component listenTo, int order){
        if(order<LAST){
            throw new IllegalArgumentException("Invalid order");
        }
        
        //get the arraylist
        ArrayList list=(ArrayList)table.get(listenTo);
        if(list==null){
            list=new ArrayList();
            table.put(listenTo,list);
            listenTo.addMouseListener(new ContextMenuManager());
        }
        
        //add to arraylist, and then sort
        CMMListener wrapper=new CMMListener(listener, order);
        list.add(wrapper);
        //put it in order by order int.
        Collections.sort(list);
    }
    
    /**
     * Unregister as a context menu listener.
     */
    public static void removeContextMenuListener(ContextMenuListener listener, Component listenTo, int order){
        CMMListener l=new CMMListener(listener,order);
        ArrayList list=(ArrayList)table.get(listenTo);
        if(list==null){
            return;//all removed already (or never added)
        }
        int index=list.indexOf(l);
        if(index==-1){//does not exist
            return;
        }
        list.remove(index);
        if(list.isEmpty()){
            table.remove(listenTo);//remove the table from the hashtable
        }
    }
    
    public void mousePressed(MouseEvent e){
        if (e.isPopupTrigger()){
            e.consume();//hog it.
            showPopup(e);
        }
    }
    public void mouseReleased(MouseEvent e){
        if (e.isPopupTrigger()){
            e.consume();//hog it.
            showPopup(e);
        }
    }
    
    /**
     * Contacts the ContextMenu listeners for this component, gets their menu items,
     * assembles them, and finally shows the menu.
     */
    private void showPopup(MouseEvent e){
        /*NOTE: makes some assumptions about the values of the constants.
         * LAST comes at the very first, followed by AUTO_ORDER, then FIRST, then numbers
         * specified.
         */
        Component comp=e.getComponent();
        ArrayList list=(ArrayList)table.get(comp);//list is assumed to be sorted.
        int autoEnd=-1,lastEnd=-1;
        int i,j,size=list.size(),order,actualOrder=0;
        int x=e.getX(),y=e.getY();
        ContextMenuEvent thisEvent=new ContextMenuEvent(comp,x,y);
        CMMListener cmml;
        JMenu menu=new JMenu();
        JMenuItem[] items;
        boolean doSep=true;
        
        for(i=0;i<size;++i){
            cmml=(CMMListener)list.get(i);
            order=cmml.getOrder();
            if(order==AUTO_ORDER){
                autoEnd=i;
            }
            else if(order==LAST){
                lastEnd=i;
            }
            else{//the rest are already in correct order... just add them.
                thisEvent.setOrder(order);
                thisEvent.setActualOrder(actualOrder);
                
                items=cmml.getListener().getContextMenuItems(thisEvent);
        
                doSep=(items.length>0);//show a separator only if this isn't empty
                
                for(j=0;j<items.length;++j){
                    if(items[j]==Separator){
                        menu.addSeparator();
                        //don't reflect the separator on the actual order
                        --actualOrder;//it will be added back again, and will end up being correct.
                    }
                    else{
                        menu.add(items[j]);
                    }
                }
                
                actualOrder+=items.length;
                
                //add a separator between them all.
                if(i<size-1 && doSep){//not last
                    menu.addSeparator();
                }
            }
        }
        //add in the AUTO_ORDER items (if any)
        if(autoEnd!=-1){//some exist
            menu.addSeparator();
            thisEvent.setOrder(AUTO_ORDER);
            //figure out where to start
            i=0;
            if(lastEnd!=-1){
                i=lastEnd+1;
            }
            for(;i<=autoEnd;++i){
                thisEvent.setActualOrder(actualOrder);
                cmml=(CMMListener)list.get(i);
                items=cmml.getListener().getContextMenuItems(thisEvent);
                doSep=(items.length>0);//show a separator only if this isn't empty
                for(j=0;j<items.length;++j){
                    if(items[j]==Separator){
                        menu.addSeparator();
                        //don't reflect the separator on the actual order
                        --actualOrder;//it will be added back again, and will end up being correct.
                    }
                    else{
                        menu.add(items[j]);
                    }
                }
                actualOrder+=items.length;
                if(i<autoEnd && doSep){
                    menu.addSeparator();
                }
            }
        }
        //add in the LAST items (if any)
        if(lastEnd!=-1){
            menu.addSeparator();
            thisEvent.setOrder(LAST);
            
            for(i=0;i<=lastEnd;++i){
                thisEvent.setActualOrder(actualOrder);
                cmml=(CMMListener)list.get(i);
                items=cmml.getListener().getContextMenuItems(thisEvent);
                doSep=(items.length>0);//show a separator only if this isn't empty
                for(j=0;j<items.length;++j){
                    if(items[j]==Separator){
                        menu.addSeparator();
                        //don't reflect the separator on the actual order
                        --actualOrder;//it will be added back again, and will end up being correct.
                    }
                    else{
                        menu.add(items[j]);
                    }
                }
                actualOrder+=items.length;
                if(i<lastEnd && doSep){
                    menu.addSeparator();
                }
            }
        }
        
        //make an actual popup menu, and show it.
        menu.getPopupMenu().show(comp,x,y);
    }
}

/**
 * A small wrapper class for a listener and its order
 */
class CMMListener implements Comparable{
    private ContextMenuListener listener;
    private int order;
    public CMMListener(ContextMenuListener l, int o){
        listener=l;
        order=o;
    }
    
    public ContextMenuListener getListener(){
        return listener;
    }
    
    public int getOrder(){
        return order;
    }
    
    /**
     * Note: breaks the "equal contract."
     */
    public int compareTo(Object o){
        return order-((CMMListener)o).order;
    }
    
    
    public boolean equals(Object o){
        if(o instanceof CMMListener){
            CMMListener other=(CMMListener)o;
            if(other.order==order && other.listener.equals(listener)){
                return true;
            }
        }
        return false;
    }
}