package edu.gwu.blaise.version2.util.find;
/* 
 * DocumentFind.java
 * 
 * Portions of DocumentFind use StringSearch-derived code.
 *
 * eaio: StringSearch - high-performance pattern matching algorithms in Java
 * Copyright (c) 2003, 2004 Johann Burkard (jb@eaio.com) http://eaio.com
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 */

import javax.swing.text.*;
/**
 * DocumentFind finds a pattern of text within a javax.swing.text.Document.
 * It uses the Boyer-Moore-Horspool-Raita algorithm.
 * 
 * Portions of the implementation have been modified from http://johannburkard.de/ (StringSearch).
 * 
 * While DocumentFind does not handle "replacing" text once it has been found, it has been designed to be
 * compatible with a changing Document. (As such, you can basically just instantiate one DocumentFind per 
 * Document.
 *
 * @author Michael Gaiman mgaiman@gwu.edu
 * @version 1.1 3/12/2004
 */
public class DocumentFind{
    /**The Document that is searched in*/
    private Document document;

    /**The current search pattern*/
    private char[] pattern;
    private String patternWithCase;
    private int pLength;
    
    /**
     * The shift table
     */
     private CharIntMap table;
     private boolean ignoreCase=true;
     
     
     private static final boolean debug=true;
     
    
    /**
     * Constructs a new DocumentFind.  Ignoring case by default.
     */
    public DocumentFind(Document d){
        setDocument(d);
    }
    
    public Document getDocument(){
        return document;
    }
    
    public void setDocument(Document d){
        document=d;
    }
    
    /**
     * Changes whether or not to ignore case when searching.
     * By default DocumentFind searches with case ignored.
     * It is faster to search without case ignored.
     */
    public void setIgnoreCase(boolean c){
        if(c!=ignoreCase){//a change...
            ignoreCase=c;
            if(patternWithCase!=null){//it is null if this is called before setPattern
                setPattern(patternWithCase);
            }
        }
    }
    
    public boolean isIgnoreCase(){
        return ignoreCase;
    }

    /**
     * Sets the pattern to search with as well as builds the shift table.
     */
    public void setPattern(String p){
        patternWithCase=p;
        if(ignoreCase){
            p=p.toLowerCase();
        }
        pattern=p.toCharArray();
        pLength=pattern.length;
        table = createCharIntMap(pattern, pLength);
        for (int i = 0; i < pLength - 1; ++i) {
            table.set(pattern[i], pLength - i - 1);
        }
    }
    
    /**
     * Returns a {@link CharIntMap} of the extent of the given pattern, using the
     * specified default value.
     * 
     * @param pattern the pattern
     * @param defaultValue the default value
     * @return a CharIntMap
     * @see CharIntMap#CharIntMap(int, char, int)
     */
    protected CharIntMap createCharIntMap(char[] pattern, int defaultValue) {
        char min = Character.MAX_VALUE;
        char max = Character.MIN_VALUE;
        for (int i = 0; i < pattern.length; i++) {
        max = max > pattern[i] ? max : pattern[i];
        min = min < pattern[i] ? min : pattern[i];
        }
        return new CharIntMap(max - min + 1, min, defaultValue);
    }
    
    /**
     * Finds the first occurence of pattern starting searching at offset.
     * It is advantageous to call setPattern and then use the simpler findNext(int,int) method
     * as this will save having to recompute the shift table with each call.
     * The is just a convience method--it calls setPattern and then the other find.
     * @see setPattern(String)
     * @see findNext(int,int)
     */
    public int findNext(String pattern, int offset, int end) throws BadLocationException{
        setPattern(pattern);
        return findNext(offset,end);
    }
    
    /**
     * Searches starting at offset, until the end of the document.
     * Uses {@link findNext(int,int)}.
     * @see findNext(int, int)
     */
    public int findNext(int offset) throws BadLocationException{
        return findNext(offset,document.getLength());
    }
    
    /**
     * Finds the first occurence of pattern starting searching at offset.
     * End must be greater than offset.
     * @returns the starting offset of the pattern in the text, or -1 if not found after offset.
     * @throws BadLocationException if you have an invalid initial offset or if there is a location problem
     * getting the text from the document.
     */
    public int findNext(int offset, int end) throws BadLocationException{
        
        if(offset>=document.getLength() || offset<0){
            throw new BadLocationException("Invalid starting offset for search",offset);
        }
        if(table==null){
            throw new IllegalStateException("Must set pattern before calling findNext.");
        }
        if(end>document.getLength() || end<=offset){
            throw new BadLocationException("Invalid end value for search",end);
        }
        
        Segment text=new Segment();
        int textLength=end-offset;
        if(debug)System.out.println("Getting text offset:"+offset+" doc length:"+textLength);
        document.getText(offset,textLength,text);
        if(debug){System.out.print("Got text:'");
            for(int i=text.offset;i<text.count+text.offset;++i){
                System.out.print(text.array[i]);
            }
            System.out.println("' text.count:"+text.count+" textLength:"+textLength+" text.offset:"+text.offset);
        }
        
        if(ignoreCase){
            int textStart=text.offset;
            int textEnd=text.count+text.offset;
            int i, j, k, mMinusOne;
            char last, first;

            i = pattern.length - 1;
            mMinusOne = i - 1;

            last = Character.toLowerCase(pattern[i]);
            first = Character.toLowerCase(pattern[0]);

            i += textStart;

            while (i < textEnd) {
                if (Character.toLowerCase(text.array[i]) == last) {
                    if (Character.toLowerCase(text.array[i - (pLength - 1)]) == first) {
                        k = i - 1;
                        j = mMinusOne;
                        while (k > -1 && j > -1 && Character.toLowerCase(text.array[k]) == Character.toLowerCase(pattern[j])) {
                            --k;
                            --j;
                        }
                        if (j == -1) {
                            //k+1 is the original, but (k-text.offset+offset+1) gets you into
                            //a usable result.
                            return k-text.offset+offset+1;
                        }
                    }
                }
                i += table.get(Character.toLowerCase(text.array[i]));
            }

            return -1;
        }
        else{
            int textStart=text.offset;
            int textEnd=text.count+text.offset;
            int i, j, k, mMinusOne;
            char last, first;

            i = pattern.length - 1;
            mMinusOne = i - 1;

            last = pattern[i];
            first = pattern[0];

            i += textStart;

            while (i < textEnd) {
                if (text.array[i] == last) {
                    if (text.array[i - (pLength - 1)] == first) {
                        k = i - 1;
                        j = mMinusOne;
                        while (k > -1 && j > -1 && text.array[k] == pattern[j]) {
                            --k;
                            --j;
                        }
                        if (j == -1) {
                            //k+1 is the original, but (k-text.offset+offset+1) gets you into
                            //a usable result.
                            return k-text.offset+offset+1;
                        }
                    }
                }
                i += table.get(text.array[i]);
            }

            return -1;
        }
        
    }

}