package edu.gwu.blaise.version2.util.find;

import javax.swing.*;
import javax.swing.text.*;
import java.awt.event.*;
import java.awt.*;
import java.beans.*;
import edu.gwu.blaise.version2.util.WindowOpener;

/**
 * A JFrame that has find and replace options.
 *
 * To show a FindFrame call: <code>FindFrame.showFind(myJTextComponent,true);</code> this will bring
 * up a FindFrame, or bring the current one to the front if one already exists.
 *
 * The actual finding is done by DocumentFind.
 *
 * @author Michael Gaiman
 * @version 1.1 3/12/2004
 */
 
public class FindFrame extends JFrame implements ActionListener, PropertyChangeListener{
    private JTextComponent component=null;
    private JTextField patternField;
    private JTextField replaceField;
    private JLabel foundLabel;
    private DocumentFind find;
    private String lastPattern="";
    private JButton replaceButton;
    private JButton replaceAllButton;
    
    private static final String REPLACE_ALL     ="Replace All";
    private static final String REPLACE         ="Replace";
    private static final String NEXT            ="Next";
    private static final int NUM_COLS           =25;
    private static final int PAD                =5;
    private static final String EMPTY_NOT_FOUND ="         ";
    private static final String NOT_FOUND       ="Not found";
    
    private static FindFrame existing=null;//only one FindFrame can exist at a time.  this is it.
    
    private static final boolean debug=true;
    
    /**
     * Constructor.  Can only be made from within the class
     */
    private FindFrame(JTextComponent c){
        super("Find");
        setJTextComponent(c);
        //create the GUI 
        //javax.swing.SwingUtilities.invokeLater(new Runnable(){
        //    public void run() {
                createGUI();
        //    }
        //});
    }
    
    /**
     * Shows a FindFrame.  If one exists, it brings makes it visible.
     * If it was already visible, it brings it to front.
     * WindowOpener is used to Show FindFrame.
     */
    public static FindFrame showFind(final JTextComponent c){
        if(existing==null){
            existing=new FindFrame(c);
            WindowOpener.open(existing,"FindFrame");
        }
        else{
            if(existing.getJTextComponent()!=c){
                existing.setJTextComponent(c);
            }
            if(existing.isVisible()){
                //existing.toFront();
            }
            else{
                existing.setVisible(true);
            }
        }
        
        existing.toFront();
        return existing;
    }
    
    
    public void setJTextComponent(JTextComponent c){
        if(component!=null){
            component.removePropertyChangeListener(this);//unregister with the old JTC
        }
        component=c;
        component.addPropertyChangeListener("document", this);//register with the new one.
        find=new DocumentFind(c.getDocument());
        
        //adjust the enabledness of the replace stuff based upon editablility of the new document
        if(replaceField!=null){
            replaceField.setEnabled(component.isEditable());
        }
        if(replaceButton!=null){
            replaceButton.setEnabled(component.isEditable());
        }
        if(replaceAllButton!=null){
            replaceAllButton.setEnabled(component.isEditable());
        }
    }
    
    public JTextComponent getJTextComponent(){
        return component;
    }
    
    /**
     * Called when the Document changes.
     */
    public void propertyChange(PropertyChangeEvent e){
        find.setDocument((Document)e.getNewValue());
    }
    
    private void createGUI(){
        Container container=getContentPane();
        container.setLayout(new BoxLayout(container,BoxLayout.Y_AXIS));
        container.add(Box.createVerticalStrut(PAD));
        container.add(Box.createVerticalGlue());
        
        //create the find row
        Box box=Box.createHorizontalBox();
        box.add(Box.createHorizontalStrut(PAD));
        box.add(Box.createHorizontalGlue());
        box.add(new JLabel("Find:"));
        box.add(Box.createHorizontalStrut(PAD));
        patternField=new JTextField(NUM_COLS);
        patternField.setMaximumSize(patternField.getPreferredSize());
        box.add(patternField);
        box.add(Box.createHorizontalStrut(PAD));
        container.add(box);
        
        container.add(Box.createVerticalGlue());
        //create replace row
        box=Box.createHorizontalBox();
        box.add(Box.createHorizontalStrut(PAD));
        box.add(Box.createHorizontalGlue());
        box.add(new JLabel("Replace:"));
        box.add(Box.createHorizontalStrut(PAD));
        replaceField=new JTextField(NUM_COLS);
        replaceField.setMaximumSize(replaceField.getPreferredSize());
        replaceField.setEnabled(component.isEditable());
        box.add(replaceField);
        box.add(Box.createHorizontalStrut(PAD));
        container.add(box);
        
        container.add(Box.createVerticalGlue());
        //create label row
        foundLabel=new JLabel(EMPTY_NOT_FOUND);
        container.add(foundLabel);
        
        container.add(Box.createVerticalGlue());
        //create button row
        box=Box.createHorizontalBox();
        box.add(Box.createHorizontalGlue());
        JButton b=new JButton(REPLACE_ALL);
        b.setEnabled(component.isEditable());
        b.addActionListener(this);
        box.add(b);
        replaceAllButton=b;
        b=new JButton(REPLACE);
        b.setEnabled(component.isEditable());
        b.addActionListener(this);
        box.add(b);
        replaceButton=b;
        box.add(Box.createHorizontalGlue());
        JButton nextButton=new JButton(NEXT);
        nextButton.addActionListener(this);
        box.add(nextButton);
        box.add(Box.createHorizontalGlue());
        container.add(box);
        container.add(Box.createVerticalGlue());
        container.add(Box.createVerticalStrut(PAD));
        
        //set frame attributes
        setSize(400,120);
        setResizable(false);
        getRootPane().setDefaultButton(nextButton);
    }
    
    /**
     * Perfoms a find next operation using the search term that is currently in the 
     * Find field.
     */
    public void doFindNext(){
        Caret caret=component.getCaret();
        if(debug)System.out.println("caret mark:"+caret.getMark()+" dot:"+caret.getDot());
        int offset=Math.max(caret.getMark(), caret.getDot());//start at end of selection (if one exists)
        int start=Math.min(caret.getMark(), caret.getDot());
        
        //make sure that the PatternField has something useful in it.
        String pattern=patternField.getText();
        if(pattern==null || pattern.equals("")){
            Toolkit.getDefaultToolkit().beep();
            setLabelFound(false);
            return;
        }
        try{
            int result;
            if(!pattern.equals(lastPattern)){//if the pattern has changed, notify find.
                find.setPattern(pattern);
                lastPattern=pattern;
            }
            if(offset>=component.getDocument().getLength()){//starting at the end.  Search from the beginning.
                result=find.findNext(0,offset);
            }
            else{
                result=find.findNext(offset);
            }
            
            if(debug){
                System.out.println("FF Result:"+result+" when offset:"+offset);
            }
            
            if(result>=0){//valid result.
                component.select(result,result+pattern.length());
                setLabelFound(true);
                return;
            }
            else{//not found.
                //try searching again from the beginning, stopping at the beginning of the selection (if any).
                if(start!=0){
                    result=find.findNext(0,start);
                    if(debug)System.out.println("FF (re-search from beginning) Result:"+result+" when start:"+start);
                }
                else{
                    result=-1;
                    if(debug)System.out.println("FF (re-search from beginning) selected at beginning--not searched");
                }
                
                if(result>=0){
                    component.select(result,result+pattern.length());
                    setLabelFound(true);
                    return;
                }
                else{
                    Toolkit.getDefaultToolkit().beep();
                    setLabelFound(false);
                    return;
                }
            }
        }catch(BadLocationException ex){
            System.out.println(ex);
            return;
        }
    }
    
    private void doReplace(){
        int start=component.getSelectionStart();
        int end=component.getSelectionEnd();
        if(start==end){//nothing selected..nothing to replace.
            return;
        }
        if(!component.isEditable()){
            return;
        }
        String replace=replaceField.getText();
        component.replaceSelection(replace);
        component.setCaretPosition(start+replace.length());
        return;
    }
    
    private void doReplaceAll(){
        String pattern=patternField.getText();
        String replace=replaceField.getText();
        if(pattern==null || pattern.equals("")){//invalid find...
            Toolkit.getDefaultToolkit().beep();
            setLabelFound(false);
            return;
        }
        if(!component.isEditable()){
            return;
        }
        if(!pattern.equals(lastPattern)){//replace it in DocumentFind
            find.setPattern(pattern);
            lastPattern=pattern;
        }
        
        try{
            int result;
            int offset=0;
            do{
                result=find.findNext(offset);
                offset=result+pattern.length();
                if(result<0){
                    break;
                }
                if(debug)System.out.println("REPLACE ALL: result:"+result+" offset:"+offset);
                component.select(result,offset);
                component.replaceSelection(replace);
            }while(result>=0);
        }
        catch(BadLocationException e){}
    }

    public void actionPerformed(ActionEvent e){
        String command=e.getActionCommand();
        if(command.equals(NEXT)){
            doFindNext();
            return;
        }//end next
        else if(command.equals(REPLACE)){
            doReplace();
            return;
        }//end replace
        else if(command.equals(REPLACE_ALL)){
            doReplaceAll();
            return;
        }
    }
    
    private void setLabelFound(boolean f){
        if(f){
            foundLabel.setText(EMPTY_NOT_FOUND);
        }
        else{
            foundLabel.setText(NOT_FOUND);
        }
    }
    
    public static void main(String[] argv){
        JFrame f=new JFrame("A random Document (find test)");
        final JTextPane p=new JTextPane();
        JButton b=new JButton("Show find dialog");
        b.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent e){
                FindFrame.showFind(p);
            }
        });
        Container c=f.getContentPane();
        c.add(b,BorderLayout.NORTH);
        c.add(new JScrollPane(p),BorderLayout.CENTER);
        f.setSize(400,400);
        f.setVisible(true);
    }

}