package edu.gwu.servlets;
import javax.servlet.*;
import javax.servlet.http.*;
import java.io.*;
import java.util.*;
/** 
 * Every servlet extends this one. It maintains several useful
 * references in static variables that can be used across all
 * servlets.
 *
 * @version 2, 6/2003
 */
public abstract class BaseServlet extends HttpServlet {

  // Static data (one copy only) available to all servlets.
  /**
   * The copy of {@link StaticContext} available to BaseServlet extenders.
   */
  protected static StaticContext staticContext = new StaticContext();
  
  //How access it provided to the other subclassed servlets.
  private static int numActiveServlets = 0;
  private static Vector servletRegistry = new Vector();

  //constructor
  public BaseServlet(){
    addServlet(this);
    if(debugOn()){
        writeDebug("BaseServlet: Added "+this+" to servletRegistry");
    }
  }
  
  // Static methods.
  /**
   *
   * @param bs The <code>BaseServlet</code>
   */
  private synchronized static void addServlet(BaseServlet bs){
    servletRegistry.addElement (bs);
    numActiveServlets ++;
  }
  
  /**
   * @param bs The <code>BaseServlet</code> to be removed from the registry.
   */
  private synchronized static boolean removeServlet(BaseServlet bs){
    numActiveServlets --;
    return servletRegistry.removeElement (bs);
  }
  
  /**
   * Gives access to all of the other <code>BaseServlet</code> subclasses on 
   * the current JVM.
   * @return An {@link java.util.Enumeration} of all of the <code>BaseServlet</code>s.
   */
  public static Enumeration servlets(){
    return servletRegistry.elements();
  }


  // Helper methods that are inherited.
  /**
   * Outputs a message to the debug log.  Should check with
   * {@link #debugOn} before calling this method.
   * For normal logging calls, use {@link #writeLog}.
   * @see #debugOn
   * @see #writeLog
   * @param str The String to be written to the debug log.
   */
  public static void writeDebug (String str)
  {
    StaticContext.writeDebug (str);
  }
  
  /**
   * tells whether debugging is turned on or not.
   * @return <code>boolean</code> <code>true</code> if debugging is on, <code>false</code> otherwise.
   */
  public static boolean debugOn ()
  {
    return StaticContext.debugOn();
  }

  /**
   * Writes to a log using {@link StaticContext}, which performs buffering, etc.
   * @param str The String to output to the log.
   */
  public static void writeLog (String str)
  {
    StaticContext.writeLog (str);
  }

  /**
   * Cleans up BaseServlets.  This method cannot be overwritten,
   * so subclasses should use {@link #shutdown} instead.
   * @see #shutdown
   */
  final public void destroy ()
  {
    if(debugOn()){
        writeDebug(this+" being destroyed");
    }
    shutdown();
    removeServlet(this);
    if(numActiveServlets==0){
        //so that it only gets shutdown once, at the end.
        StaticContext.shutdown();
    }
  }

  // Abstract methods.
  /**
   * Perform any stopping related housework here.
   * This method will be called if shutdown occurs normally (but may not be, incase of a server crash).
   */
  protected abstract void shutdown ();
  
  /**
   * Reset your servlet with this method call.  After this call, your servlet should be in the
   * same state as when it was started up.
   */
  protected abstract void reset ();
  /**
   * Perform any saving related housework here.  This method will not be called except if the implementing
   * subclass calls it (from {@link #shutdown} for example.)
   */
  protected abstract void save ();
  
  /**
   * Print out a status message here.  This method is called by other {@link BaseServlet} subclasses
   * who access other children via {@link #servlets}.
   */
  abstract public void test (PrintWriter out);
}
