package edu.gwu.servlets;
/*
 * @(#)ConnectionPool
 *
 * Copyright (c) 1998 Karl Moss. All Rights Reserved.
 *
 * You may study, use, modify, and distribute this software for any
 * purpose provided that this copyright notice appears in all copies.
 *
 * This software is provided WITHOUT WARRANTY either expressed or
 * implied.
 *
 * @author  Karl Moss
 * @version 1.0
 * @date    11Mar98
 *
 * For some reason, this code does not work with mysql. Will try removing the
 * destruction of connections - instead just close them. RS.
 *
 */

import java.sql.*;
import java.io.*;
import java.util.*;

/**
 * This class handle the actual implementation of the connection pooling.
 * Do not interact directly with <code>ConnectionPool</code>, instead use
 * {@link StaticContext} or {@link dbBroker}.
 */
class ConnectionPool implements TimerListener
{
  DebugListener debug;

  // JDBC Driver name
  String m_JDBCDriver;

  // JDBC Connection URL
  String m_JDBCConnectionURL;

  // User
  String m_JDBCUser;

  // Password
  String m_JDBCPassword;

  // Minimum size of the pool
  int m_ConnectionPoolSize;

  // Maximum size of the pool
  int m_ConnectionPoolMax;

  // Maximum number of uses for a single connection, or -1 for
  // none
  int m_ConnectionUseCount = -1;

  // Maximum connection idle time (in minutes)
  int m_ConnectionTimeout_seconds;

  int m_ConnectionTimeout_millis;

  // The Connection pool. This is a vector of ConnectionObject
  // objects
  java.util.Vector m_pool;

  // The maximum number of simultaneous connections as reported
  // by the JDBC driver
  int m_MaxConnections = -1;

  // Our Timer object
  Timer m_timer;
  
  /**
    * <p>Initializes the ConnectionPool object
    */
  public ConnectionPool (
			 DebugListener debug,
			 String m_JDBCDriver,
			 String m_JDBCConnectionURL,
			 String m_JDBCUser,
			 String m_JDBCPassword,
			 int m_ConnectionPoolSize,
			 int m_ConnectionPoolMax,
			 int m_ConnectionTimeout_seconds
                        ) 
    throws dbException
  {
    this.debug = debug;
    this.m_JDBCDriver = m_JDBCDriver;
    this.m_JDBCConnectionURL = m_JDBCConnectionURL;
    this.m_JDBCUser = m_JDBCUser;
    this.m_JDBCPassword = m_JDBCPassword;
    this.m_ConnectionPoolSize = m_ConnectionPoolSize;
    this.m_ConnectionPoolMax = m_ConnectionPoolMax;
    this.m_ConnectionTimeout_seconds = m_ConnectionTimeout_seconds;
    this.m_ConnectionTimeout_millis = m_ConnectionTimeout_seconds * 1000;

    createPool();
    
    // Start our timer so we can timeout connections. The
    // clock cycle will be 20 seconds
    m_timer = new Timer(this, 20);
    m_timer.start();
  }

  void writeDebug (String s)
  {
    if (debug != null)
      debug.debugWrite (s);
  }

  /**
    * <p>Destroys the pool and it's contents. Closes any open
    * JDBC connections and frees all resources
    */
  public synchronized void destroy()
    {
      trace("destroying connections");
      try {

        // Stop our timer thread
        if (m_timer != null) {
          m_timer.stopThread();
          m_timer = null;
        }

        synchronized(m_pool){
            // Clear our pool
            if (m_pool != null) {
    
                // Loop throught the pool and close each connection
                for (int i = 0; i < m_pool.size(); i++) {
                    ConnectionObject co = (ConnectionObject) m_pool.elementAt(i);
                    if (co != null) {
                        if (co.con != null) {
                            try {
                                co.con.close();
                            }
                            catch (Exception e) {
                            // Ignore.
                            }
                        }
                        co = null;
                    }
                    trace("destroyed connection #"+i);
                }
            }
        }
        m_pool = null;
      }
      catch (Exception ex) {
        trace(ex.getMessage());
      }
    }

  /**
    * <p>Gets an available JDBC Connection. Connections will be
    * created if necessary, up to the maximum number of connections
    * as specified in the configuration file.
    *
    * @return JDBC Connection, or null if the maximum
    * number of connections has been exceeded
    */
  public synchronized java.sql.Connection getConnection()
    {
      // If there is no pool it must have been destroyed
      if (m_pool == null) {
        trace("ConnectionPool.getConnection(): Pool destroyed");
        return null;
      }

      java.sql.Connection con = null;
      ConnectionObject connectionObject = null;
      int poolSize = m_pool.size();

      // Get the next available connection
      for (int i = 0; i < poolSize; i++) {

        // Get the ConnectionObject from the pool
        ConnectionObject co = (ConnectionObject) m_pool.elementAt(i);

        // If this is a valid connection and it is not in use,
        // grab it
        if (co.isAvailable()) {
          connectionObject = co;
          break;
        }
      }

      writeDebug ("ConnPool.getC: poolsize=" + poolSize 
		  + " max=" + m_ConnectionPoolMax + " connObj="  + connectionObject);

      // No more available connections. If we aren't at the
      // maximum number of connections, create a new entry
      // in the pool
      if (connectionObject == null) {
        if ((m_ConnectionPoolMax < 0) ||
            ((m_ConnectionPoolMax > 0) &&
             (poolSize < m_ConnectionPoolMax))) {
          
          // Add a new connection.
          int i = addConnection();
          
          // If a new connection was created, use it
          if (i >= 0) {
            connectionObject = (ConnectionObject) m_pool.elementAt(i);
          }
        }
        else {
          trace("Maximum number of connections exceeded");
	  return null;
        }
      }

      // If we have a connection, set the last time accessed,
      // the use count, and the in use flag
      if (connectionObject != null) {
        connectionObject.inUse = true;
        connectionObject.useCount++;
        touch(connectionObject);
        con = connectionObject.con;
      }
      
      return con;
    }

  /**
    * <p>Places the connection back into the connection pool</p>
    * 
    *
    * @param Connection object to close
    */
  public synchronized void close(java.sql.Connection con)
    {
      // Find the connection in the pool
      int index = find(con);
      
      if (index != -1) {
        ConnectionObject co = (ConnectionObject) m_pool.elementAt(index);
	// Clear the use count and reset the time last used
	touch(co);
	co.inUse = false;
      }
    }
  
  /**
    * <p>Prints the contents of the connection pool to the
    * standard output device
    */
  public void printPool (PrintWriter out)
    {
      out.println("--ConnectionPool--");
      if (m_pool != null) {
        for (int i = 0; i < m_pool.size(); i++) {
          ConnectionObject co = (ConnectionObject)
            m_pool.elementAt(i);
          out.println("<li>" + i + "=" + co);
        }
      }
      out.flush();
    }
  
  
  /**
    * <p>Creates the initial connection pool. A timer thread
    * is also created so that connection timeouts can be
    * handled.
    *
    * @return true if the pool was created
    */
  private void createPool() throws dbException
    {
      // Sanity check our properties
      if (m_JDBCDriver == null) {
        throw new dbException("JDBCDriver property not found");
      }
      if (m_JDBCConnectionURL == null) {
        throw new dbException("JDBCConnectionURL property not found");
      }
      if (m_ConnectionPoolSize < 0) {
        throw new dbException("ConnectionPoolSize property not found");
      }
      if (m_ConnectionPoolSize == 0) {
        throw new dbException("ConnectionPoolSize invalid");
      }
      if (m_ConnectionPoolMax < m_ConnectionPoolSize) {
        trace("WARNING - ConnectionPoolMax is invalid and will " +
              "be ignored");
        m_ConnectionPoolMax = -1;
      }
      if (m_ConnectionTimeout_millis < 0) {
        // Set the default to 10 seconds
        m_ConnectionTimeout_millis = 10000;
      }

      // Dump the parameters we are going to use for the pool.
      // We don't know what type of servlet environment we will
      // be running in - this may go to the console or it
      // may be redirected to a log file
      trace("JDBCDriver = " + m_JDBCDriver);
      trace("JDBCConnectionURL = " + m_JDBCConnectionURL);
      trace("ConnectionPoolSize = " + m_ConnectionPoolSize);
      trace("ConnectionPoolMax = " + m_ConnectionPoolMax);
      trace("ConnectionUseCount = " + m_ConnectionUseCount);
      trace("ConnectionTimeout_millis = " + m_ConnectionTimeout_millis +
            " seconds");

      // Attempt to create a new instance of the specified
      // JDBC driver. Well behaved drivers will register
      // themselves with the JDBC DriverManager when they
      // are instantiated
      trace("Registering " + m_JDBCDriver);
      try {
	java.sql.Driver d = (java.sql.Driver)
	  Class.forName(m_JDBCDriver).newInstance();
      }
      catch (Exception e) {
	throw new dbException ("ConnPool: Driver not found");
      }
      
      // Create the vector for the pool
      m_pool = new java.util.Vector (m_ConnectionPoolMax);

      // Bring the pool to the minimum size
      fillPool (m_ConnectionPoolSize);//MG: was m_ConnectionPoolMax, fixed.
    }

  /**
    * <p>Adds a new connection to the pool
    *
    * @return Index of the new pool entry, or -1 if an
    * error has occurred
    */
  private int addConnection()
    {
      int index = -1;

      try {
        // Calculate the new size of the pool
        int size = m_pool.size() + 1;

        // Create a new entry
        fillPool (size);

        // Set the index pointer to the new connection if one
        // was created
        if (size == m_pool.size()) {
          index = size - 1;
        }
      }
      catch (dbException ex) {
        ex.printStackTrace();
      }
      return index;
    }
  
  /**
    * <p>Brings the pool to the given size
    */
  private synchronized void fillPool(int size) throws dbException
    {
      // Loop while we need to create more connections
      while (m_pool.size() < size) {
        
        ConnectionObject co = new ConnectionObject();

        // Create the connection
	try {
	  co.con =
	    java.sql.DriverManager.getConnection(m_JDBCConnectionURL,
						 m_JDBCUser,
						 m_JDBCPassword);
	  // System.out.println ("FillPool: mpool.size=" + m_pool.size() + " size=" + size + " co.con=" + co.con);	
        }
	catch (SQLException e2) {
	  System.out.println ("In Exception");
	  throw new dbException ("connPool.fillPool: couldn't get first connection. (original Exception:"+e2.getMessage()+")");
	}

        // Do some sanity checking on the first connection in
        // the pool
        if (m_pool.size() == 0) {

          // Get the maximum number of simultaneous connections
          // as reported by the JDBC driver
	  try {
	    java.sql.DatabaseMetaData md = co.con.getMetaData();
	    m_MaxConnections = md.getMaxConnections();
	  }
	  catch (SQLException e) {
	  }

        }

        // Give a warning if the size of the pool will exceed
        // the maximum number of connections allowed by the
        // JDBC driver
        if ((m_MaxConnections > 0) &&
            (size > m_MaxConnections)) {
          trace("WARNING: Size of pool will exceed safe maximum of " +
                m_MaxConnections);
        }

        // Clear the in use flag
        co.inUse = false;
        
        // Set the last access time
        touch(co);
        
        m_pool.addElement(co);
      }
    }

  /**
    * <p>Find the given connection in the pool
    *
    * @return Index into the pool, or -1 if not found
    */
  private int find (java.sql.Connection con)
    {
      int index = -1;

      synchronized(m_pool){
        // Find the matching Connection in the pool
        if ((con != null) && (m_pool != null)) {
            for (int i = 0; i < m_pool.size(); i++) {
                ConnectionObject co = (ConnectionObject) m_pool.elementAt(i);
                if (co.con == con) {
                    index = i;
                    break;
                }
            }
        }
      }
      return index;
    }
  
  /**
    * <p>Called by the timer each time a clock cycle expires.
    * This gives us the opportunity to timeout connections
    */
  public synchronized void TimerEvent(Object object)
    {
      // No pool means no work
      if (m_pool == null) {
        return;
      }

      // Get the current time in milliseconds
      long now = System.currentTimeMillis();
      synchronized(m_pool){
        // Check for any expired connections and remove them
        for (int i = m_pool.size() - 1; i >= 0; i--) {
            ConnectionObject co = (ConnectionObject) m_pool.elementAt(i);
            // System.out.println ("TimerEvent: check: cobj=" + co);
            // If the connection is not in use and it has not been
            // used recently, remove it
            if (!co.inUse) {
                if ((m_ConnectionTimeout_millis > 0) && (co.lastAccess + (m_ConnectionTimeout_millis) < now)) {
                    //System.out.println (">>Remove cobj=" + co);
                    try {
                        //Remove the old connection.
                        writeDebug  ("TimerEvent: Expired connection: cobj=" + co);
                        int index=find(co.con);
                        m_pool.remove(index);
                        co.con.close();
                        //Previously a new connection was reestablished here, but this functionality has been
                        //removed.  This allows a high number of connections to be kept with the database when
                        //load is high, but when load dimishes it will drop unused connections.
                    }
                    catch (SQLException e) {
                        // Not much we can do here, but give up.
                        trace ("Error: could not create new connection in TimerEvent.  More Info:");
                        trace (e.getMessage());
                        while((e=e.getNextException()) != null){
                            trace(e.getMessage());
                        }
                    }
                }
            }
        }
        
    
    
        // Now ensure that the pool is still at it's minimum size
        try {
            if (m_pool != null) {
                if (m_pool.size() < m_ConnectionPoolSize) {
                    //System.out.println (">> Calling fillPool: mpoolsize=" + m_pool.size());
                    fillPool (m_ConnectionPoolSize);
                }
            }
        }
        catch (dbException ex) {
            ex.printStackTrace();
        }
      }//synchronized block
    }
  
  /**
    * <p>Sets the last access time for the given ConnectionObject
    */
  private void touch (ConnectionObject co)
    {
      if (co != null) {
        co.lastAccess = System.currentTimeMillis();
      }
    }
  
  /**
    * <p>Trace the given string
    */
  private void trace(String s)
    {
      writeDebug ("ConnectionPool: " + s);
    }

}

// This package-private class is used to represent a single
// connection object
class ConnectionObject
{
  // The JDBC Connection
  public java.sql.Connection con;

  // true if this connection is currently in use
  public boolean inUse;

  // The last time (in milliseconds) that this connection was used
  public long lastAccess;

  // The number of times this connection has been used
  public int useCount;

  /**
    * <p>Determine if the connection is available
    *
    * @return true if the connection can be used
    */
  public boolean isAvailable()
    {
      boolean available = false;

      try {

        // To be available, the connection cannot be in use
        // and must be open
        if (con != null) {
          if (!inUse &&
              !con.isClosed()) {
            available = true;
          }
        }
      }
      catch (SQLException ex) {
      }
      
      return available;
    }
  
  /**
    * <p>Convert the object contents to a String
    */
  public String toString()
    {
      return "Connection=" + con + ",inUse=" + inUse +
        ",lastAccess=" + lastAccess + ",useCount=" + useCount;
    }

}
