package edu.gwu.servlets;
import javax.servlet.*;
import javax.servlet.http.*;
import java.io.*;
import java.util.*;
import java.sql.Connection;

/** 
 * <p>This class is intended to maintain a static context available
 * to all servlets that extend {@link BaseServlet}. The class contains 
 * a DB Connection pooler, a threaded log writer, a debug information writer,
 * a security handler and a threaded properties reader.</p>
 * <p>To use StaticContext, the {@link PropFileLocator#propFileDir}, 
 * and {@link PropFileLocator#propFileName} variables 
 * in {@link PropFileLocator} need to be configured
 * to point to the properties configuration file.
 *
 * @author Rahul Simha
 * @author Michael Gaiman
 * @version 1.1 June 2003
 *
 */
public class StaticContext implements PropertyChangeListener{
  /**
   * propSleep controls how often the properties file is reread for changes.  Recommended value is 10000,
   * which is every 10 seconds.
   */
  static final int propSleep = 10000; // Every 10 seconds
  
  //Do not configure other values below...
  
  static ThreadedPropertyReader propReader;
  static Thread propThread;
  static Vector notificationRegistry=null;

  // Log file stuff:
  static ThreadedLogWriter logWriter = null;
  static Thread logThread;
  
  //defaults
  static final boolean DEBUG_ON_DEFAULT=true;
  static final String DEBUG_FILE_NAME_DEFAULT="debug.txt";
  static final String WRITE_DIR_DEFAULT="";

  // Static data (one copy only) available to all servlets.
  static boolean debugOn = DEBUG_ON_DEFAULT;
  static String debugFileName = DEBUG_FILE_NAME_DEFAULT;

  static Date startDate = new Date ();

  static String writeDir = WRITE_DIR_DEFAULT;

  // Dbase classes.
  static dbBroker db = null;
  private static Thread dbBThread;

  // Security.
  static ThreadedSecurityHandler securityHandler = null;
  private static Thread tshthread;
  
  static ThreadedEmailSender emailSender = null;
  static Thread emailThread;


  // Static initializer.  Sets everything up.
  static {
    try {

      // First,  prepare property notification registry, then load up properties.
      notificationRegistry=new Vector();
      File propFile = null;
      try{
        propFile = new File (PropFileLocator.propFileDir + PropFileLocator.propFileName);
      }
      catch(Exception eProp){
        //write to the debug file, which will output to the default file in the default place.
        System.out.println("edu.gwu.servlets:  Cannot find properties file.  Please check your"+
                   " PropFileLocator.java file, or consult the documentation for more help.");
      }
      if(!propFile.exists()){
        System.out.println("edu.gwu.servlets:  Cannot find properties file.  Please check your"+
                   " PropFileLocator.java file, or consult the documentation for more help.");
      }
      else{//always write just a little info the the debug file.
        System.out.println("edu.gwu.servlets: Properties File ("+propFile+") Found.");
      }
      
      //Loading properties file using getResource (where propFileName was hardcoded above):
      //java.net.URL propURL=StaticContext.class.getResource("/"+propFileName);
      //File propFile = new File(propURL.getFile());
      
      propReader = new ThreadedPropertyReader (propFile, propSleep);
      propThread = new Thread (propReader);
      propThread.start ();
      
      // Where do we write logs, debug etc?
      validateWriteDir(propReader.getStringProperty (SCProperties.writeDir));

      // Debugging stuff
      validateDebugOn(propReader.getStringProperty (SCProperties.debugOn));
      validateDebugFileName(propReader.getStringProperty (SCProperties.debugFileName));
      if(debugOn()){
        FileWriter fw = new FileWriter (writeDir + debugFileName);
        PrintWriter pw = new PrintWriter (fw, true);
        pw.println ("DEBUG FILE: Date = " + startDate);
        pw.close();
      }

      //From here, don't validate any of the properties, that is the responsability
      //of the property users.

      // Create the log file.
      String logFileName = propReader.getStringProperty (SCProperties.logFileName);
      int logFlushSize = propReader.getIntProperty (SCProperties.logFlushSize);
      int logMaxSize = propReader.getIntProperty (SCProperties.logMaxSize);
      int logSleepSeconds = propReader.getIntProperty (SCProperties.logSleepSeconds);
      logWriter = new ThreadedLogWriter (writeDir, logFileName, logFlushSize, logMaxSize, logSleepSeconds);
      logThread = new Thread (logWriter);
      logThread.start ();
      
      if (debugOn)
	writeDebug ("StaticContext: log File created");

      // Next, the database connections.
      db = new dbBroker ();
      dbBThread=new Thread(db);
      dbBThread.start();
      if (debugOn)
	writeDebug ("StaticContext: db broker created");
      writeLog ("StaticContext: db broker created");

      // Passwords etc.
      String passwordFileName = propReader.getStringProperty ("passwordFileName");
      int securitySleepTimeSeconds = propReader.getIntProperty ("securitySleepTimeSeconds");
      securityHandler = new ThreadedSecurityHandler (passwordFileName, securitySleepTimeSeconds);
      tshthread = new Thread (securityHandler);
      tshthread.start();
      writeLog ("StaticContext: Security Handler created");
      
      //Email sender
      int emailSleepSeconds = propReader.getIntProperty (SCProperties.emailSleepTimeSeconds);
      emailSender = new ThreadedEmailSender (emailSleepSeconds);
      emailThread = new Thread (emailSender);
      emailThread.start ();      
      if (debugOn)
	writeDebug ("StaticContext: email Sender created");
    }
    catch (Exception e) {}
  }
  
  /**
   * Creates a new StaticContext object.
   */
  public StaticContext(){
    //Register for the properties you want to know about.
    registerForProperty(SCProperties.debugOn, this);
    registerForProperty(SCProperties.debugFileName, this);
    registerForProperty(SCProperties.writeDir, this);
  }

  /**
   * Resets the database connection pool, causing all connections to be destroyed
   * and then remade.
   */
  public static void resetConnectionPool ()
  {
    // At this time, the only way is to destroy what exists and re-create.
    db.destroy();
    // Create again.
    db = new dbBroker ();
    writeLog ("StaticContext: RECREATED connection pool after resetConnectionPool method call.");
  }
  
  /**
   * Used to obtain a connection to the database from the connection pool.
   * The pool must already be established.  A connection should be freed after use.
   * @return a <code>Connection</code> object if available, or <code>null</code> if unavailable.
   * @see #freeSQLConnection(Connection)
   * @see java.sql.Connection
   */
  public static Connection getSQLConnection(){
    return db.getConnection();
  }
  
  /**
   * Used to return a database connection to the connection pool.
   * The connection need not be valid upon return.
   * @param c The <code>Connection</code> to be freed.
   */
  public static void freeSQLConnection(Connection c){
    db.freeConnection(c);
  }
  
  /**
   * Checks to see if a password is correct for the user.
   * @return <code>boolean true</code> if the password is correct.  <code>false</code>
   * if the password is incorrect or if the user doesn't exist.
   * @param user The username.
   * @param pwtest The password to check.
   */
  public static boolean checkPassword (String user, String pwtest){
    return securityHandler.checkPassword(user,pwtest);
  }
  
  /**
   * Tests to see if a password is used by any user.
   * @return <code>true</code> if the password exists <code>false</code> otherwise.
   * @param pwtest The password to check.
   * @see SCProperties#passwordFileName
   */
  public static boolean passwordExists (String pwtest){
    return passwordExists(pwtest);
  }
  
  /**
   * Write a message to the debug log.
   * The debug log outputs immediately, with no buffering.
   * You should check to see if debugging is turned first.
   * The can be accomplished by calling {@link debugOn()}.
   * @param str The String to write to the debug log.
   * @see debugOn()
   */
  synchronized static void writeDebug (String str)
  {
    try {
      FileWriter fw = new FileWriter (writeDir+debugFileName, true);
      PrintWriter pw = new PrintWriter (fw, true);
      pw.println (str);
      pw.flush();
      pw.close ();
      fw.close ();
    }
    catch (IOException e) { }
    // writeDrasticDebug (str);
  }

  /**
   * tells whether debugging is turned on or not.
   * @return <code>boolean true</code> if debugging is on, <code>false</code> otherwise.
   */
  synchronized static boolean debugOn ()
  {
    return debugOn;
  }
  /**
   * Writes to a log using {@link ThreadedLogWriter}, which performs buffering, etc.
   * @param str The String to output to the log.
   * @see ThreadedLogWriter
   */
  synchronized static void writeLog (String str)
  {
    logWriter.appendLog (str);
  }

  /**
   * Allows access to all of the properties given in the StaticContext configuration file.
   * @see SCProperties
   * @return {@link java.util.Enumeration} of all the properties.
   */
  public static Enumeration propertyNames ()
  {
    return propReader.propertyNames ();
  }
  
  /**
   * Allows for polling for property changes.
   * @param propName The property you wish to know the value of.
   * @return The value of the property.
   * @see SCProperties
   */
  public static String getProperty (String propName)
  {
    return propReader.getStringProperty (propName);
  }
  
  /**
   * Used to register for a property change notification.  Generally called from the constructor.
   * @param property The property you wish to receive notification of.
   * @param requester The {@link PropertyChangeListener} implementor that should be called when a change happens.
   * @see SCProperties
   * @see PropertyChangeListener
   */
  static void registerForProperty(String property, PropertyChangeListener requester){
    if(debugOn()){
        writeDebug("StaticContext: "+requester+" has registered to be notified for "+property);
    }
    notificationRegistry.add(new NotificationRequest(property, requester));
  }
  
  /**
   * From this method requesters will receive their <code>PropertyChangeNotification()</code> calls.
   * @see SCProperties
   * @see PropertyChangeListener
   * @param property The property that changed.
   * @param value The new value of the changed property.
   */
  static void propertyHasChanged(String property, String value){
    synchronized(notificationRegistry){
        int size=notificationRegistry.size();
        NotificationRequest nr=null;
        //search through the registry looking for objects that want to be notified
        for(int i=0;i<size;i++){
            nr=(NotificationRequest)notificationRegistry.elementAt(i);
            if(nr.property.equals(property)){
            //if they want this property, tell them of the change.
                if(debugOn()){
                    writeDebug("StaticContext: found "+nr.requester+" who wants to know about this property");
                }
                nr.requester.propertyChangeNotification(property, value);
            }
        }
    }
  }
    
  //StaticContext Personal Property Changes...
  /**
   * StaticContext's personal property change notification.  Called when a property
   * it has registered to be alerted to changes.
   * @param property The property is cares about.
   * @param value The new value of the property.
   */
  public void propertyChangeNotification(String property, String value){
    if(debugOn()){
        writeDebug("StaticContext: Property Change Notification: Property "+property+" value="+value);
    }
    
    if(property.equals(SCProperties.debugOn)){
        validateDebugOn(value);
    }
    else if(property.equals(SCProperties.debugFileName)){
        validateDebugFileName(value);
    }
    else if(property.equals(SCProperties.writeDir)){
        validateWriteDir(value);
    }
  }
  //validate methods make sure the new value of a property is valid.
  //if it is not valid, it reverts to a default value.
  static void validateDebugOn(String value){
    if(value==null){
        debugOn=DEBUG_ON_DEFAULT;
        return;
    }
    if(value.equalsIgnoreCase("true")){
        debugOn=true;
        return;
    }
    if(value.equalsIgnoreCase("false")){
        debugOn=false;
        return;
    }
    //here it is an invalid boolean. Use Default.
    debugOn=DEBUG_ON_DEFAULT;
  }
  
  static void validateDebugFileName(String value){
    if(value==null){
        debugFileName=DEBUG_FILE_NAME_DEFAULT;
    }
    else{
        debugFileName=value;
    } 
  }
  static void validateWriteDir(String value){
    if(value==null){
        writeDir=WRITE_DIR_DEFAULT;
    }
    else{
        writeDir=value;
    }
  }
  /**
   * Shuts down the various services it is running.
   * First the property change checking, then the security handler.
   * Next comes the database connections, and finally logging support goes down.
   */
  public synchronized static void shutdown ()
  {
    if(debugOn()){
        writeDebug("StaticContext shutting down.");
    }
    // Stop the threads.
    propReader.stop ();
    securityHandler.stop ();
    db.destroy();//destroy the connections to the database.
    logWriter.stop ();
  }

}

/**
 * The inner datastructure used to hold Property Notification requests.
 * @author Michael Gaiman
 * @see PropertyChangeListener
 */
class NotificationRequest{
    public String property;
    public PropertyChangeListener requester;
    public NotificationRequest(){}
    public NotificationRequest(String p, PropertyChangeListener r){
        property=p;
        requester=r;
    }
}
