package edu.gwu.servlets;
import java.io.*;
import java.util.*;

/**
 * File: ThreadedLogWriter.java
 *
 * Handles the periodic writing of log files.  Do not interact directly with this
 * class, instead use {@link StaticContext} or {@link BaseServlet}.
 */

class ThreadedLogWriter implements Runnable, PropertyChangeListener {

  //Defaults (for use with PropertyChange)
  static final int FLUSH_SIZE_DEFAULT=100; // 100 lines
  static final int MAX_FILE_SIZE_DEFAULT=10000; // 10000 lines
  static final String BASE_NAME_DEFAULT="BaseServletLog";
  static final int SLEEP_TIME_MILLIS_DEFAULT=60000; // 1 minutes
  static final String WRITE_DIR_DEFAULT="";//current dir
  
  static final int bufSize = 1000;

  StringBuffer buf;
  int flushSize = FLUSH_SIZE_DEFAULT;
  int maxFileSize = MAX_FILE_SIZE_DEFAULT; 

  String writeDir=WRITE_DIR_DEFAULT;
  String baseName=BASE_NAME_DEFAULT;
  
  String logFileName;
  int currentLinesInBuffer = 0;
  int currentLinesWrittenOut = 0;
  int fileCounter = 0;

  boolean stopped = false;
  int sleepTimeMillis = SLEEP_TIME_MILLIS_DEFAULT;

  public ThreadedLogWriter (String writeDir, String baseName, int flushSize, int maxFileSize, int sleepTimeSeconds)
  {
    validateLogFlushSize(Integer.toString(flushSize));
    validateLogMaxSize(Integer.toString(maxFileSize));
    validateWriteDir(writeDir);
    validateLogFileName(baseName);
    validateLogSleepSeconds(Integer.toString(sleepTimeSeconds));
    buf = new StringBuffer (bufSize);
    //makeNewLogFileName (); Not needed anymore--validateLogFileName calls it.
    registerForProperties();
  }
  
  //Property Related:
  void registerForProperties(){
    if(StaticContext.debugOn()){
        StaticContext.writeDebug("ThreadedLogWriter: Registering Properties");
    }
    StaticContext.registerForProperty(SCProperties.writeDir,this);
    StaticContext.registerForProperty(SCProperties.logFileName, this);
    StaticContext.registerForProperty(SCProperties.logFlushSize, this);
    StaticContext.registerForProperty(SCProperties.logMaxSize, this);
    StaticContext.registerForProperty(SCProperties.logSleepSeconds, this);
  }
  
  //needed to implement PropertyChange
  public void propertyChangeNotification(String property, String value){
    
    if(StaticContext.debugOn()){
        StaticContext.writeDebug("ThreadedLogWriter: Property Change Notification: Property "+property+" value="+value);
    }
    
    if (property.equals(SCProperties.writeDir)){
        validateWriteDir(value);
    }
    else if(property.equals(SCProperties.logFileName)){
        validateLogFileName(value);
    }
    else if(property.equals(SCProperties.logFlushSize)){
        validateLogFlushSize(value);
        //after current sleep is over, this change will be propagated
    }
    else if(property.equals(SCProperties.logMaxSize)){
        validateLogMaxSize(value);
        //after current sleep is over, this change will be propagated
    }
    else if(property.equals(SCProperties.logSleepSeconds)){
        validateLogSleepSeconds(value);
        //change will be propagated once current sleep is done.
    }
  }
  
  void validateWriteDir(String value){
    if(value==null){
        writeDir=WRITE_DIR_DEFAULT;
    }
    else{
        writeDir=value;
    }
    fileCounter=0;
    makeNewLogFileName();
    currentLinesWrittenOut=0;
  }
  
  void validateLogFileName(String value){
    if(value==null){
        baseName=BASE_NAME_DEFAULT;
    }else{
        baseName=value;
    }
    fileCounter=0;
    makeNewLogFileName();
    currentLinesWrittenOut=0;
  }
  void validateLogFlushSize(String value){
    if(value==null){
        flushSize=FLUSH_SIZE_DEFAULT;
        return;
    }else{
        try {
            flushSize = Integer.parseInt (value);
        }
        catch (NumberFormatException e) {
            // If there's a problem, use the default
            flushSize=FLUSH_SIZE_DEFAULT;
            return;
        }
        
        //now do sanity check on value.
        if(flushSize<0 || flushSize>100000){
            flushSize=FLUSH_SIZE_DEFAULT;
            return;
        }
    }
  }
  
  void validateLogMaxSize(String value){
    if(value==null){
        maxFileSize=MAX_FILE_SIZE_DEFAULT;
        return;
    }else{
         try {
            maxFileSize = Integer.parseInt (value);
        }
        catch (NumberFormatException e) {
            // If there's a problem, use the default
            maxFileSize=MAX_FILE_SIZE_DEFAULT;
            return;
        }
        
        //now do sanity check on value.
        if(maxFileSize<100){
            maxFileSize=MAX_FILE_SIZE_DEFAULT;
            return;
        }
    }
  }
  
  void validateLogSleepSeconds(String value){
    int sleepTimeSeconds=0;
    if(value==null){
        sleepTimeMillis=SLEEP_TIME_MILLIS_DEFAULT;
        return;
    }else{
        try {
            sleepTimeSeconds = Integer.parseInt (value);
        }
        catch (NumberFormatException e) {
            // If there's a problem, use the default
            sleepTimeMillis=SLEEP_TIME_MILLIS_DEFAULT;
            return;
        }
        
        //now do sanity check on value.
        if(sleepTimeSeconds<1 || sleepTimeSeconds>600){
            sleepTimeMillis=MAX_FILE_SIZE_DEFAULT;   //sets sleep time to be ten seconds (why not use default sleep time?)
            return;
        }
        //move valid sleepTime from seconds to MilliSeconds
        sleepTimeMillis = sleepTimeSeconds * 1000;
    }
  }
  
  private void makeNewLogFileName () 
  {
    do{
    fileCounter ++;
    logFileName = writeDir + baseName + fileCounter;
    }while((new File(logFileName)).exists());
    if(StaticContext.debugOn())
        StaticContext.writeDebug("ThreadedLogWriter's new log file is "+logFileName);
  }
  

  public synchronized void appendLog (String s)
  {
    buf.append (s);
    buf.append ("\n");
    currentLinesInBuffer +=countLines(s);
  }
  
  int countLines(String s){
    if (s==null) return 0;
    LineNumberReader lr=new LineNumberReader(new StringReader(s));
    int lines=0;
    try{
        String ls=lr.readLine();
        while(ls!=null){
            lines++;
            ls=lr.readLine();
        }
    }catch (IOException e) {
      e.printStackTrace ();
    }
    return lines;
  }
  
  private void writeOut () 
  {
    PrintWriter pw=null;
    try {
        // Get file ready for appending.
        FileWriter logFileWriter = new FileWriter (logFileName, true);
        pw = new PrintWriter (logFileWriter, true);
        
        //Give some date information
        Date current=new Date();
        pw.println(current+":");
        currentLinesWrittenOut++;
        
        // Get String from buffer
        String toOut=buf.toString();
        
        
        LineNumberReader lr=new LineNumberReader(new StringReader(toOut));
        String aLine=lr.readLine();
        while(aLine!=null){
            pw.println(aLine);
            currentLinesWrittenOut++;
            aLine=lr.readLine();

            // See if new log file needs to be created.
            if (currentLinesWrittenOut > maxFileSize) {
                // Create new file.
                makeNewLogFileName ();
                currentLinesWrittenOut=0;
                
                //open a new stream from it so output can continue.
                pw.flush();
                pw.close();
                logFileWriter=new FileWriter(logFileName,true);
                pw=new PrintWriter(logFileWriter,true);
                
                //Give some date information
                current=new Date();
                pw.println(current+":");
                currentLinesWrittenOut++;
            }
        }//endwhile
        lr.close();
        // Clear buffer.
        buf = new StringBuffer (bufSize);
        currentLinesInBuffer = 0;
    }
    catch (IOException e) {
      e.printStackTrace ();
    }
    finally{
        if(pw!=null){
            pw.flush();
            pw.close();
        }
    }
  }


  public void stop ()
  {
    if(StaticContext.debugOn()){
        StaticContext.writeDebug("ThreadedLogWriter has been asked to stop.");
    }
    stopped = true;
  }

  
  public void run ()
  {
    while (! stopped) {
        if (currentLinesInBuffer >= flushSize) {
            writeOut ();
        }
        // Sleep
        try {
            Thread.sleep (sleepTimeMillis);
        }
        catch (InterruptedException e) {}

    } //endwhile
    //before leaving, dump rest of buffer to log.
    if(StaticContext.debugOn()){
        StaticContext.writeDebug("ThreadedLogWriter is dumping remaining log contents.");
    }
    if(currentLinesInBuffer>0){
        writeOut();
    }
    if(StaticContext.debugOn()){
        StaticContext.writeDebug("ThreadedLogWriter stopped.");
    }
  }
}
