package edu.gwu.servlets;
import java.io.*;
import java.util.*;

/**
 * Class <code>ThreadedPropertyReader</code> is intended for
 * periodic re-reading of a given properties file. This allows
 * a programmer to make changes to a system while it is running
 * via the properties file.
 *
 */
class ThreadedPropertyReader implements Runnable {

  int sleepTimeMillis = 10000;
  boolean isComplete = false;
  File propFile = null;
  boolean stopped = false;

  Properties props = null;

  public synchronized boolean isComplete () 
  {
    return isComplete;
  }


  synchronized void setComplete (boolean isComplete)
  {
    this.isComplete = isComplete;
  }
  

  public ThreadedPropertyReader (File propFile, int sleepTimeMillis)
    throws FileNotFoundException
  {
    if (sleepTimeMillis > 1000)
      this.sleepTimeMillis = sleepTimeMillis;
    if (propFile == null) {
      throw new FileNotFoundException ("ThreadPropertyReader: propFile null");
    }
    this.propFile = propFile;

    readProps ();
  }
  

  public void stop ()
  {
    stopped = true;
  }


  // To implement Runnable:

  public void run ()
  {
    while (! stopped) {

      setComplete (false);

      readProps ();

      setComplete (true);

      // Sleep
      try {
        Thread.sleep (sleepTimeMillis);
      }
      catch (InterruptedException e) {
        System.out.println ("ThreadedPropertyReader: sleep exception in run()");
        // Not much else you can do, since this is early on. Log file is not
        // yet created.
      }
    } // endwhile
    
  }
  
  private void readProps ()
  {
    Properties newProps = new Properties ();
    try {
      FileInputStream fis = new FileInputStream (propFile);
      // Load the properties
      newProps.load (fis);
      fis.close();
    }
    catch (IOException e) {
      System.out.println ("ThreadedPropertyReader.readProps(): Could not load Config file. Exception:"+e);
    }
    if((props!=null) && (!props.equals(newProps))){
    //a change has happened...find which properties have changed
        StaticContext.writeLog("PropertyReader: Property change detected.");
        findChanged(newProps);
        props = newProps;
    }
    else{
        props = newProps;
    }
  }
  
  //finds which properties have changed and lets StaticContext know
  void findChanged(Properties newProps){
    //first check the new properties for changes
    Enumeration names=newProps.propertyNames();
    String key;
    String value;
    while(names.hasMoreElements()){
        key=(String)names.nextElement();
        value=newProps.getProperty(key);
        if(!(value.equals(props.getProperty(key)))){
        //will work if it is a new property.
            if(StaticContext.debugOn()){
                StaticContext.writeDebug("ThreadedPropertyReader: Property="+key+" changed from "
                +props.getProperty(key)+" to "+value);
            }
            
            StaticContext.propertyHasChanged(key,value);
        }
    }
    //now check the old properties for removals
    names=props.propertyNames();
    while(names.hasMoreElements()){
        key=(String)names.nextElement();
        if(newProps.getProperty(key)==null){
        
            if(StaticContext.debugOn()){
                StaticContext.writeDebug("ThreadedPropertyReader: Property="+key
                +" has been removed from property file.");
            }
        
            StaticContext.propertyHasChanged(key,null);
        }
    }
  }
    

  // Public Methods.
  public String getStringProperty (String prop)
  {
    return props.getProperty (prop);
  }

  public boolean getBooleanProperty (String prop)
  {
    String val = props.getProperty (prop);
    if (val == null)
      return false;
    if (val.equalsIgnoreCase ("true"))
      return true;
    else 
      return false;
  }

  public int getIntProperty (String prop)
  {
    String val = props.getProperty (prop);
    if (val == null)
      return -1;
    try {
      int k = Integer.parseInt (val);
      return k;
    }
    catch (NumberFormatException e) {
      // If there's a problem, return -1
      return -1;
    }
  }

  public double getDoubleProperty (String prop)
  {
    String val = props.getProperty (prop);
    if (val == null)
      return -1;
    try {
      Double D = Double.valueOf (val);
      return D.doubleValue();
    }
    catch (NumberFormatException e) {
      // If there's a problem, return -1
      return -1;
    }
  }

  public Enumeration propertyNames ()
  {
    return props.propertyNames();
  }

  public void writeProperties (PrintWriter out)
  {
    out.println ("ThreadedPropertyReader: PROPERTIES");

    if (props != null) {
      Enumeration e = props.propertyNames(); 
      while (e.hasMoreElements()) {
	String prop_name = (String) e.nextElement();
	String prop_val = props.getProperty (prop_name);
	out.println ("  " + prop_name + "=" + prop_val);
      }
      out.println ("");
    }
    else {
      out.println ("... props pointer is null");
    }
  }
  

}

