package edu.gwu.servlets;
/**
 * File: SecurityHandler.java
 *
 * Reads from a password file and checks passwords.
 * NOTE: implementation is only temporary; a more secure version
 * is desired, with an encrypted property file.
 */

import java.io.*;
import java.util.*;

class ThreadedSecurityHandler implements Runnable, PropertyChangeListener {

  static final String PASSWORD_FILE_NAME_DEFAULT="pw.props";
  static final int SLEEP_TIME_SECONDS_DEFAULT=100;
  
  String passwordFileName=PASSWORD_FILE_NAME_DEFAULT;
  Properties pwprops;
  int sleepTimeSeconds = SLEEP_TIME_SECONDS_DEFAULT;
  boolean stop=false;

  public ThreadedSecurityHandler (String passwordFileName, int sleepTimeSeconds)
  {
    validatePasswordFileName(passwordFileName);
    validateSleepTimeSeconds(Integer.toString(sleepTimeSeconds));
    
    StaticContext.registerForProperty(SCProperties.passwordFileName, this);
    StaticContext.registerForProperty(SCProperties.securitySleepTimeSeconds, this);
    loadProps();
  }

  public void propertyChangeNotification(String property, String value){
    if(property.equals(SCProperties.passwordFileName)){
      validatePasswordFileName(value);
    }
    else if(property.equals(SCProperties.securitySleepTimeSeconds)){
      validateSleepTimeSeconds(value);
    }
  }
  
  void validatePasswordFileName(String value){
    if(value==null){
      StaticContext.writeLog("No passwordFileName provided.  Using Default.");
      passwordFileName=PASSWORD_FILE_NAME_DEFAULT;
    }
    else if(!(new File(value).exists())){
      StaticContext.writeLog("Invalid passwordFileName.  Using Default.");
      passwordFileName=PASSWORD_FILE_NAME_DEFAULT;
    }
    else{
      passwordFileName=value;
    }
    //now load the new passwords...
    loadProps();
  }
  
  void validateSleepTimeSeconds(String value){
    if(value==null){
      StaticContext.writeLog("securitySleepTimeSeconds is not a valid integer.  Using default.");
      sleepTimeSeconds=SLEEP_TIME_SECONDS_DEFAULT;
      return;
    }
    
    int parsedValue=0;
    try{
      parsedValue=Integer.parseInt(value);
    }
    catch(NumberFormatException e){
      StaticContext.writeLog("securitySleepTimeSeconds is not a valid integer.  Using default.");
      sleepTimeSeconds=SLEEP_TIME_SECONDS_DEFAULT;
      return;
    }
    
    if(parsedValue<1 || parsedValue>600){//less than 1 second, greater than 10 minutes.
      StaticContext.writeLog("Invalid securitySleepTimeSeconds value.  Using default.");
      sleepTimeSeconds=SLEEP_TIME_SECONDS_DEFAULT;
      return;
    }
    
    sleepTimeSeconds=parsedValue;
  }
    

  private synchronized void loadProps ()
  {
    Properties tempProps = new Properties ();
    try {
      // ### 
      // Have to code entire path:
      File f = new File (passwordFileName);
      FileInputStream fis = new FileInputStream (f);

      // Load the properties
      tempProps.load (fis);
      if((pwprops==null) || (!tempProps.equals(pwprops))){
        if (StaticContext.debugOn()) {
            StaticContext.writeDebug ("New Passwords: ("+new Date()+")");
            Enumeration e = tempProps.propertyNames(); // To start, get a new Enumeration.
            while (e.hasMoreElements()) {
            String user = (String) e.nextElement();
            String pw = tempProps.getProperty (user);
            StaticContext.writeDebug ("user=" + user + " pw=" + pw);
            }
        }
      }
      pwprops=tempProps;
    }
    catch (IOException e) {
      StaticContext.writeLog ("SecurityHandler: COULD NOT LOAD PASSWORD FILE");
    }
  }

  /**
   * Checks to see if a password is correct for the user.
   * @return <code>boolean true</code> if the password is correct.  <code>false</code>
   * if the password is incorrect or if the user doesn't exist.
   * @param user The username.
   * @param pwtest The password to check.
   */
  public synchronized boolean checkPassword (String user, String pwtest)
  {
    String pw = pwprops.getProperty (user);
    if (pw == null) {
      StaticContext.writeLog ("SecurityHandler: user=" + user + " not found");
      return false;
    }
    if (! pw.equalsIgnoreCase (pwtest) ) {
      StaticContext.writeLog ("SecurityHandler: password supplied for user=" + user + " not correct.");
      return false;
    }
    return true;
  }

  /**
   * Tests to see if a password is used by any user.
   * @return <code>true</code> if the password exists <code>false</code> otherwise.
   * @param pwtest The password to check.
   */
  public synchronized boolean passwordExists (String pwtest)
  {
    if (!pwprops.contains (pwtest)) {
      StaticContext.writeLog ("SecurityHandler: password=" + pwtest + " does not exist");
      return false;
    }
    return true;
  }

  public void stop(){
    if(StaticContext.debugOn()){
        StaticContext.writeDebug("ThreadedSecurityHandler has been asked to stop");
    }
    stop=true;
  }

  public void run ()
  {
    while (!stop) {
      // Sleep for 100 seconds
      try {
	Thread.sleep (sleepTimeSeconds * 1000);
      }
      catch (InterruptedException e) {
	StaticContext.writeLog ("Interrupted Exception e=" + e + " in securityHandler thread");
      }

      // Reload properties.
      loadProps();
    }//end while
  }
}
