package edu.gwu.servlets;
import java.sql.*;
import java.net.*;
import java.io.*;

/**
 * <code>dbBroker</code> is the component of <code>StaticContext</code> that handles
 * the database connection pooling.  All interaction with the database should actually be
 * accomplished through {@link edu.gwu.servlets.StaticContext} rather than through <code>
 * dbBroker</code> directly.
 */

class dbBroker extends StaticContext implements DebugListener, PropertyChangeListener, Runnable {

  // Some defaults.
  static final String JDBC_DRIVER_DEFAULT="com.mysql.jdbc.Driver";
  static final String JDBC_URL_DEFAULT="jdbc:mysql://localhost:3306/";
  static final String JDBC_DB_DEFAULT="test";
  static final String JDBC_USER_DEFAULT="defaultUser";
  static final String JDBC_PASSWORD_DEFAULT="";
  static final int JDBC_MIN_CONNECTIONS_DEFAULT=5;
  static final int JDBC_MAX_CONNECTIONS_DEFAULT=10; 
  static final int JDBC_CONNECTION_TIMEOUT_DEFAULT=600;//10 minutes
  static final int JDBC_CONNECTION_TRIALS_DEFAULT=10; 
  
  String jdbcDriver=JDBC_DRIVER_DEFAULT;
  String jdbcURL=JDBC_URL_DEFAULT;
  String jdbcDB=JDBC_DB_DEFAULT;
  String jdbcUser=JDBC_USER_DEFAULT;
  String jdbcPassword=JDBC_PASSWORD_DEFAULT;
  int jdbcMinConnections=JDBC_MIN_CONNECTIONS_DEFAULT;
  int jdbcMaxConnections=JDBC_MAX_CONNECTIONS_DEFAULT;
  int jdbcConnectionTimeout=JDBC_CONNECTION_TIMEOUT_DEFAULT;
  int jdbcConnectionTrials=JDBC_CONNECTION_TRIALS_DEFAULT;

  ConnectionPool connpool;
  
  boolean stopped=false; //used with Thread
  int sleepTimeMillis=60000; //1 minute
  
  boolean propertiesChanged=false;

  int numConnections = 0;

  public dbBroker ()
  {
    writeLog ("Setting up DB Connection Broker...");
    validateJdbcDriver(propReader.getStringProperty (SCProperties.jdbcDriver));
    validateJdbcURL(propReader.getStringProperty (SCProperties.jdbcURL));
    validateJdbcDB(propReader.getStringProperty (SCProperties.jdbcDB));
    validateJdbcUser(propReader.getStringProperty (SCProperties.jdbcUser));
    validateJdbcPassword(propReader.getStringProperty (SCProperties.jdbcPassword));
    validateJdbcMinConnections(propReader.getStringProperty (SCProperties.jdbcMinConnections));
    validateJdbcMaxConnections(propReader.getStringProperty (SCProperties.jdbcMaxConnections));
    
    validateJdbcConnectionTimeout(propReader.getStringProperty (SCProperties.jdbcConnectionTimeout));
    validateJdbcConnectionTrials(propReader.getStringProperty (SCProperties.jdbcConnectionTrials));
    
    //register to be notified of changes to our properties.
    registerForProperties();
    
    //make the actual pool of connections
    makeNewPool();
  }  
  
  
  protected void makeNewPool(){
    try {
      writeLog ("Attempting creation of new dbase connection ...");
      
      numConnections=0;
      
      connpool = new ConnectionPool (
                                     this,                   // A debug listener.
                                     jdbcDriver,             // JDBC driver
                                     jdbcURL+jdbcDB,         // Connection URL
                                     jdbcUser,               // User
                                     jdbcPassword,           // Password
                                     jdbcMinConnections,     // Min# conn's at start.
                                     jdbcMaxConnections,     // Max
                                     jdbcConnectionTimeout   // #seconds to keep alive.
                                     );
      writeLog ("Dbase connections created to DB=" + jdbcDB);
      writeLog ("dbBroker: const: created pool");
      writeLog ("dbBroker: testing pool with max connections ...");
      Connection[] con = new Connection[jdbcMaxConnections];
      for (int i=0; i<jdbcMaxConnections; i++) {
        writeLog ("Generating connection " + i);
        con[i] = connpool.getConnection();
      }
      
      for (int i=0; i<jdbcMaxConnections; i++) {
        writeLog ("Freeing connection " + i);
        connpool.close(con[i]);
      }
      writeLog ("dbBroker: test complete");
    }
    catch (dbException e)  { 
      writeLog ("dbBroker const: " + e);
    }
  }
  
  void registerForProperties(){
    if(StaticContext.debugOn()){
        StaticContext.writeDebug("dbBroker: Registering Properties");
    }
    StaticContext.registerForProperty(SCProperties.jdbcDriver,this);
    StaticContext.registerForProperty(SCProperties.jdbcURL, this);
    StaticContext.registerForProperty(SCProperties.jdbcDB, this);
    StaticContext.registerForProperty(SCProperties.jdbcUser, this);
    StaticContext.registerForProperty(SCProperties.jdbcPassword, this);
    StaticContext.registerForProperty(SCProperties.jdbcMinConnections, this);
    StaticContext.registerForProperty(SCProperties.jdbcMaxConnections, this);
    StaticContext.registerForProperty(SCProperties.jdbcConnectionTimeout, this);
    StaticContext.registerForProperty(SCProperties.jdbcConnectionTrials, this);
  }
  
  public void propertyChangeNotification(String property, String value){
    if(StaticContext.debugOn()){
        StaticContext.writeDebug("dbB: Property Change Notification: Property "+property+" value="+value);
    }
    
    if (property.equals(SCProperties.jdbcDriver)){
        if(validateJdbcDriver(value)) propertiesChanged=true;
    }
    else if(property.equals(SCProperties.jdbcURL)){
        if(validateJdbcURL(value)) propertiesChanged=true;
    }
    else if(property.equals(SCProperties.jdbcDB)){
        if(validateJdbcDB(value)) propertiesChanged=true;
    }
    else if(property.equals(SCProperties.jdbcUser)){
        if(validateJdbcUser(value)) propertiesChanged=true;
    }
    else if(property.equals(SCProperties.jdbcPassword)){
        if(validateJdbcPassword(value)) propertiesChanged=true;
    }
    else if(property.equals(SCProperties.jdbcMinConnections)){
        if(validateJdbcMinConnections(value)) propertiesChanged=true;
    }
    else if(property.equals(SCProperties.jdbcMaxConnections)){
        if(validateJdbcMaxConnections(value)) propertiesChanged=true;
    }
    else if(property.equals(SCProperties.jdbcConnectionTimeout)){
        if(validateJdbcConnectionTimeout(value)) propertiesChanged=true;
    }
    else if(property.equals(SCProperties.jdbcConnectionTrials)){
        validateJdbcConnectionTrials(value);
        //no need to restart the ConnectionPool for a change in this property.
    }
  }
  
  //NOTE: the boolean returned is NOT whether the value is valid.  
  //the boolean returned is whether there was a change or not.
  //true if the variable changed, false if it stayed the same.
  boolean validateJdbcDriver(String value){
    String oldValue=jdbcDriver;
    if(value==null){
        writeLog("using default jdbcDriver value");
        jdbcDriver=JDBC_DRIVER_DEFAULT;
    }
    else{
        jdbcDriver=value;
    }
    return !(oldValue.equals(jdbcDriver));
  }
  
  boolean validateJdbcURL(String value){
    String oldValue=jdbcURL;
    if(value==null){
      writeLog("using default jdbcURL value");
      jdbcURL=JDBC_URL_DEFAULT;
    }
    else{
      jdbcURL=value;
    }
    return !(oldValue.equals(jdbcURL));
  }
  
  boolean validateJdbcDB(String value){
    String oldValue=jdbcDB;
    if(value==null){
      writeLog("using default jdbcDB value");
      jdbcDB=JDBC_DB_DEFAULT;
    }
    else{
      jdbcDB=value;
    }
    return !(oldValue.equals(jdbcDB));
  }
  boolean validateJdbcUser(String value){
    String oldValue=jdbcUser;
    if(value==null){
      writeLog("using default jdbcUser value");
      jdbcUser=JDBC_USER_DEFAULT;
    }
    else{
      jdbcUser=value;
    }
    return !(oldValue.equals(jdbcUser));
  }
  
  boolean validateJdbcPassword(String value){
    String oldValue=jdbcPassword;
    if(value==null){
      writeLog("using default jdbcPassword value");
      jdbcPassword=JDBC_PASSWORD_DEFAULT;
    }
    else{
      jdbcPassword=value;
    }
    return !(oldValue.equals(jdbcPassword));
  }
  
  boolean validateJdbcMinConnections(String value){
    
    int oldValue=jdbcMinConnections;
    int parseValue=0;
    if(value==null){
      writeLog("invalid jdbcMinConnections value, using default.");
      jdbcMinConnections=JDBC_MIN_CONNECTIONS_DEFAULT;
    }else{
      try{
        parseValue=Integer.parseInt(value);
      }catch(NumberFormatException e){
        writeLog("invalid jdbcMinConnections value, using default.");
        jdbcMinConnections=JDBC_MIN_CONNECTIONS_DEFAULT;
      }
    
      if(parseValue<0){
        writeLog("invalid jdbcMinConnections, using default value.");
        jdbcMinConnections=JDBC_MIN_CONNECTIONS_DEFAULT;
      }
      else{
        jdbcMinConnections=parseValue;
      }
    }
    
    return oldValue!=jdbcMinConnections;
  }
  
  boolean validateJdbcMaxConnections(String value){
    
    int oldValue=jdbcMaxConnections;
    int parseValue=0;
    if(value==null){
      writeLog("invalid jdbcMaxConnections value, using default.");
      jdbcMinConnections=JDBC_MAX_CONNECTIONS_DEFAULT;
    }
    else{
      try{
        parseValue=Integer.parseInt(value);
      }catch(NumberFormatException e){
        writeLog("invalid jdbcMaxConnections value, using default.");
        jdbcMinConnections=JDBC_MAX_CONNECTIONS_DEFAULT;
      }
    
      if(parseValue<1){
        writeLog("invalid jdbcMaxConnections, using default value.");
        jdbcMaxConnections=JDBC_MAX_CONNECTIONS_DEFAULT;
      }
      else{
        jdbcMaxConnections=parseValue;
      }
    }
    
    return oldValue!=jdbcMaxConnections;
  }
  
  boolean validateJdbcConnectionTimeout(String value){
    int oldValue=jdbcConnectionTimeout;
    int parseValue=0;
    if(value==null){
      writeLog("invalid jdbcConnectionTimeout value, using default.");
      jdbcConnectionTimeout=JDBC_CONNECTION_TIMEOUT_DEFAULT;
    }
    else{
      try{
        parseValue=Integer.parseInt(value);
      }catch(NumberFormatException e){
        writeLog("invalid jdbcConnectionTimeout value, using default.");
        jdbcConnectionTimeout=JDBC_CONNECTION_TIMEOUT_DEFAULT;
      }
    
      if(parseValue<60 || parseValue>3600){ //less then 20 seconds, greater than 1 hour
        writeLog("invalid jdbcConnectionTimeout, using default value.");
        jdbcConnectionTimeout=JDBC_CONNECTION_TIMEOUT_DEFAULT;
      }
      else{
        jdbcConnectionTimeout=parseValue;
      }
    }
    return oldValue!=jdbcConnectionTimeout;
  }
  
  boolean validateJdbcConnectionTrials(String value){
    int oldValue=jdbcConnectionTrials;
    int parseValue=0;
    if(value==null){
      writeLog("invalid jdbcConnectionTrials value, using default.");
      jdbcConnectionTrials=JDBC_CONNECTION_TRIALS_DEFAULT;
    }
    else{
      try{
        parseValue=Integer.parseInt(value);
      }catch(NumberFormatException e){
        writeLog("invalid jdbcConnectionTrials value, using default.");
        jdbcConnectionTrials=JDBC_CONNECTION_TRIALS_DEFAULT;
      }
    
      if(parseValue<1 || parseValue>500){
        writeLog("invalid jdbcConnectionTrials, using default value.");
        jdbcConnectionTrials=JDBC_CONNECTION_TRIALS_DEFAULT;
      }
      else{
        jdbcConnectionTrials=parseValue;
      }
    }
    
    return oldValue!=jdbcConnectionTrials;
  }
  
  
  synchronized void incrConnections ()
  {
    numConnections++;
  }

  synchronized void decrConnections ()
  {
    numConnections--;
  }

  /**
   * @return The number of active (checked out) connections to the database.
   */
  public int getNumConnections ()
  {
    return numConnections;
  }

  /**
   *  Prints some helpful information about the connections.
   */
  public void printConnections (PrintWriter out)
  {
    out.println ("Number of active connections: " + numConnections);
  }

  /**
   * This method is attempts to get a connection to the database.
   * If all connections are in use, it will attempt again until it has 
   * attempted {@link SCProperties#jdbcConnectionTrials} amount (or until
   * it has gotten a connection--which ever comes first.)
   * @return A {@link java.sql.Connection} if available, or <code>null</code>
   * if all connections are in use.
   */
  public Connection getConnection()
  {
    Connection con = null;
    int i = 0;
    if(connpool==null){
      writeDebug("dbB: getConn: connpool is null, returning NULL.");
      writeLog("dbB: getConn: connpool is null, returning NULL.");
      return con;
    }
    while (i < jdbcConnectionTrials) {
      con = connpool.getConnection();
      if (con != null) {
	incrConnections();
	if (debugOn())
	  writeDebug ("dbB: getCONN: con=" + con);
	return con;
      }
      try {
        writeDebug("dbB: getCONN: Pool empty--all connections in use.");
	Thread.sleep (100);
        writeDebug("dbB: getCONN: trying again. ("+i+" of "+jdbcConnectionTrials+")");
      }
      catch (InterruptedException e) {
      }
      i++;
    }
    writeDebug("dbB: getCONN: unable to return a connection.  Returning NULL.");
    writeLog("dbB: getCONN: All Connections are inuse.  Returning NULL.");
    return con;
  }

  /**
   * Returns a {@link java.sql.Connection} back to the connection pool.
   * @param The {@link java.sql.Connection} to return.
   */
  public void freeConnection (Connection con)
  {
    if (debugOn())
      writeDebug  ("dbB: FREE-CONN: con=" + con);
    connpool.close (con);
    decrConnections();
  }

  //used to check to see if a new ConnectionPool should be spawned
  //to do property changes.
  public void run (){
    while(!stopped){
        if(propertiesChanged){
            if(debugOn()){writeDebug("dbB: Starting new ConnectionPool due to property changes.");}
            propertiesChanged=false;
            if(connpool!=null){
              connpool.destroy();//destroy old pool first.
            }
            makeNewPool();
        }
        // Sleep
        try {
            Thread.sleep (sleepTimeMillis);
        }
        catch (InterruptedException e) {}
    }
  }
  
  public void stop(){
    if(StaticContext.debugOn()){
        StaticContext.writeDebug("dbBroker has been asked to stop.");
    }
    stopped=true;
  }

  public void destroy ()
  {
    // Remove connections.
    debugWrite("dbB: Destroying Pool.");
    stop();
    if(connpool!=null){
        connpool.destroy();
    }
  }

  // Implements DebugListener
  public void debugWrite (String s)
  {
    if (debugOn())
      super.writeDebug (s);
  }

}
